# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without
# modifications, as long as this notice is preserved.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# initial revision: 2003-09-29
#
# Plugin for the dependency synthesiser to convert the negra corpus to cdg.
#
# TODO
# - pod this plugin
# - put it into a separate namespace
# - move extralevels into a callback

#------------------------------------------------------------------------------
=head2 Global Variables

=over 10

=item phrase tree conversion callbacks

=cut
#------------------------------------------------------------------------------

%phraseConvTable =
  (
   AP   => \&apConversion,
   AVP  => \&avpConversion,
   CAC  => \&cjConversion,
   CAP  => \&cjConversion,
   CAVP => \&cjConversion,
   CCP  => \&cjConversion,
   CNP  => \&cjConversion,
   CO   => \&cjConversion,
   CPP  => \&cjConversion,
   CS   => \&cjConversion,
   CVP  => \&cjConversion,
   CVZ  => \&cjConversion,
   ISU  => \&ISUConversion,
   MPN  => \&mpnConversion,
   MTA  => \&mtaConversion,
   NP   => \&npConversion,
   NM   => \&nmConversion,
   PP   => \&ppStructureConversion,
   PPER => \&explConversion,
   S    => \&S_conversion,
   VP   => \&vpConversion,

);

=item Regular expression for matching German scopus adverbs

=cut
my $scope_adv_re = join '|',
  qw(
     noch nur schon auch so immer ganz da fast heute je erst dann hier
     weiter nie
    );


=item Regular expression for matching German auxiliary verbs

=cut
my $aux_re = join '|',
  qw(
     habe hast hat habt habet haben
     hatte hattest hattet hatten
     bin bist ist sind seid sind sein
     sei seist seien
     war warst waren wart wre wrst wren wrt
     werde wirst wird werden werdet
     wurde wurdest wurden wurdet
     wrde wrdest wrden wrdet
    );

=item Regular expression for matching German modal verbs

=cut
my $mod_re = join '|',
  qw(
     will willst wollen wollt
     wollte wolltest wollten wolltet
     gewollt

     mu(|ss) mu(|ss)t mssen m(|ss)t
     mu(|ss)te mu(|ss)test mu(|ss)ten mu(|ss)tet
     m(|ss)te m(|ss)test m(|ss)ten m(|ss)tet
     gemu(|ss)t

     darf darfst drfen drft
     durfte durftest durften durftet
     drfte drftest drften drftet
     gedurft

     soll sollst sollen sollt
     sollte solltest sollten solltet
     gesollt

     kann kannst knnen knnt
     konnte konntest konnten konntet
     knnte knntest knnten knntet
     gekonnt

    );

=item Regular expression for matching German separable verb prefixes

=cut
my $avz_re = join '|',
  qw(

     ab allein an aneinander auf aufeinander aufrecht aus auseinander
     bei beisammen beiseite bekannt bereit bergauf bestehen bevor
     blank brach da dabei dagegen daher dahin dar davon dazu dicht
     dran drauf durch durcheinander ein einher entgegen entlang
     entzwei fallen feil fern fertig fest flott fort frei gefangen
     gegen gegenber geheim gerade gleich gro(|ss) gut heim her herab
     heran herauf heraus herbei herber herein herum herunter hervor
     herber hin hinauf hinaus hindurch hinein hinterher hinunter
     hinweg hinzu hinber hoch hof hngen ineinander inne instand kalt
     kaputt kehrt kennen klar lahm laufen leer leid locker los luxus
     mit mitein nach nahe nieder nher offen preis quer raus rein rum
     rund runter rber schlechter schwer selbst sicher spazieren stand
     stark statt stecken stehen still teil tot uff um umher unter
     verloren voll vor voran voraus vorbei vorher vorber wach wahr
     weg weh weiter wett wider wieder wiederauf wunder zu zufrieden
     zugrunde zurecht zurck zusammen zustande zuvor zuwider zweck
     zwei zwischen ber berein berhand brig

    );

#------------------------------------------------------------------------------
=item dependency tree conversion callbacks

=cut
#------------------------------------------------------------------------------
%depConvTable = (
  "ADV" => \&advConversion,
  "APP" => \&appoConversion,
  "AUX" => \&auxConversion,
  "ATTR" => \&attrConversion,
  "CJ" => \&cjLabelConversion,
  "DET" => \&detLabelConversion,
  "EXPL" => \&explLabelConversion,
  "GMOD" => \&gmodConversion,
  "KOM" => \&komConversion,
  "KON" => \&konLabelConversion,
  "KONJ" => \&konjConversion,
  "NEB" => \&nebLabelConversion,
  "OBJC" => \&objcConversion,
  "OBJI" => \&objcConversion,
  "PN" => \&pnConversion,
  "PP" => \&ppLabelConversion,
  "PRED" => \&predLabelConversion,
  "RE" => \&reLabelConversion,
  "REL" => \&relLabelConversion,
  "ROOT" => \&rootConversion,
  "S" => \&sConversion,
  "SUBJ" => \&subjConversion,
  "ZEIT" => \&zeitLabelConversion,
  #"OBJA" => \&objaConversion,
  #"OBJA" => \&objaConversion,
);

#------------------------------------------------------------------------------
=item negra head definitions

format of an entry:
   <nodeTag> => ["r|l", "<nodeTag>-<edgeTag>", "<nodeTag>-<edgeTag>", ...]

=cut
#------------------------------------------------------------------------------
%headTable =
  (
   "--" =>   ["l"], # unassigned
   "AA" =>   ["l", "-HD"], # first head to the left
   "AP" =>   ["l", "-CM", "-HD"], # conjunct to the left or first head to the left

   "AVP" =>  ["l", "-HD", # first head or
              "-PH"], # placeholder to the left

   "CH" =>   ["l"], # chunk


   # Coordinated phrases come under a number of different labels, but
   # they all follow the same rules. In a normal conjunction (`A und
   # B') the first constituent is marked CJ and the second and third
   # together occupy a nested phrase also marked CJ. Since we want the
   # dependency structure A<--und<--B, we take the first CJ node from
   # the left. In structures with two-part conjunctions (`sowohl A als
   # B') the left part is labelled KON and the second part CD, so if
   # there is a CD, we must find it as a head first.

   "CAC" =>  ["l", "-CD", "-CJ"],
   "CAP" =>  ["l", "-CD", "-CJ"],
   "CAVP" => ["l", "-CD", "-CJ"],
   "CCP" =>  ["l", "-CD", "-CJ"],
   "CDP" =>  ["l", "-CD", "-CJ"],
   "CNP" =>  ["l", "-CD", "-CJ"],
   "CO" =>   ["l", "-CD", "-CJ"],
   "CPP" =>  ["l", "-CD", "-CJ"],
   "CS" =>   ["l", "-CD", "-CJ"],
   "CVP" =>  ["l", "-CD", "-CJ"],
   "CVZ" =>  ["l", "-CD", "-CJ"],

   "DL" =>   ["l", "-DH"], # first discourse head to the left
   "ISU" =>  ["l", "-UC"], # first unit to the left
   "MPN" =>  ["l", "-PNC"], # first proper noun component to the left
   "MTA" =>  ["l", "-ADC"], # first token to the left
   "NM" =>   ["l",
              "NN-NMC", # first article, noun, then numbers
              ".*-NMC"],

   "NP" =>   ["l", "(NE|NN|FM|PPER|PIS|PDS|MPN|CNP|TRUNC|PTKANT|ISU)-NK",
              # first noun, substituting pronoun or NP

              "-RE",
              "(NP|PN)-NK",      # "die zuletzt unter Druck geratenen _Titel_" (s884)
              "NM-NK",      # "weitere 200 _Millionen_" (s915)
              "^ADJ.*-NK",  # `der _grosse_'
              "AP-NK",      # "jene ganz _andere_"
              "PIDAT-NK",   # "ein _paar_"
              "-NK"],       # or the first noun kernel that we have
   "PN" =>   ["l", "PNC"],

   "PP" =>   ["l", "KOKOM-CM",
              "^(APPR.*|CAC)-AC",
              "PROAV-PH",
              "CPP-NK"
             ],

   "QL" =>   ["l", "-UC"], # first component to the left
   "S" =>    ["l", "V.*FIN-HD",
              "V.*-HD",
              "-HD",
              "-PD",
#                 "ADJD-PD",
#                 "NP-PD",
#                 "-SB"
             ],

   "VP" =>   ["l", "-HD"], # verb phrase (non-finite)
   "VZ" =>   ["l", '-HD$'], # zu-marked infinitive
  );

#------------------------------------------------------------------------------
=item negra label definition

format:
   1: <dependency label>
   2: <modifier-posTag>-<modifier-edgeTag>
   3: <modifiee-posTag>-<modifiee-edgeTag>
   4: <domiNode-nodeTag>-<domiNode-edgeTag>
=cut
@labelTable =
  (

   # Punctuation always gets the empty label.
   ['', '^\$', '.', '.'],

   # The edge label ROOT does not appear in NEGRA. We insert it during
   # phrase tree conversion in some places to mark constituents as
   # fragments, even though they are usually subordinated by NEGRA
   # because fragments are not allowed there.
   ['S', '.*', 'ROOT', '.*'],

   # "Panal, _glaube_ ich" (s96)
   ['PAR', 'V.FIN-HD', '.', '^S-DH$'],

   # detect nodes explicitly marked by us as fragments,
   # e.g. by S_conversion()
   ['ROOT', '.', '.', '-ROOT'],

   # `_Ja_, das ist unlogisch.' (s1585)
   ['ROOT', '^\$.|PTKANT', '.*', '.*',],

   # `die Patrouillen, ein _Sicherheitsdienst_, der...' (s158)
   ['APP', '(NN|NE|FM|CARD).', '.', '-APP' ],

   # `der Trommler, _das_ ist Maik' (s237)
   ['APP', '(PPER|PDS)-(NK|PH)', 'NN-NK', '-(NK|PH)'],

   # We do not actually use the dependency label RE; each of these
   # assignments is changed to either REl or S later.

   # `da, wo wir angefangen _hatten_' (s94)
   ['RE', '.*', '.*', '-RE'],

   # `dermaen obskur, da selbst Ravel sich amsiert _htte_' (s25)
   ['RE', '(FIN|INF)-HD', '(ADV|PROAV)-HD', 'S-CC'],

   # `gewagte Verbindungen und _Risiken_' (s2)
   ['CJ', '.*', 'KON', '.'],

   # `Folklore, _Rock_, Klassik und Jazz' (s3)
   ['KON', 'KON', '.', '.'],

   # `sie gehen Risiken ein, _versuchen_ ihre Mglichkeiten auszureizen" (s2)
   ['KON', '.', '.', '-C[DJ]'],

   # `Schmutzpartikel _wie_ Fublle wegkicken' (s96)
   ['KOM', 'KOKOM-CM', '.*', '.*'],

   # `in den Jahren _1990_ und 1991' (s18933)
   ['APP', 'CARD', 'NN', 'NP'],#1045

   # `_Anfang_ September beginnt...' (s117)
   ['ZEIT', 'NN-', '.', 'NP-MO'],

   # `_1744_ kam die Husarenreiterei hinzu' (s163)
   ['ZEIT', 'CARD', '^(V|ADJA)', '-MO'],

   # `_Hans_, sag uns das Schwein.' (s1660)
   ['VOK', '.', '.', '-VO$' ],

   # SUBJ
   #
   # both nouns and verbs can be subjects:
   #
   # "Wir gehen in den Zoo"
   # "Da keine akuten Sicherheitsmngel bestehen, ist auch dem
   #  Gutachten zu entnehmen"
   #
   # Where NEGRA labels a phrase "SB" it is almost always right. But
   # the subordination verb--SUBJ-->verb is also caught by the AUX
   # rules, therefore this rule must come earlier.

   # `_Mag_ sein, da einer schchtern ist' (s234)
   ['SUBJC', '^V', '.*', '-SB$'],

   # `Mgen _Puristen_ auch die Nase rmpfen' (s1)
   ['SUBJ', '.*', '.*', '-SB$'],

   # `wenn die Komponisten zu singen _beginnen_, gewinnt die Aufnahme...' (s69)
   ['NEB', '^V.(FIN|INF|IZU)', '.', '-MO$'],

   # `...was den Drusen mehr Parlamentssitze _sicherte_' (s1462)
   ['REL', '^V.FIN', '^VV.*', '-RC$',],

   # `versuchen ihre Mglichkeiten _auszureizen_' (s2)
   ['OBJI', '^V.(INF|IZU)', '^(VV|ADJD)', '-OC$'],

   # "...hat es sich zur Aufgabe gemacht, der Musik zu Ruhm zu _verhelfen_' (s71)
   ['OBJI', '^V.(INF|IZU)', '^V', '-OA$'],

   # `der Impuls, das Gert _auszuschalten_' (s105)
   ['OBJI', '^V.(INF|IZU)', '^NN.*', '(S|VP|VZ)-OC'],

   # `Soll blo einer sagen, ich _wrde_ wahllos fernsehen!' (s89)
   # `...sind sich einig, da Adorno kindliche Zge gezeigt _habe_' (s1924)
   ['OBJC', '^V.(FIN|IMP)', '^(VV|ADJD)', '-OC$'],

   # `Da dieser wenig pointiert _wurde_, das hatten sie geahnt' (s758)
   ['OBJC', '^V.(FIN|IMP)', '^V', '-OA$'],

   # 'der Anla, da thiopien _auseinanderbricht_' (s564)
   ['OBJC', '^V.FIN', '^NN.*', '(S|VP)-OC'],

   # `"dann _schaffen_ wir eben den Klassenerhalt", war Grimm um heitere
   # Einlagen bemht' (s776)
   ['OBJC', '^VVFIN-HD', '^VAFIN-HD', 'S-RS$'],

   # `ein Label, das klassische Musik _produziert_' (s11)
   ['REL', '^V.*-HD$', '.*', '-RC$',],

   # `Philip Glass wurde mit Tonbndern _berschttet_'  (s9)
   ['AUX', '^V.*', '^VA.*', '.*'],

   # `die Produktion wollte sich dem Ethno-Trend _anschlieen_' (s74)
   ['AUX', '^V.*(INF|PP)', '^VM.*', '.*'],

   # `eher eine Minute als eine _Sekunde_' (s111)
   ['CJ', '.', 'KOKOM-CM', '.*'],

   # `von _damals_' (s157)
   ['PN', '^ADV-NK', 'APPR-AC', '.*'],

   # `da er _recht_ hatte' (s101)
   ['ADV', 'ADJD-', 'VAFIN-HD', '-OA$'],

   # `eine eigene _Handschrift_ aufweisen' (s4)
   ['OBJA', '.*', '.*', '-OA$'],

   # `_Hauptsache_, das Kind ist gesund' (s1287)
   ['OBJA', 'NN-PD', 'VVFIN-HD', 'S-OC'],

   # `...den der Konjunkturrckgang das _Amt_ kostete' (s862)
   ['OBJA2', '.*', '.*', '-OA2$'],

   # `den _Arbeiten_ zu einer VerFfentlichung verhelfen' (s7)
   ['OBJD', '.*', '.*', '-DA$'],

   # `der _Tat_ berfhrt' (s179)
   ['OBJG', '.*', '.*', '-OG$'],

   # `ohne da die Arbeitszeit verlngert _werde_' (s916)
   ['NEB', 'FIN-HD','NN-NK','S-MNR'],

   # `in _Schwung_' (s840)
   ['PN', '^(N|FM|P[^T]|CARD|TRUNC)', '^(APP.*|ADJD)-AC', '.*'],

   # `fr _unbedenklich_ halten' (s1107)
   ['PN', 'ADJ[AD]-(NK|CJ)', 'APPR-AC', '.*'],

   # `vor _allem_' (s297)
   ['PN', '^APPR-AC', 'APPR-AC', '.*'],

   # `bis _zum_ Facelifting' (s181)
   ['PN', '^APPRART-AC', 'APPR-AC', '.*'],

   # `nach und von _Sarajewo_' (s706)
   ['PN', '^(N|FM|P[^T]|CARD)', '^APP.*-CJ', '.*'],

   # `Frankfurt am _Main_' (s454)
   ['PN', 'N[NE]-PNC', '^APP.*-PNC', '.*'],

   # `Tonband nach _Tonband_' (s8)
   ['PN', '-UC$', 'APPR-UC', '.'],

   # `_allzu_ spektakulr' (s227)
   ['ADV', '^PTKA-', '.*', '.*'],

   # `fiel mir, offen _gestanden_, ziemlich schwer' (s82)
   ['ADV', '^VVPP', '.*', '.-MO'],

   # `_klar_, die Fans waren zunchst irritiert' (s250)
   ['ADV', '^ADJD-MO$', '^VA.*-HD', '.*'],

   # `...die nicht unbedingt _akademisch_ sein mu' (s13)
   ['PRED', '^ADJD.*', '^VA.*-HD', '-PD'],

   # `Die Klnge sind _voller_ Spielfreude.' (s19)
   # `Die Naivitt wirkt nie _lcherlich_' (s24)
   # `Man wird _zum_ anderen Wesen.' (s204)
   ['PRED', '.*', '.*', '-PD'],

   # `Glass, fr neue Talente sehr _aufgeschlossen_' (S10)
   ['ADV', '^ADJD-HD', '.*', '^AP-.*'],

   # `ein _wenig_ mutig' (s1550)
   ['ADV', '^PIS-NK', '^ADJ', 'NP-MO'],

   # `_Morans_ wichtigste Arbeitsgrundlage' (s32)
   ['GMOD', '.*', '.*', '-(GL|GR|AG)$'],

   # `zur besten aller mglichen _Welten_' (s1374)
   ['GMOD', 'NN-NK', '-HD', 'NP-CC'],

   # `sie gehen gewagte Verbindungen und Risiken _ein_' (s2)
   ['AVZ', '^PTKVZ-', '.*', '.*'],

   # `dreiviertel _zehn_' (s19137)
   ['APP', 'CARD', 'CARD', ''],

   # `80 _Millionen_ Mexikaner' (s663)
   ['DET', 'NN-NMC', 'NN-', '.*'],

   # `_junge_ Komponisten' (s1)
   ['ATTR', 'ADJA|CARD', '.*', '.*'],

   # `_die_ Nase' (s1)
   ['DET', '^(ART|P.*AT)', '.*', '.*'],

   # `_es_ spielt keine Rolle, ob...' (s4)
   ['EXPL', '(PPER|PDS)-PH', '.*', '.*'],

   # `Dafr gibt _es_ Grnde.' (TIGER-s86)
   ['SUBJ', '(PPER|PDS)-EP', '.*', '.*'],

   # `drei _Monate_ spter' (s179)
   ['GRAD', 'NN-', '.*', '-AMS'],

   # `_Warum_ sollen wir den Wohlstand eintauschen?' (s613)
   ['ADV', 'PWAV-(MO|CP)', '.', '.*'],

   # `..._da_ kaum jemand teilnehmen kann.' (s5)
   ['KONJ', 'KOUS|PWAV', '^V.*FIN', '.*'],

   # `_ohne_ da die Arbeitszeit verlngert werde' (s916)
   ['KONJ', 'APPR-AC','FIN-HD','APPR-AC'],

   # `_anstatt_ Komponisten Anerkennung zu verschaffen' (s14)
   ['KONJ', 'KOUI', '^V.*(INF|ZU)-', '.*'],

   # `_obzwar_ auf den ersten Blick waghalsig' (s2388)
   ['KONJ', 'KOUS', '^ADJD-', '.*'],

   # `_zu_ verhelfen' (s7)
   ['PART', '^(PTKZU|APZR)-', '.*', '.*'],

   # `Spekulationen _darber_
   ['PP', '.', 'NN-', '-OP'],

   # `orientieren sich _daran_ (TIGER/s22)
   ['OBJP', '.', '.', '-OP'],

   # `_im_ Crossover-Bereich' (s1)
   ['PP', '^APP.*-(AC|CJ)$', '.*', '^PP.*'],

   # `die ffnung _von_ kommunikativen Beziehungen' (s1827)
   ['PP', '^APPR-AC$', '.*', '-PG$'],

   # `von Zeit _zu_ Zeit' (s637)
   ['PP', '^APP.*-AC', '.*', 'ISU'],

   # `auf das Abbild, _auf_ Photographien oder Malerei' (s137)
   ['PP', '^APP.*-AC', '.*', 'CPP.*'],

   # `Frankfurt _am_ Main' (s474)
   ['PP', '^APP.*-PNC', '.*', 'MPN'],

   # `_daran_ teilhaben' (s5)
   ['PP', 'PROAV', '.*', '.*',],

   # `Arm _in_ Arm' (s2288)
   ['PP', '^APPR-UC', '.', '^ISU-'],

   # `nicht _in_ gleichem Mae, aber sprbar' (s2204)
   ['PP', '^APP.*-AC', '.*', 'CO-MO',],

   # `das Quartett _Uakti_' (s17)
   ['APP', '(NE|NN|FM|PPER)-(NK|CJ)', '(NE|NN|FM)-(NK|PNC|CJ)', '.*'],

   # `es spielt _eben_ keine Rolle' (s4)
   ['ADV', 'ADV|ADJD|V.PP|PTKNEG', '.*', '.*'],

   # `Philip _Glass_' (s9)
   ['APP', 'PNC', '^N.*', '.*'],

   # `The _Manson_ Family' (s42)
   ['APP', '(FM|NE)', '.*', '.*'],

   # `bei jedem "wollen"' (s273)
   ['PN', '.*', '^APP.*-AC', '.*'],

   # All subordinations that match no other rule are labelled ROOT.
   # Since this label does not exist, it clearly indicates that there
   # was a problem translating this word.
   ['ROOT', '.*', '.*', '.*'],


  );

=item general callbacks

=cut
%callbacks = (
#  "initialize" => \&initialize,
#  "finalize" => \&finalize,
  "readSentence" => \&negraReadSentence,
#  "processSentence" => \&negraProcessSentence,
);

#------------------------------------------------------------------------------
=head2 Implementation

=cut

# ----------------------------------------------------------------------
# Callback function after reading entire sentence.
# ----------------------------------------------------------------------
sub negraReadSentence {
  my $nodes = shift;


  # STTS says that forms of `sein'/`haben'/`werden' are always VA,
  # and forms of `mssen'/`knnen' etc. are always VM.

  # However, they say that compounds of these verbs are VV if they
  # are in one word, and VA otherwise. I.e., the tagging guide says
  #
  #  `wir/PPER wollen/VMFIN ihn/PPER loswerden/VVINF'
  #
  # but
  #
  #  `wir/PPER werden/VAFIN ihn/PPER los/PTKVZ'
  #
  # Unfortunately the NEGRA corpus does not follow this rule
  # consistently; `loswerden' is VVINF, but `dabeisein' is VAINF.
  # Therefore we enforce the much simpler convention in the deutsch
  # grammar: All compounds of `sein' etc. are always VA, period.
  for my $w (values %$nodes) {
    next if $$w{id} >=500;
    my $c = $$w{nodeTag};
    my $p = $$w{word};

    if ($c =~ /VV/ &&
        $p =~ /^ ($avz_re) ($aux_re) $/x) {
      debug "Changing VV to VA for word `$p'\n";
      $$w{nodeTag} =~ s/VV/VA/;
    }
    if ($c =~ /VV/ &&
        $p =~ /^ ($avz_re) ($mod_re) $/x) {
      debug "Changing VV to VA for word `$p'\n";
      $$w{nodeTag} =~ s/VV/VA/;
    }
  }
  return 1;
}


# ----------------------------------------------------------------------
# Handle improperly applied MNR tag.
#
# The NEGRA guidelines say explicitly that MNR marks postnominal
# modifiers, such as `[das Haus [in Frankfurt]MNR]'. But over 500 times
# the relation is used for prenominal modifiers, such as `[als Diplomat
# ]MNR hatte [er] das Recht...'. Most of the time these labellings are
# doubtful for other reasons, e.g. the unnecessary discontinuity of 'er
# als Diplomat'. Therefore we try to move the misplaced MNR one level
# up in the tree if this improves things.
#----------------------------------------------------------------------
sub mnrConversion {
  my $n = shift;
  return 0 unless $$n{edgeTag} eq 'MNR';
  my $result = 0;
  my $p = $$n{parent};
  for my $c(@{$$p{children}}) {
    if($$c{min} > $$n{max} &&
       $$c{edgeTag} eq 'NK') {
      debug "Prenominal MNR #$$n{id} detected";
      $result = 1;
      my $r = $$p{parent};
      $$n{edgeTag} = 'MO';
      moveNode $n, $r;
      last;
    }
  }
  return $result;
}

# ----------------------------------------------------------------------
# Embedded phrases starting with `denn', `doch' etc. are treated by
# NEGRA as if the conjunction were absent. It is is included, with the
# label JU, right into the normal phrase:
#
# Eine Wohnung zu finden [PP und zu einem erschwinglichen Preis] ist
# fast unmglich. (NEGRA s485)
#
# To convert this into a proper coordination, we need to raise the
# conjunction so that it does not interfere with the head selection
# among the other members of the phrase.
sub JU_raising {
  my $n = shift;

  my $JU = $$n{children}[0];

  if(defined $$JU{features}{cat} &&
     $$JU{features}{cat} eq 'KON' &&
     $$JU{edgeTag} =~ /^(JU|MO)$/) {
    # If we have a simple case like s485, it is sufficient to raise
    # the `JU' word one step.
    my $p = $$n{parent};


    # But if a noun phrase starts with the junctor, and it is
    # ultimately a child of verb phrase, we have to raise the junctor
    # all the way out of the NP (NEGRA s239).
    if($$n{nodeTag} eq 'NP' && $$n{parent}{id} != -1) {
      while($$p{nodeTag} eq 'NP' && -1 != $$p{id}) {
        $p = $$p{parent};
      }
      $n = $p;
      $p = $$p{parent};
    }

    # If the entire phrase is itself coordinated with another
    # conjunction, we must raise the junctor one more level:
    #
    #  [CS [S Und doch bezaubert ihre Naivitt] und [S wirkt nie lcherlich]]
    #  (NEGRA s24)
    #
    # to get both conjunctions to the right place.
    elsif($$p{nodeTag} =~ /^C/) {
      $n = $p;
      $p = $$p{parent};
    }

    my $new = insertInnerNode $rootNode, $p,
    { nodeTag => 'CS', edgeTag => $$n{edgeTag} };

    debug "Raise initial junctor $$JU{word}";

    $$JU{edgeTag} = 'CD';
    moveNode $JU, $new;
    moveNode $n, $new;
    return 1;
  }
}


sub apConversion {
  my $n = shift;
  my $result = mnrConversion $n;
  $result |= JU_raising $n;
  return $result;
}

#------------------------------------------------------------------------


#----------------------------------------------------------------------
# Change head marker in some AVPs.
#----------------------------------------------------------------------
sub avpConversion {
  my $n = shift;
  my $result = 0;

  # NEGRA says that in `wie immer', `immer' is the head, but we say
  # it's the `wie', since otherwise it would be missing its
  # complement.
  my $lonely_KOKOM = '';
  my $false_head = '';

  for my $c(@{$$n{children}}) {

    if(isLeaf($c) &&
       $$c{edgeTag} eq 'CM' &&
       $$c{features}{cat} eq 'KOKOM') {
      $lonely_KOKOM = $c;
    }
    if(isLeaf($c) &&
       $$c{edgeTag} eq 'HD' &&
       $$c{features}{cat} =~ /ADV|PROAV/) {
      $false_head = $c;
    }
  }

  if($false_head && $lonely_KOKOM) {
    debug "shift head marker in `$$lonely_KOKOM{word} $$false_head{word}'";
    $$false_head{edgeTag} = 'MO';
    $result = 1;
  }

  # The construction `zu wenig, um zu gewinnen' receives a weird
  # nonprojective modelling in NEGRA (s774) where `zu/PTKA' is
  # considered the head of `um zu/PTKZU gewinnen'. We want `zu/PTKA'
  # to modify the adjective, so we raise it.
  my $zu = $$n{children}[0];
  if(defined $$zu{features}{cat} &&
     $$zu{features}{cat} eq 'PTKA' &&
     $$zu{edgeTag} eq 'HD' &&
     $$n{parent}{nodeTag} eq 'AP') {
    debug "Raise zu/PTKA";
    moveNode $zu, $$n{parent};
    $$zu{edgeTag} = 'MO';
    $result = 1;
  }

  # AVP nodes sometimes contain AVC subordinations to indicate that
  # several adverbs together form an adverbial expression, such as
  # `immer wieder', `um so'. We always want such adverbs to be
  # siblings, so we split such nodes into several MO subordinations.
  my $split = 1;
  for my $c(@{$$n{children}}) {
    if($$c{edgeTag} ne 'AVC') {
      $split = 0;
      last;
    }
  }
  if($split) {
    $result = 1;
    for my $c(@{$$n{children}}) {
      my $new = insertInnerNode $rootNode, $$n{parent},
        { nodeTag => $$n{nodeTag}, edgeTag => 'MO' };
      moveNode $c, $new;
    }
    deleteNode $n;
  } else {
    $result |= JU_raising $n;
    $result |= mnrConversion $n;
  }
  return $result;
}


# ----------------------------------------------------------------------
# Convert phrases with KOKOM (`wie'/`als').
#
# This works exactly the same way as PP conversion, except that
# KOKOM may actually take entire sentences as their kernel, so
# we copy the new node's tag from the old node.
#
# Note that a KOKOM can take an entire PP as its kernel:
#
#  [PP wie im Film]
#
# therefore ppStructureConversion calls this first.
# ----------------------------------------------------------------------
sub KOKOMconversion {
  my $n = shift;
  # detect KOKOM edge tag
  my $KOKOM = '';
  for my $c(@{$$n{children}}) {
    if ($$c{edgeTag} eq 'CM') {
      $KOKOM = $c;
      last;
    }
  }

  # is a conversion necessary ?
  return 0 unless $KOKOM;

  debug 'KOKOMconversion';

  my $new = insertInnerNode $rootNode, $n,
    { nodeTag => $$n{nodeTag}, edgeTag => 'CM' };

  for my $c(@{$$n{children}}) {
    if($c != $new && $$c{min} > $$KOKOM{min}) {
      moveNode $c, $new;
    }
  }
  return 1;
}

# ----------------------------------------------------------------------
# Convert PP nodes.
#
# NEGRA annotates prepositional phrases without embedded NP:
#
#  [PP AC/fr NK/viele NK/junge NK/Komponisten]
#
# We reconstruct the PP so that it contains a proper NP:
#
#  [PP AC/fr PN[NP NK/viele NK/junge NK/Komponisten]]
#
# so that the same NP conversion rules apply for PP kernels and other NP.
# ----------------------------------------------------------------------
sub ppStructureConversion {
  my $n = shift;
  my $result = mnrConversion $n;

  $result |= KOKOMconversion $n;
  $result |= JU_raising $n;

  debug 'ppStructureConversion';
  for my $c (@{$$n{children}}) {

    # Pairings of PH/RE, such as for "damit, da..."
    # do not actually contain noun phrases and are handled correctly
    # by the normal rules.
    if ($$c{edgeTag} =~ /^(PH)/){
      return $result;
    }

    # The same goes if the actual kernel is a coordinated PP (s766).
    if ($$c{edgeTag} eq 'NK' && $$c{nodeTag} eq 'CPP'){
      return $result;
    }
  }

  # extract the noun kernel
  my @lowerChildren;
  for my $c (@{$$n{children}}) {
    if ($$c{edgeTag} !~ /^(AC|MO|CM|NG|KON)$/) {
      push @lowerChildren, $c;
    }
  }

  # is a conversion necessary?
  return $result unless @lowerChildren;

  my $new =
    insertInnerNode $rootNode, $n, { nodeTag => 'NP', edgeTag => 'PN' };

  for(@lowerChildren) {
    moveNode $_, $new;
  }

  return 1;
}

#----------------------------------------------------------------------
# Split coordinated adverb phrases.
#
# Adverbs are sometimes co-ordinated without a junctor:
#
#  [AP [CAVP sehr, sehr] eingngig]
#
# In this case we want both adverbs to be siblings, since `sehr'
# doesn't modify `sehr' but `eingngig'. Therefore we split the node
# if no junctor is found.
# ----------------------------------------------------------------------
sub CAVPConversion {
  my $n = shift;
  return 0 unless $$n{nodeTag} eq 'CAVP';
  for my $c(@{$$n{children}}) {
    return 0 if $$c{nodeTag} =~ /^(KON|APPR)$/;
  }

  for my $c(@{$$n{children}}) {
    moveNode $c, $$n{parent};
    $$c{edgeTag} = 'MO';
  }
  deleteNode $n;
  return 1;
}


# ----------------------------------------------------------------------
# Convert coordinated phrases.
#
# NEGRA stuffs all conjuncts into a big phrase. We want to chain the
# elements of the conjunction from left to right, so to make the head
# attachment algorithm do the right thing we ironically have to insert
# more sub-phrases before we can leave the phrase space.
#
# Used for nodeTags CAC,CAP,CAVP,CCP,CNP,CO,CPP,CS,CVP,CVZ.
# ----------------------------------------------------------------------
sub cjConversion {
  my $n = shift;
  my $result = mnrConversion $n;

  return 1 if CAVPConversion $n;

  debug "cjConversion(#$$n{id})";

  # insert nested conjunction nodes while there are more than two children
  if(@{$$n{children}} > 2) {

    my @c = @{$$n{children}};

    # detect coordinations that start with `entweder', etc. In these
    # phrases the second constituent should be the head, not the
    # initial conjunction.
    if($$n{children}[0]{word} =~ /^(entweder|weder|sowohl)$/i) {
      my $first = $$n{children}[0];
      my $second = $$n{children}[1];
      if(!isLeaf $second) {
        moveNode $first, $second;
      }
      # If we left the edge tag as 'CD', `weder' would be selected as
      # the head, but we want it to be subordinated. So we change it
      # to something else.
      $$first{edgeTag} = 'KON';
      shift @c;
    }


    shift @c;
    my $new = insertInnerNode $rootNode, $n,
      { nodeTag=> $$n{nodeTag}, edgeTag => 'CJ' };
    for(@c) {
      moveNode $_, $new;
    }

    cjConversion( $new );
  }
  return 1;
}


# ----------------------------------------------------------------------
# Convert ideosyncratic phrases.
#
# Phrases such as `Na und?' or `ein fr allemal' are labelled as `ISU'
# in NEGRA, and all elements are stuffed into one phrase. As with
# conjunctions, we want to have chains, so we perform the same
# conversion as with coordinated nodes. This is not always the right
# thing to do, but more often than not; and ISU nodes are so rare
# that it is hardly worth expending any more effort.
# ----------------------------------------------------------------------
sub ISUConversion {
  my $n = shift;
  my $result = 0;

  debug "ISUConversion(#$$n{id})";

  # insert nested nodes while there are more than two children
  if(@{$$n{children}} > 2) {
    my @c = @{$$n{children}};
    shift @c;
    my $new = insertInnerNode $rootNode, $n,
      { nodeTag=> $$n{nodeTag}, edgeTag => 'NK' };
    for(@c) {
      moveNode $_, $new;
    }

    ISUConversion( $new );
  }
  return 1;
}


#------------------------------------------------------------------------------
# phrase conversion expl hack:
# move the expletive 'es' one phrase level up
# used for nodeTags PPER-PH
#------------------------------------------------------------------------------
sub explConversion {
  my $n = shift;

  return 0 if $$n{edgeTag} ne "PH";

  debug 'explConversion';
  moveNode $n, $$n{parent}{parent};
  return 1;
}


# ----------------------------------------------------------------------
# NMC marks multi-token numbers. Often this is a reasonable decision,
# such as in `eine Million'. But sequences such as `100 000' are
# tokenizer errors that we undo here.
# ----------------------------------------------------------------------
sub nmConversion {
  my $n = shift;
  my $word;

  # First of all, check whether the tokens all consist of digits and
  # punctuation. We want to splice "60" "000", but not "200" "Millionen".
  for my $c (@{$$n{children}}) {
    return if($$c{word} =~ /[^0-9,\/-]/);
  }

  # Check whether we should include more tokens than the children of
  # this NMC into the new word (cf. s1411).
  my $min = $$n{min};
  while($min > 0) {
    my $l = $$rootNode{leaves}[$min-1];
    debug "Test extranodal word $$l{word}";
    last unless $$l{word} =~ /^\($/;
    last unless $$l{parent}{id} == -1;
    debug "Found extranodal word $$l{word}";
    $min--;
  }
  my $max = $$n{max};
  while($max < $#{$$rootNode{leaves}}) {
    my $l = $$rootNode{leaves}[$max+1];
    debug "Test extranodal word $$l{word}";
    last unless $$l{word} =~ /^\)$/;
    last unless $$l{parent}{id} == -1;
    debug "Found extranodal word $$l{word}";
    $max++;
  }

  # Punctuation that occurs between the words tokens within an NMC
  # phrase should be moved down into the NMC phrase.
  for my $i($min..$max) {
    my $l = $$rootNode{leaves}[$i];
    if($$l{parent}{id} == -1) {
      moveNode $l, $n;
    }
    # assemble the new word
    $word .= $$l{word};
  }
  spliceWords $n, $min, $max,
    { word => $word, edgeTag => 'NK', nodeTag => 'CARD'};
}


# ----------------------------------------------------------------------
# MTA is the node label for nodes that group multi-token adjectives,
# e.g. [NP die/PPER [MTA New/FM Yorker/ADJA] Brse.]
#
# Since they are easily recognizable, we simply convert them to a
# single adjective.
# ----------------------------------------------------------------------
sub mtaConversion {
  my $n = shift;
  my $word = '';
  for my $c(sort {$$a{id} <=> $$b{id}} @{$$n{children}}) {
    $word .= "$$c{word} ";
  }
  $word =~ s/ +$//;

  spliceWords $n, $$n{min}, $$n{max},
    { word => $word, nodeTag => 'ADJA', edgeTag => 'NK' };
}


# ----------------------------------------------------------------------
# When a compound name is compounded with a noun, the NEGRA corpus
# introduces a word boundary inside the name, but not after it:
# 'Jimi'/NE, 'Hendrix-Fans'/NN
# We repair this by making a single word out of it.
# ----------------------------------------------------------------------
sub repair_NE_NN_split {
  my $n = shift;

  my $last = $$rootNode{leaves}[$$n{max}];
  my $prev = $$rootNode{leaves}[$$n{max}-1];

  if($$last{word} =~ /-/ &&
     $$last{nodeTag} eq 'NN' &&
     $$prev{nodeTag} eq 'NE') {

    debug 'repair NE/NN split';

    spliceWords $n, $$n{max}-1, $$n{max},
      { word => "$$prev{word} $$last{word}",
        nodeTag => 'NN',
        edgeTag => 'NK',
        morphTag => $$last{morphTag}}
  }
}

#------------------------------------------------------------------------------
# Construct a subordinated PNC in an MPN. An MPN node with more than two PNC
# direct children is split into a new subtree with a PNC child and an MPN node
# including the rest. The procedure is called recursively by negra2cdg.
#
# There is no restriction for using the procedure
# used for nodeTags MPN with more than two PNC direct children.
#------------------------------------------------------------------------------
sub mpnConversion {
  my $n = shift;

  repair_NE_NN_split $n;

  # is a conversion nesessary?
  return 0 if $#{$$n{children}} < 2;

  debug "MPN conversion";

  # the first PNC becomes the new head
  my $head = $$n{children}[0];

  # the other children are moved down in the tree
  my @rest = @{$$n{children}};
  shift @rest;

  # insert a new node in between
  my $new = insertInnerNode
    $rootNode, $n, { nodeTag => "MPN", edgeTag => "PNC" };

  for(@rest) {
    moveNode $_, $new;
  }

  return 1;
}


# ----------------------------------------------------------------------
# Detect TRUNC followed by nominal material; such constellations are
# tokenizer errors and are mercilessly undone by us.
# ----------------------------------------------------------------------
sub TRUNCconversion {
  my $n = shift;

  my @c = @{$$n{children}};

  for my $c (@c) {
    if ($$c{nodeTag} eq 'TRUNC' && $$c{id} < $$n{max}) {
      # look for following NN or NE
      my $next = $$rootNode{leaves}[$$c{id}+1];
      next unless $$next{features}{cat} =~ /^(NN|NE|FM|XY)$/;

      # concatenate words
      my $word = $$c{word};
      if ($word !~ /-$/) {
        $word .= ' ';
      }
      $word .= $$next{word};

      # steal features from right word
      my $morph = $$next{morphTag};
      my $cat = $$next{nodeTag};

      spliceWords $n, $$c{id}, $$c{id}+1,
        { word => $word, nodeTag => $cat,
          morphTag => $morph, edgeTag => 'NK' };
      return;
    }
  }
}


sub npConversion {
  my $n = shift;
  my $result = TRUNCconversion $n;
  $result |= JU_raising $n;
  $result |= KOKOMconversion $n;
  return $result;
}

# ----------------------------------------------------------------------
# phrase conversion:
# if there is no head in a 'S'entence, we have to split it in two parts.
#
# NEGRA claims that the constituents of a fragmentary sentence can be
# subjects or objects, but we allow them only as fragments.
# ----------------------------------------------------------------------
sub sSplitting {
  my $n = shift;
  my $result = 0;

  # nothing to split if <2 children
  return 0 if ($#{$n->{children}} <1);

  for my $c (@{$$n{children}}) {
    # check if there is a head, for the second rule
    if ($$c{edgeTag} eq 'HD') {
      return 0;
    }
  }

  # if there was no head: split
  my @raiseNodes; #the Nodes to be raised to the root node
  for my $c (@{$$n{children}}) {
    if ($c->{edgeTag} =~ /^(OC|SB|DM|SP|CJ)$/ || $$c{nodeTag} =~ /APPR/) {
      push @raiseNodes, $c;
      debug 'sSplitting';
      $result = 1;
    }
  }

  for(@raiseNodes) {
    moveNode $_, $rootNode;
    convertPhraseTree $_;
  }

  return $result;
}

# ----------------------------------------------------------------------
# Detect [APP] children of S; this label is only used for constituents
# that are really fragments.
# ----------------------------------------------------------------------
sub detectAppFragments {
  my $n = shift;
  my @new_children;
  for my $c(@{$$n{children}}) {
    if($$c{edgeTag} eq 'APP') {
      # Mark the node as fragmentary; label selection will propagate
      # this to a `S' dependency label, and deptree conversion will
      # turn the edge to NIL. Clever, heh?
      $$c{edgeTag} = 'ROOT';
    }
  }
}

sub S_conversion {
  my $n = shift;
  my $result = 0;

  # The edge label `SP' means `don't know whether this is subject or
  # predicate'. We don't allow uncertainty, so we resolve it;
  # `das/PDS' or `was/PWS' are likely predicates, everything else is
  # subjects.
  for my $c(@{$$n{children}}) {
    if($$c{edgeTag} eq 'SP') {
      if($$c{nodeTag} =~ /^P[WD]S$/) {
        $$c{edgeTag} = 'PD';
      } else {
        $$c{edgeTag} = 'SB';
      }
      $result = 1;
    }
  }

  $result |= KOKOMconversion $n;
  $result |= JU_raising $n;
  $result |= detectAppFragments $n;
  $result |= sSplitting $n;
}

sub vpConversion {
  my $n = shift;
  my $result = detectAppFragments $n;
  $result |= mnrConversion $n;
  $result |= KOKOMconversion $n;
  return $result;
}


# Move attachments down the AUX group.
sub AUX_lowering {
  my $e = shift;
  return 0 unless $$e[1] =~ /^(PP|ADV|ZEIT)$/;

  # We want to lower all adverbial attachments to the full verb in two
  # cases:
  #
  # - when the attachment is within the verb bracket, but not at the
  # bottom of the AUX chain. This is detected by finding an adverbial
  # attachment on the right where the AUX chain also forks to the
  # right.
  #
  # - when the attachment is to a verb group in verb-last order, but
  # not at the bottom of the AUX chain. This is detected by finding an
  # adverbial attachment on the left where the AUX chain also forks to
  # the left.
  #
  # These two situations are flagged by the constraints `Adverb am
  # Hilfsverb' and `Adverb am mittleren Hilfsverb', respectively, and
  # we satisfy the constraints by moving the attachment down as far
  # as possible.
  #
  # There are other situations where attachment is still ambiguous -
  # topicalized adverbials and scrambled verb groups, for instance -
  # but in those cases projectivity and topology impose further
  # constraints, so there are no normalization constraints for them.

  my $deps = shift;
  my $foot = shift;

  if($foot) {
    debug "lowering `$$e[0]{word}' for consistency\n";
    moveEdge $e, $foot;
    return 1;
  }
  return 0;
}

sub AUX_raising {
  my $edge = shift;
  my $deps = shift;

  return 0 unless $$edge[1] =~ /^(PP|ADV|ZEIT|KOM)$/;

  #debug "AUX_raising($$edge[0]{word}) ($$edge[0]{id}-$$edge[2]{id})\n";

  my $e0 = $$edge[0]{parent}{dep};


  while($$e0[1] =~ /^(ADV|AUX|PRED|OBJI|SUBJ)$/) {
    $e0 = $$e0[0]{parent}{dep};
  }

#  debug "verb phrase head is `$$e0[0]{word}' ($$e0[0]{id})\n";

  # If a participle is fronted, adverbials that are not fronted are
  # nonprojective and must be raised:
  #
  #  `Selbst besucht hat er ihn nie, aber...' (s31)
  if($$e0[0]{id} > $$edge[2]{id} && $$e0[0]{id} < $$edge[0]{id}) {
    debug
      "Raise attachment `$$edge[0]{word}' to fronted verb `$$edge[2]{word}'";
    moveEdge $edge, $$e0[0];
    return 1;
  }

  # The same goes if if the adverbial itself is fronted:
  #
  #  `Von dort aus wurde es zum Schornstein verlegt' (s87)
  if ($$e0[1] eq 'S') {
    for my $c (@{$$e0[0]{children}}) {
      if ($$c{id} > $$edge[0]{id} &&
          $$c{id} < $$edge[2]{id} &&
          $$e0[0]{id} > $$edge[0]{id} &&
          $$edge[2] != $$e0[0]) {
        debug "Raise fronted adverbial $$edge[0]{word}";
        moveEdge $edge, $$e0[0];
        return 1;
      }
    }
  }

  # If the verb group is scrambled, the adverbial should modify the
  # finite verb:
  #
  # `Wie er mit dieser Einstellung in die SED habe eintreten
  # knnen...' (s1649)
  if($$edge[0]{id} < $$e0[0]{id} &&
     $$edge[2]{id} > $$e0[0]{id}) {
    debug "Raise adverb `$$edge[0]{word}' out of scrambled auxiliary";
    moveEdge $edge, $$e0[0];
    return 1;
  }

  return 0;
}


# normalize attachment to conjoined phrases (cf. KON_Sperre)
sub KON_lowering {
  my $edge = shift;
  if($$edge[0]->{id} > $$edge[2]->{id}) {
    for my $c (@{$$edge[2]->{children}}) {
      my $e = $c->{dep};
      my $l = $$e[1];
      if($l eq 'KON' && $$e[0]->{id} < $$edge[0]->{id}) {
        for my $c2 (@{$$e[0]->{children}}) {
          my $e2 = $c2->{dep};
          my $l2 = $$e2[1];
          if($l2 eq 'CJ' && $$e2[0]{features}{cat} ne 'TRUNC') {
            moveEdge $edge, $$e2[0];
            debug "lower KON attachment\n" ;
            return 1;
          }
        }
      }
    }
  }
  return 0;
}

#------------------------------------------------------------------------------
# Normalize complicated APP groups to simple chains.
sub appoConversion;
sub appoConversion {
  my $edge = shift;

  my $result = 0;

  for my $c (@{$$edge[2]->{children}}) {
    my $e = $c->{dep};
    my $l = $$e[1];
    my $w = $$e[0]->{word};
    my $cat = $$e[0]->{features}{cat};

    if($l =~ /^(APP|KON|CJ)$/ &&
       $$e[0]->{id} < $$edge[0]->{id} &&
       $$e[0]->{id} > $$e[2]->{id}) {
      moveEdge $edge, $$e[0];
      debug "Enforce APP chain\n" ;
      $result = 1;
      appoConversion $edge;
    }
  }

  $result = 1 if KON_lowering $edge;

  return $result;
}

# Enforce consistent adverbial modification in the verb phrase.
#
# Where the German verb phrase is distributed over more than one word,
# it is always questionable whether adverbial modifications should
# modify the finite verb or the full verb. In general, they should
# modify the full verb so that more meaningful statistics about PP-V
# coocurrence can be gained. But sometimes this would lead to
# nonprojective trees that could be avoided with the other reading,
# so we choose the other attachment.
#
# The general attachment rules are:
#
#  1. Adverbials in sentence-initial position modify the finite verb.
#  But if the full verb is fronted and the adverbial is not, it is
#  raised to the finite verb (s31).
#
#  2. Adverbials after the finite verb also modify the full verb.
#
#  3. Adverbials within the verb bracket are attached to the full
#  verb. If this makes them nonprojective to the subject, this is
#  tolerated.
#
#  4. If the verb group is scrambled (s1649), they modify the finite
#  verb instead.
#
# They are enforced by AUX_lowering, which lowers adverbials to the
# foot of the AUX group, and AUX_raising, which maybe raises them to
# the head instead.
sub auxConversion {
  my $edge = shift;
  my $deps = shift;
  my $result = 0;

  #debug "auxConversion $$edge[0]{word}";

  # Find AUX sibling
  my $aux;
  for(@{$$edge[2]{children}}) {
    if($$_{dep}[1] eq 'AUX') {
      $aux = $$_{dep};
      last;
    }
  }

  my $foot;
  if($aux) {

    # Find the foot of the AUX chain
    $foot = $aux;
  STEP:
    while(1) {
      for(@{$$foot[0]{children}}) {
        if($$_{dep}[1] eq 'AUX') {
          $foot = $$_{dep};
          next STEP;
        }
      }
      last STEP;
    }

    # lower all adverbials in the verb bracket to the foot
    for my $c (@{$$edge[2]->{children}}) {
      my $e = $c->{dep};
      if($$aux[0]{id} < $$aux[2]{id} && $$e[0]{id} < $$e[2]{id}
         ||
         $$aux[0]{id} > $$aux[2]{id} && $$e[0]{id} > $$e[2]{id}) {
        $result = 1 if AUX_lowering $e, $deps, $$foot[0];
      }
    }
  }

  # raise those adverbials that need it
  for my $c (@{$$foot[0]{children}}) {
    my $e = $c->{dep};
    $result = 1 if AUX_raising $e, $deps;
  }

  return $result;
}

# ----------------------------------------------------------------------
# Adverbials are also not allowed to modify a conjunction; they
# should always modify the coordinated word instead.
sub lower_into_coordination {
  my $edge = shift;
  if($$edge[2]{id} != -1 &&
     $$edge[2]{features}{cat} eq 'KON') {
    for(@{$$edge[2]{children}}) {
      if($$_{dep}[1] eq 'CJ' &&
         $$_{dep}[0]{id} > $$_{dep}[2]{id}) {
        debug "Lower `$$edge[0]{word}' into coordination";
        moveEdge $edge, $$_{dep}[0];
        return 1;
      }
    }
  }
}


#------------------------------------------------------------------------------
# KOM conversion: if there is a KOKOM without children ("als", "wie")
# modifiying a verb, this KOKOM should be modifiee of the verb, and
# should modify the verb's former modifiee.
sub komConversion {
  my $e = shift;
  my $result = 0;

  if (($$e[0]->{features}{cat} eq 'KOKOM') &&
      !find_label_under($$e[0], 'CJ') &&
      ($$e[2]->{features}{cat}=~/V.FIN/)) {

    debug "Raise KOKOM `$$e[0]{word}' with no CJ dependency\n" ;

    my $p = $$e[2]->{dep}; # the edge of the verb
    moveEdge $e, $$p[2];
    $$e[1] = 'KOM';
    moveEdge $p, $$e[0];
    $$p[1] = 'CJ';
    $result = 1;
  }

  #s1130 "denn je"
  if (($$e[0]->{features}{cat} eq 'KOKOM') &&
      (! $$e[0]->{children}) &&
      ($$e[2]->{features}{cat}=~/ADV/)) {
    my $parent = $$e[2]->{dep}; # the edge of the verb
    moveEdge $e, $$parent[2];   # change modifiee of current edge

    moveEdge $parent, $$e[0]; # let the KOKOM be the modifiee of
    $$parent[1] = 'CJ';       # the verb
    $result = 1;
  }

  $result = 1 if lower_into_coordination $e;

  # correct nonprojective KOM attachment.
  my $p = $$e[2]{dep};
  if (exists $$p[2]{id} && -1 != $$p[2]{id} &&
      $$p[2]{id} > $$p[0]{id} &&
      $$p[2]{id} < $$e[0]{id}) {
    debug "Raise nonprojective KOM $$e[0]{word} from `$$e[2]{word}' to `$$p[2]{word}'";
    moveEdge $e, $$p[2];
    $result = 1;
    &komConversion($e);
  }

  return $result;
}

# ----------------------------------------------------------------------
# Move EDGE up until it modifies a finite verb, if possible.
sub raise_to_finite_verb {
  my $e = shift;
  my $d = $$e[2]{dep};
  while($$d[2]{id} != -1 && $$d[2]{features}{cat} !~ /FIN/) {
    $d = $$d[2]{dep};
  }
  if($$d[2]{id} != -1) {
    debug "Raise nonprojective `$$e[0]{word}' to `$$d[2]{word}'";
    moveEdge $e, $$d[2];
  }
}

#------------------------------------------------------------------------------
sub ppLabelConversion {
  my $edge = shift;
  my $deps = shift;
  return 0 if $$edge[1] ne "PP";
  my $result = 0;
  # correct systematic error: `als' has a top edge called PP,
  # but it should be always showed as KOM.
  if($$edge[0]->{word} =~ /[Aa]ls/){
    $$edge[1] = 'KOM';
    debug "Change `als' label to KOM\n" ;
    komConversion $edge;
    $result = 1;
  }

  # Prepositions attached to prepositions or adverbs should instead
  # modify whatever the other preposition modifies:
  #
  # *[[NP Die Serie] ist [auch] [bald [PP im Fernsehen]]]
  #  [[NP Die Serie] ist [auch] [bald] [PP im Fernsehen]]
  #
  # Only fixed constructions like `vor allem' can attach to other
  # adverbials:
  #
  # [[PP [PP vor allem] im Ausland] gab es hohe Erwartungen]
  if($$edge[2]{features}{cat} =~ /APPR|ADV/) {
    my $is_adverbial = 0;
    for my $c(@{$$edge[0]{children}}) {
      my $e = $$c{dep};
      if($$e[1] eq 'PN' && $$e[0]{features}{cat} eq 'ADJA') {
        $is_adverbial = 1;
        last;
      }
    }
    if(!$is_adverbial) {
      $result = 1;
      my $counter = 0;
      while(-1 != $$edge[2]{id} &&
            $$edge[2]{features}{cat} =~ /APPR|ADV/) {
        moveEdge $edge, $$edge[2]{dep}[2];
        last if($counter++ > $#$deps);
      }
      return 1;
    }
  }

  # If an auxiliary verb has no full verb, and no other complement
  # that might substitute for it, and is not co-ordinated, then it
  # might be a construction where the PP really fills the role of the
  # auxiliary complement:
  #
  # `...damit die Konvois _ins Innere_ knnen.' (s598)
  if($$edge[2]{features}{cat} =~ /VM.+/ &&
    $$edge[2]{dep}[1] !~ /^(KON|CJ)$/) {
    my $auxFound = 0;
    my $maxPPChild = $$edge[0];
    foreach my $child(@{$$edge[2]->{children}}){
      my $childEdge = $child->{dep};
      my $childLabel = $$childEdge[1];
      if($childLabel =~ /^(AUX|AVZ|OBJA|PRED)$/) {
        $auxFound = 1;
        last;
      }
      if($childLabel eq 'PP' && $child->{id} > $maxPPChild->{id}){
        $maxPPChild = $child;
      }
    }
    if (!$auxFound) {
      debug "Altering label $maxPPChild->{dep}[1] to AUX at $maxPPChild->{id}";
      $maxPPChild->{dep}[1] = 'AUX';
      $result =1 ;
    }
  }

  # if a PP modifies an NP to the right, it is almost always wrong.
  # Try to move it up if there is a Verb above it.
  if($$edge[2]{features}{cat} =~ /NN|NE|FM/ &&
     $$edge[2]{word} !~ /nde[rnms]?$/ &&
     $$edge[2]{id} > $$edge[0]{id}) {
    my $parent = $$edge[2]->{dep};
    if(exists $$parent[2]{features} &&
       $$parent[2]{features}{cat} =~ /^V/) {
      debug "Raise left PP from `$$edge[2]{word}' to `$$parent[2]{word}'";
      moveEdge $edge, $$parent[2];
      $result = 1;
    }
  }

  #s948,203 by the verb 'werden' PP's with the prepositions 'aus' and 'zu' are to be PRED
  if($$edge[0]->{word} =~ /^((a|A)us|(z|Z)u[m])$/ &&
     $$edge[2]->{word} =~ /^(werde[n|t]|wird|wirst|w(u|)rde[n|t|st]|[ge]worden|ward)$/)
	{
	 my $found = 0;
	 for my $c (@{$$edge[2]->{children}}) {
	  my $e = $c->{dep};
          my $l = $$e[1];
          if ($l eq 'PRED'){
            $found = 1;
	    last;
	  }
         }
	 if($found != 1){
	  $$edge[1] = 'PRED';
	 }
	}	



  # correct nonprojective attachment
  my $par = $$edge[2]{dep};
  if($$par[2]{id} != -1 &&
     $$par[1] ne 'AUX') { # AUX groups are already treated by auxConversion
    if(($$edge[2]{id} > $$edge[0]{id} &&
        $$par[2]{id} < $$edge[2]{id} &&
        $$par[2]{id} > $$edge[0]{id})
        ||
       ($$edge[2]{id} < $$edge[0]{id} &&
        $$par[2]{id} > $$edge[2]{id} &&
        $$par[2]{id} < $$edge[0]{id})) {
      raise_to_finite_verb $edge;
      $result = 1;
    }
  }



  $result = 1 if KON_lowering $edge;

  return $result;
}


# -----------------------------------------------------------------------------
sub pnConversion {
  my $edge = shift;

  return 0 if $$edge[1] ne "PN";
  my $result = 0;

  # correct systematic error: `als' is often called APPR, but the STTS
  # says that it is always KOKOM.
  if(($$edge[2]->{features}{cat} eq 'APPR') &&
     ($$edge[2]->{word} =~ /[Aa]ls/)){
    $$edge[2]->{features}{cat} = 'KOKOM';
    $$edge[1] = 'CJ';
    debug "Change `als' to KOKOM\n" ;
    komConversion $edge;
    return 1;
  }

  #s1123 "bis zum .. Jahr"
  if($$edge[0]->{features}{cat} eq 'APPRART'){
    foreach my $child(@{$$edge[2]->{children}}){
      my $childEdge = $child->{dep};
      if(($$childEdge[1] eq 'PN') && ($$childEdge[0]->{id} > $$edge[0]->{id})){
        moveEdge $childEdge, $$edge[0];
        $result = 1;
	last;
      }
    }
  }

  #after APPRART all atributives should be marked as ATTR
  #s1123 "bis zum naechsten Jahr"
  if($$edge[2]->{features}{cat} eq 'APPRART'){
    foreach my $child(@{$$edge[0]->{children}}){
      my $childEdge = $child->{dep};
      if(($$childEdge[1] eq 'DET') &&
         ($$childEdge[0]->{id} < $$edge[0]->{id})){
        $$childEdge[1] = 'ATTR';
        $result = 1;
	last;
      }
    }
  }

  # When several PN co-occur, resolve the conflict in one of three
  # ways.
  #
  # (1) `vor _allem_ _in_ den Vereinigten Staaten':
  #
  # This is a normal PP, so `vor' must be `PP'.
  #
  # (2) `_innerhalb_ von zwlf _Jahren_':
  #
  # when prepositions take other prepositions as an argument, the
  # structure must change so that they are subordinated under each other.
  #
  # (3) `von Zeit zu Zeit':
  #
  # This PP has a noun as its argument, so it must be raised to the
  # governor of the other PP.
  foreach my $c(@{$$edge[2]->{children}}) {
    my $e = $$c{dep};
    if($$e[0]{id} != $$edge[0]{id} && 'PN' eq $$e[1]) {
      if($$e[0]{id} < $$e[2]{id}) {
        $$e[1] = 'PP';
        $result = 1;
        ppLabelConversion $e;
      } elsif($$e[0]{id} > $$e[2]{id} &&
              $$e[0]{id} < $$edge[0]{id}) {
        if($$e[0]{features}{cat} =~ /^APPR/) {
          moveEdge($edge, $$e[0]);
          $result = 1;
        } else {
          moveEdge($edge, $$e[2]);
          $$edge[1] = 'PP';
          ppLabelConversion $edge;
          $result = 1;
        }
      }
    }
  }

  if($$edge[2]{id} == -1) {
    $$edge[1] = 'S';
  }

  return $result;
}

# If a word directly or indirectly subordinated under $w belongs to
# the category $cat, it is returned, otherwise ''.
#
# If a third argument is passed, it is a regular expression of labels
# that should stop the search.
sub find_under;
sub find_under {
  my $w = shift;
  my $cat = shift;
  my $scope_re = shift;

  return $w if $w->{features}{cat} =~ /$cat/;

  for my $c (@{$w->{children}}) {
    next if $scope_re && $$c{dep}[1] =~ /$scope_re/;
    my $w = find_under($c, $cat);
    return $w if $w;
  }

  return '';
}

# Return 1 if $w is modified by at least one word with the label $label_re.
sub find_label_under {
  my $w = shift;
  my $label_re = shift;
  for my $c (@{$w->{children}}) {
    return 1 if $$c{dep}[1] =~ /$label_re/;
  }
  return 0;
}

sub nebLabelConversion {
  my $edge = shift;
  my $result = 0;

  # NEB should modify the finite verb if there is one.
  if($$edge[2]{features}{cat} !~ /FIN/) {
    my $p = $$edge[0]{parent}{dep};
    while($$p[1] =~ /AUX/) {
      $p = $$p[0]{parent}{dep};
    }
    if($edge != $p) {
      moveEdge $edge, $$p[0];
      debug "Raise NEB to finite verb `$$p[0]{word}'\n" ;
      $result = 1;
    }
  }

  # NEB should be REL if it carries a PWAV/PWS.
  if(find_under $$edge[0], 'PW', 'OBJC|REL|NEB') {
    debug 'Change NEB to REL';
    $$edge[1] = 'REL';
    $result = 1;
  }

  return $result;
}

#------------------------------------------------------------------------------
# Post-process ROOT edges.
#
# ROOT is not actually a valid dependency label. The label table
# assigns this label to indicate that either the translator or the
# dependency grammar is incomplete, since we could not find a fitting
# label. For simplicity, we force all those untranslatable
# constructions to be fragments.
sub rootConversion {
  my $e = shift;
  debug "Fragment: `$$e[0]{word}'";
  moveEdge $e, $rootNode;
  $$e[1] = 'S';
  return 1;
}

#------------------------------------------------------------------------------
# Post-process S edges.
#
# If the matrix clause is embedded into its own object clause, we
# label it as PAR instead of S, and subordinate it to the next word
# on the left. This is to avoid nonprojective dependencies.
sub sConversion {
  my $e = shift;

  # See if the word labelled S is a finite verb.
  if($$e[0]{id} =~ /FIN$/) {
     return 0;
  }

  # Unfortunately the phenomenon is handled inconsistently by NEGRA.
  # Sometimes the matrix clause is superordinated, so that the phrase
  # tree becomes nonprojective, and we have to change both label and
  # structure (s42, s314). We detect the nonprojectivity by looking
  # for an edge that crosses the word that is supposed to be the root
  # of the sentence.
  my $deps = shift;
  my $x;
 EDGE:
  for my $d (@$deps) {
    if($$d[2]{id} != -1 &&
       ( $$d[0]{id} > $$e[0]{id} && $$d[2]{id} < $$e[0]{id}
         ||
         $$d[0]{id} < $$e[0]{id} && $$d[2]{id} > $$e[0]{id}) ) {

      # Now we must find that daughter edge of $e which ultimately
      # dominates the nonprojective edge, because that is the edge that
      # must change.
      $x = $d;
      while($$x[2]{id} != -1 &&
            $$x[2]{id} != $$e[0]{id} &&
            $$x[2]{word} ne 'so') {
        $x = $$x[2]{dep};
      }

      if($$x[2]{id} == -1) {
        $x = '';
      } elsif ($$x[1] !~ /^(S|OBJC|RE)$/) {
        $x = '';
      } else {
        last EDGE;
      }
    }
  }

  if($x) {

    debug "Change S to PAR";

    # The subordinated S edge becomes the new root...
    moveEdge $x, $$e[2];

    # ...and the former regent must subordinate under the next available
    # word to the left. Available means: the word must be a proper
    # word, not punctuation, and it should not be subordinated under
    # the regent itself -- that would create a cycle (s532).
    my $i = $$e[0]{id} - 1;
    my $w = $$e[0]{parent}{leaves}[$i];
    while($$w{features}{cat} =~ /\$/ || $$w{parent} eq $$e[0]) {
      $i--;
      $w = $$e[0]{parent}{leaves}[$i];
    }
    moveEdge $e, $w;
    $$e[1] = 'PAR';
    return 1;
  }

  # But other times the matrix clause is subordinated as a `discourse
  # head' (DH), so that we need only change S to PAR (s96).
  if($$e[0]{features}{cat} eq 'VVFIN' &&
     $$e[2]{id} != -1 &&
     $$e[2]{features}{cat} !~ '^V') {
    debug "Change S to PAR\n";
    $$e[1] = 'PAR';
    return 1;
  }

  return 0;
}

#------------------------------------------------------------------------------
# Distinguish S, OBJC, OBJI, and NEB. All verb subordinations are
# called OBJC by the labeller for lack of evidence. Here we change
# edges if we find a subordinated conjunction.
sub objcConversion {
  my $edge = shift;

  my $has_right_child = 0;

  for my $c (@{$$edge[0]->{children}}) {
    my $e = $c->{dep};
    my $l = $$e[1];
    my $w = $$e[0]->{word};
    my $cat = $$e[0]->{features}{cat};
    if($$e[0]{id} > $$edge[0]{id}) {
      $has_right_child = 1;
    }

    # `da' / `ob' indicate object clause
    return 0 if $w =~ /^(da(ss|)|ob)$/i;

    # `weil' etc. indicate subclause
    if($cat =~ /KOU[SI]/ && $l eq 'KONJ') {
      debug "Change OBJC to NEB\n" ;
      $$edge[1] = 'NEB';
      return nebLabelConversion $edge;
    }
  }

  # if at least one word follows, we probably have verb-second
  # position, so we want S.
  if($has_right_child) {
    $$edge[1] = 'S';
    sConversion $edge;
    return 1;
  }
  return 0;
}



#------------------------------------------------------------------------------
#
sub advConversion($$);
sub advConversion($$) {
  my $edge = shift;
  my $deps = shift;
  return 0 if $$edge[1] ne "ADV";
  my $result = 0;

  # Adverbs usually do not modify other adverbs; instead, they should
  # be siblings.

  if($$edge[0]{features}{cat} =~ /^(ADV|PTKNEG)$/ &&
     $$edge[2]{id} != -1 &&
     $$edge[2]{features}{cat} =~ /^(ADV|PTKNEG)$/ &&
     $$edge[0]{word} !~ /^($scope_adv_re)$/o) {
    debug "Raise adverb $$edge[0]{word} below adverb $$edge[2]{word} to $$edge[2]{dep}[2]{word}";
    moveEdge $edge, $$edge[2]{dep}[2];
    advConversion $edge, $deps;
    return 1;
  }

  # participles, when used adverbially, are to be tagged ADJD.
  # Unfortunately NEGRA does not always follow this rule, so we
  # enforce it here.
  if($$edge[0]->{features}{cat} eq 'VVPP') {
    $$edge[0]->{features}{cat} = 'ADJD';
    debug "Change adverbial VVPP to ADJD\n" ;
  }

  #s1153 "von(APPR) heute(MO)", the correct label is to be PN instad of ADV
  my $e0 = $$edge[2]{dep};
  my $akk = 0;
  if(defined $$e0[1] && $$e0[1] eq 'PP') {
    foreach my $child(@{$$edge[2]->{children}}){
      $akk++;
    }
    if($akk == 1){
      $$edge[1] = 'PN';
      debug "Change adverbial PP kernel from ADV to PN" ;
      $result = 1;
    }
  }

  # `so da' should be treated as a group because it changes the
  # properties of `da': with `so' (or `ohne'/`auer') it introduces
  # subclauses, alone it introduces object clauses.
  if($$edge[0]{word} eq 'so') {
    for my $d (@$deps) {
      if($$d[0]{id} == $$edge[0]{id}+1) {
        if($$d[0]{word} =~ /da(ss|)/) {
          debug 'group "so da"';
          $result = 1;
          moveEdge $edge, $$d[0];
          my $p = $$d[2]{dep};
          if($$p[1] eq 'OBJC') {
            $$p[1] = 'NEB';
            nebLabelConversion $p;
          }
        }
        last;
      }
    }
  }


  $result = 1 if lower_into_coordination $edge;

  # Scope adverbs are sometimes subordinated into the subclause even
  # when they are nested deep within the main clause (s1830). In this
  # case the normal tests are not enough; we must search the entire
  # tree for edges which indicate this.
  my $x = '';
  if($$edge[0]{id} < $$edge[2]{id}) {
    for my $d (@$deps) {
      if($$d[2]{id} != -1 &&
         $$d[1] ne 'SUBJ' &&
         $$d[0]{id} > $$edge[0]{id} &&
         $$d[0]{id} < $$edge[2]{id} &&
         $$d[2]{id} < $$edge[0]{id}){
        debug "Raise extraposed scope adverb `$$edge[0]{word}'\n";
        $x = $d;
        last;
      }
    }
  }
  if($x) {
    while($$edge[2]{id} > $$x[0]{id}) {
      moveEdge $edge, $$edge[2]{dep}[2];
    }
    return 1;
  }

  return $result;
}

#------------------------------------------------------------------------------
#
sub explLabelConversion {
  my $edge = shift;

  return 0 if $$edge[1] ne "EXPL";
  my $result = 0;

  # If the verb "sein" has no PRED or AUX, but EXPL, EXPL should be
  # marked as PRED

  # "Es sind wir", "es waren gute Leute"
  if($$edge[2]->{features}{cat} eq 'VAFIN'){
    my $auxFound = 0;
    foreach my $child (@{$$edge[2]->{children}}) {
      my $childEdge = $child->{dep};
      my $childLabel = $$childEdge[1];
      if (($childLabel eq 'AUX') || ($childLabel eq 'PRED')) {
        $auxFound = 1;
        last;
      }
    }

    if (!$auxFound) {
      debug "Reverting $$edge[1] to PRED at $$edge[0]->{id}";
      $$edge[1]='PRED';
      $result = 1;
    }
  }

  # EXPL can only modify finite verbs, since its only reason for
  # existence is to fill the Vorfeld, and only finite verbs have that.
  # Therefore it MUST be raised across AUX edges, even if it is
  # already projective.
  my $p = $$edge[0]{parent}{dep};
  while($$p[1] =~ /AUX/) {
    $p = $$p[0]{parent}{dep};
  }
  if($edge != $p) {
    moveEdge $edge, $$p[0];
    debug "Raise EXPL to finite verb\n" ;
    $result = 1;
  }

  # enforce consistent case for all EXPL
  $$edge[0]->{features}{case} = "nom";
  return $result;
}



#------------------------------------------------------------------------------
# force verbs with nouns as subjects to be person:third
sub subjConversion {
  my $edge = shift;

  if($$edge[1] eq 'SUBJ' &&
     $$edge[2]->{features}{cat} =~ /FIN$/ &&
     $$edge[0]->{word} !~ /^(ich|du|wir|ihr)$/i
    ) {
    $$edge[2]->{features}{person} = 'third';
    $$edge[0]->{features}{case} = 'nom';
  }

  # Detect close conjunctions of verbs and force all modifiers
  # to the higher verb.
  my $CJ = $$edge[2]{dep};
  return 0 unless defined $CJ;
  return 0 unless $$CJ[1] eq 'CJ';
  my $KON = $$CJ[2]{dep};
  return 0 unless defined $KON;
  return 0 unless $$KON[1] eq 'KON';
  return 0 unless $$KON[2]{id} + 2 == $$edge[2]{id};

  debug "Close coordination detected, raising SUBJ";

  for my $c (@{$$edge[2]{children}}) {
    moveEdge $$c{dep}, $$KON[2];
  }
  return 1;

}

#------------------------------------------------------------------------------
# change OBJA to OBJD if case=dat
sub objaConversion {
  my $edge = shift;

  foreach my $child(@{$$edge[0]->{children}}){
      my $childEdge = $child->{dep};
      my $childLabel = $$childEdge[1];

      if(($childLabel eq 'DET') && ($$childEdge[0]->{features}{case} eq "Dat") &&
               ($$childEdge[2]->{features}{case} eq 'dat')){

         $$edge[1] = 'OBJD';
         debug "Change OBJA to OBJD\n" ;
         return 1;
      }
  }
}



# ---------------------------------------------------------------------
# Change some ATTR edges to other labels.
sub attrConversion {
  my $edge = shift;

  # The label of words with the category CARD that stand right
  # of the modifiee should be APP instead of ATTR.
  if($$edge[0]->{features}{cat} eq 'CARD' &&
     $$edge[0]->{id} > $$edge[2]->{id}) {
    $$edge[1] = 'APP';
    appoConversion $edge;
    return 1;
  }

  # The words `alle', `beide', `viele' can be used as nonprojective
  # attributes. We want them to be adverbial modifiers instead, so we
  # raise them to the next verb and relabel them.
  my $adj = $$edge[0];

  if($$adj{word} =~ /^(all|viel|beid)/ &&
     $$edge[0]{id} > $$edge[2]{id} &&
    $$edge[2]{dep}[2]{id} < $$adj{id} &&
     $$edge[2]{dep}[2]{id} > $$edge[2]{id}) {
    debug "Correct nonprojective `alle'\n";
    moveEdge $edge, $$edge[2]{dep}[2];
    $$edge[1] = 'ADV';
    return 1;
  }

}

# ----------------------------------------------------------------------
# make ATTR labels.
#
# First rule: only the first occurance of a `DET' label modifying a
# Noun will be preserved, the other labels will be changed to `ATTR'.
#
# Second rule: change to `ATTR' if there is an `APPRART' modifiee 1
# level above
#
# Third rule: change edge label to 'ATTR' if the modifier is on the
# right side of the modifiee
sub detLabelConversion {
  my $edge = shift;
  my $result = 0;
  my $sisters = $$edge[2]->{children};

  # change `DET' to `APP' for postnominal numbers
  if($$edge[0]{id} > $$edge[2]{id} &&
     $$edge[0]{features}{cat} eq 'CARD') {
    $$edge[1]='APP';
    debug "Change DET to APP\n" ;
    attrConversion $edge;
    return 1;
  }

  # first rule
  foreach my $sister (@$sisters) {
    next if ($sister == $$edge[0]);
    my $dep = $sister->{dep};
    if ($$dep[1] eq 'DET' &&
        $$edge[0]->{id} < $$dep[0]->{id}) {
      $$dep[1]='ATTR';
      debug "Change DET to ATTR\n" ;
      attrConversion $dep;
      $result = 1;
    }

  }

  # second rule
  my $p = $$edge[2]{dep};
  if (exists $$p[2]{features} && $$p[2]{features}{cat} =~ /APPRART/) {
    $$edge[1]='ATTR';
    debug "Change DET to ATTR\n" ;
    attrConversion $edge;
    $result = 1;
  }

  # third rule
  if ($$edge[2]->{id} < $$edge[0]->{id}) {
    if ($$edge[2]->{features}{cat} =~ /KOKOM/) { # s1028 "weniger als 1992"
      debug "Change DET to CJ\n" ;
      $$edge[1] = 'CJ';
    } else {
      debug "Change DET to ATTR\n" ;
      $$edge[1] = 'ATTR';
      attrConversion $edge;
    }
    $result = 1;
  }


  #"ein paar" s1068 improving a wrong dependency
  if (($$edge[2]->{word} eq 'ein') && ($$edge[0]->{word} eq 'paar')) {
    debug "Correct PART `$$edge[0]{word}' `$$edge[2]{word}'";
    my $parent = $$edge[2]->{dep};

    moveEdge $edge, $$parent[2];
    moveEdge $parent, $$edge[0];
    $$parent[1] = 'PART';
    $$edge[1] = 'DET';
    $result = 1;
  }

  #"solch einem Spiegel" s230 the modifier should modify the modifee's modifee
  #"einen solch sarkastischen Zusammenhang"
  if (($$edge[0]->{features}{cat} eq 'PIDAT') && ($$edge[2]->{features}{cat} =~ /ART|ADJA/)) {
    my $parent = $$edge[2]-> {dep};
    moveEdge $edge, $$parent[2];
    $$edge[1] = 'ATTR';
    debug "Change DET to ATTR\n" ;
    $result = 1;
  }

  return $result;
}

# ----------------------------------------------------------------------
sub konLabelConversion($);
sub konLabelConversion($) {
  my $edge = shift;
  my $result = 0;

  # When a conjunction starts a sentence, it should be labelled `S'
  # like all NIL edges, not `KON'.
  if ($$edge[2]->{word} eq 'NIL') { #if there is no modifiee
    $$edge[1]='S'; #rename edge to 'S'
    debug "Change KON to S\n" ;
    return 1;
  }

  # Where more than one adjective is coordinated without a
  # conjunction, we prefer to subordinate both under the same noun.
  # NEGRA stores them in one CAP (s733) so they become subordinated
  # under each other. We change that here.
  if($$edge[0]{features}{cat} =~ /^ADJ/ &&
     $$edge[2]{features}{cat} eq $$edge[0]{features}{cat} &&
     $$edge[2]{dep}[1] eq 'ATTR') {
    moveEdge $edge, $$edge[2]{parent};
    $$edge[1] = 'ATTR';
    $result = 1;
    konLabelConversion $edge;
  }

  $result = 1 if KON_lowering $edge;
  $result = 1 if appoConversion $edge;
  return $result;
}

# ----------------------------------------------------------------------
# Adverbs that precede a conjunction should modify the conjunction,
# not the verb.
sub konjConversion {
  my $edge = shift;
  my $result = 0;

  my $verb = $$edge[2];
  my $k = $$edge[0];
  for my $c(@{$$verb{children}}) {
    if($$c{id} < $$k{id}) {
      moveEdge($$c{dep}, $k);
      $result = 1;
      debug "Lower adverb `$$c{word}' to `$$k{word}'";
    }
  }
  return $result;
}

# ---------------------------------------------------------------------
# GMOD modifiers of conjoined phrases can modify either side equally
# well. For consistency the grammar forces them to modify the right
# hand side, so we move any additional modifiers to the right here.
sub gmodConversion {
  my $edge = shift;
  my $gmod = $$edge[0];
  my $lhs = $$edge[2];
  my $result = 0;
  for my $c (@{$lhs->{children}}) {
    my $e = $$c{dep};
    if($$e[1] eq 'KON') {
      for my $rhs (@{$c->{children}}) {
        my $e2 = $$rhs{dep};
        if($$e2[1] =~ /^(CJ|KON)$/) {

          # that only applies if the GMOD is to the right of both words, though.
          if($$rhs{id} < $$gmod{id}) {
            moveEdge $edge, $$e2[0];
            $result = 1;
            debug "Shift GMOD `$$gmod{word}'\n" ;
          }
        }
      }
    }
  }
  return $result;
}




#------------------------------------------------------------------------------
# insert auxiliary edges for relative clauses.
sub relLabelConversion {
  my $edge = shift;
  my $deps = shift;
  my $extradeps = shift;

  # The REF edge connects the relative pronoun with its antecedent, so
  # we have to step down the REL edge and then further down until we
  # find the relative pronoun. The relative pronoun is then connected
  # with the antecedent via REF.

  my $verb = $$edge[0]{id};
  my $antecedent = $$edge[2]{id};
  my $pronoun;

  my $w = find_under $$edge[0], 'PRELS';
  return if !$w;
  return unless $$w{word} =~ /^d/;
  $pronoun = $$w{id};

  $$extradeps{REF}[$pronoun]{up} = $antecedent+1;
  return 0;
}

#------------------------------------------------------------------------------
# Translate NEGRA's RE label into normal subclauses.
sub reLabelConversion {
  my $edge = shift;
  my $deps = shift;
  my $parent = $$edge[2]{dep};

  my $counter = 0;
  while($parent && $$parent[1] !~ /^(S|CJ|KON|REL|NEB|OBJC)$/) {
    $parent = $$parent[2]{dep};
    last if $counter++ > $#$deps;
  }

  if($parent) {
    moveEdge $edge, $$parent[0];
    $$edge[1] = 'NEB';
  } else {
    $$edge[1] = 'ROOT';
    rootConversion $edge;
  }

  debug "Change RE to $$edge[0]{word}--$$edge[1]-->$$edge[2]{word}";

  return 1;
}


#------------------------------------------------------------------------------
#
sub predLabelConversion {
  my $edge = shift;
  return 0 if $$edge[1] ne "PRED";
  my $result = 0;

  # the direct object of the verb "nennen" should be marked as OBJA2
  # instead of PRED

  # Unfortunately NEGRA does not always follow this
  # rule, so we enforce it here.

  if($$edge[2]->{word} =~ /^(nenn(en|e|st|t|te|test|ten|tet)|nannt(e|est|en|et)|genannt|
  				hei(ss|)(en|e|t|te|test|ten|tet)|gehie(ss|)en)$/){
     $$edge[1] = 'OBJA2';
     $result = 1;
   }

  return $result;

}

#------------------------------------------------------------------------------
#  s195 "..ein wenig lftet"
#  the label is renamed to 'ADV'
sub zeitLabelConversion {
  my $edge = shift;

  return 0 if $$edge[1] ne "ZEIT";
  my $result = 0;
  if ($$edge[0]->{word} eq 'wenig') {
    foreach my $child (@{$$edge[0]->{children}}) {
      my $childEdge = $child->{dep};
      if ($$childEdge[0]->{word} eq 'ein') {
        $$childEdge[1] = 'PART';
        $$edge[1]='ADV'; #rename edge to 'ADV'
        debug "Change ZEIT to ADV\n" ;
        $result = 1;
        last;
      }
    }
  }

  # s452,s472 "alles in allem"
  # this rule doesn't work
  if ($$edge[0]->{word} eq 'alles') {
    foreach my $child (@{$$edge[0]->{children}}) {
      debug "ZEIT `$$edge[0]{word}' `$$edge[2]{word}'";
      my $childEdge = $child->{dep};
      if ($$childEdge[0]->{word} eq 'in') {
	debug "ZEIT `$$edge[0]{word}' `$$edge[2]{word}'";
        foreach my $cd (@{$$edge[0]->{children}}) {
          my $cEdge = $cd->{dep};
          if ($$cEdge[0]->{word} eq 'allem') {
            debug "ZEIT to ADV `$$edge[0]{word}' `$$edge[2]{word}'";
            $$edge[1]='ADV'; #rename edge to 'ADV'
            debug "Change ZEIT to ADV\n" ;
            $result = 1;
            last;
          }
        }
      }
    }
  }
  return $result;
}

# -------------------------------------------------------------------------
# Correct phrases with separable conjunctions. The left parts of
# conjunctions are prone to annotator error (compare NEGRA s206).
# Such input can lead to wrong, and indeed nonprojective, attachment
# of words such as `weder'. Strictly speaking this an error in the
# input, but it is easier to silently correct such structures here.
sub cjLabelConversion {
  my $edge = shift;
  return 0 if $$edge[1] ne "CJ";

  my $result = 0;
  #debug "cjLabelConversion($$edge[0]{word})";

  if($$edge[0]{word} =~ /^(entweder|sowohl|weder)$/) {
    my $e1 = $$edge[2]->{dep};
    my $left = $$edge[0];
    my $right = $$edge[2];
    my $parent = $$e1[2];

    # raise `entweder' to be a sibling of `oder'
    if ($$right{features}{cat} eq 'KON') {
      moveEdge $edge, $parent;
      $$edge[1] = 'KON';
      $result = 1;
      debug "Repair two-part conjunction $$left{word}--$$right{word}\n";
    }
   }


  # `und zwar' is idiosyncratic and functions as an adverbial. We
  # force `und' to be the head, the lower phrase the argument, and
  # `zwar' a modifier of the argument.
  if($$edge[0]{word} eq 'zwar' &&
     $$edge[2]{word} eq 'und' &&
     $$edge[0]{id} == $$edge[2]{id}+1) {
    debug "Change `und zwar'";
    my $CJ = $edge;
    my $KON = $$CJ[2]{dep};
    my $ADV = $$KON[2]{dep};
    moveEdge $CJ, $$KON[2];
    $$CJ[1] = 'ADV';
    moveEdge $KON, $$ADV[2];
    moveEdge $ADV, $$KON[0];
    $$ADV[1] = 'CJ';
    $result = 1;
  }

  return $result;

}

# ----------------------------------------------------------------------
# Repair tokenizer errors.
#
# NEGRA blindly follows the decisions that their tokenizer makes even
# when they are obviously wrong. This routine tries to correct such
# mistakes.
#
# Not all re-tokenization is done here, only the part which is
# immediately detectable when looking at the word sequence. During
# phrase conversion, other rules that condition on phrase node labels
# (such as `MTA') may also re-tokenize.
sub re_tokenize {
  my $word = shift;

  my $next = $$rootNode{leaves}[$$word{id}+1];
  my $prev = $$rootNode{leaves}[$$word{id}-1];


  # The NEGRA tokenizer treats every instance of ' as an isolated
  # token. But ' is used in very different roles; a sentence like
  #
  # ...nach einem 'Leseroman', der die Herrschaft des literarischen...
  #
  # uses them as quotation, and in that function they should be
  # tokens. But in many other uses such as d'Acrais, or '84, or sei's
  # drum, the apostrophe belongs to one or both of the adjoining
  # tokens.
  if ($$word{word} eq "'") {

    # Flores d'Arcais s1384 (MNP)
    if ($$prev{word} eq "d" && $$next{nodeTag} eq 'NE') {
      debug "repair apostrophe in given name";
      moveNode $word, $$prev{parent};
      spliceWords $$prev{parent}, $$prev{id}, $$next{id},
        {
         word => "$$prev{word}'$$next{word}",
         edgeTag => 'PNC',
         nodeTag => 'NE',
         morphTag => $$next{morphTag}};
    }

    # '98, war 's
    elsif ($$next{nodeTag} eq 'CARD' || $$next{word} eq 's') {
      debug "repair apostrophe as elision";
      moveNode $word, $$next{parent};
      spliceWords $$next{parent}, $$word{id}, $$next{id},
        {
         word => "'$$next{word}",
         edgeTag => $$next{edgeTag},
         nodeTag => $$next{nodeTag},
         morphTag => $$next{morphTag}};
    }

    # Strau' "Fledermaus" s1055 (GMOD), ich tu'
    elsif (($$prev{nodeTag} eq 'NE' && $$prev{word} =~ /(s|sch|ch|x||z)$/) ||
           ($$prev{nodeTag} =~ /V.FIN/)) {
      debug "repair apostrophe as genitive morpheme";
      moveNode $word, $$prev{parent};
      spliceWords $$prev{parent}, $$prev{id}, $$word{id},
        {
         word => "$$prev{word}'",
         edgeTag => $$prev{edgeTag},
         nodeTag => $$prev{nodeTag},
         morphTag => $$prev{morphTag}};
    }
  }

  # Compounds with hyphen are sometimes wrongly split. This probably
  # happens when they happened to be hyphenated at line endings, so
  # that Baden-Wrttemberg becomes Baden- Wrttemberg. This is clearly
  # wrong, so we reassemble truncations if they immediately precede
  # nouns. Of course, if they precede a comma or conjunction they are
  # true truncations.
  elsif ($$word{word} =~ /^[A-Z].*-$/) {
    if ($$next{nodeTag} =~ /^(NN|NE|FM)$/) {
      debug "reassemble $$word{word} $$next{word}\n";
      moveNode $next, $$word{parent};
      spliceWords $$word{parent}, $$word{id}, $$next{id},
        {
         word => "$$word{word}$$next{word}",
         edgeTag => $$word{edgeTag},
         nodeTag => $$next{nodeTag},
         morphTag => $$next{morphTag}}
      }
  }

  # Compounds with quotation marks in them are particularly nasty. If
  # an author uses the word "Jingle"-Einspielungen, the NEGRA
  # tokenizer slashes this into four tokens. But the error can be
  # accurately diagnosed: a TRUNC starting with - after an " is always
  # wrong.
  elsif ($$word{word} =~ /^-/ &&
      $$word{nodeTag} =~ /^(TRUNC|NN)$/ &&
      $$prev{word} eq '"') {

    # find the previous "
    my $i = $$prev{id}-1;
    $i-- while($i > -1 && $$rootNode{leaves}[$i]{word} ne '"');
    if ($i > -1) {
      my $first = $$rootNode{leaves}[$i];
      my $s;
      for ($i..$$word{id}) {
        moveNode $$rootNode{leaves}[$_], $$word{parent};
        $s .= $$rootNode{leaves}[$_]{word};
      }
      spliceWords $$word{parent}, $i, $$word{id},
        {
         word => $s,
         edgeTag => $$word{edgeTag},
         nodeTag => $$word{nodeTag},
         morphTag => $$word{morphTag}};

      debug "reassemble quoted compound $s\n";
    }
  }

  # Adjectives back-formed from city names with spaces in them are
  # usually regarded as two tokens: Bad Nauheimer. We want them to be
  # single tokens, so we reassemble them.
  elsif ($$word{word} =~ /^[A-Z].+er$/ &&
      $$word{nodeTag} eq 'ADJA' &&
      $$prev{nodeTag} eq 'NE' &&
      $$prev{word} =~ /^(New|Neu|Bad)$/) {

    spliceWords $$word{parent}, $$word{id}-1, $$word{id},
        {
         word => "$$prev{word} $$word{word}",
         edgeTag => $$word{edgeTag},
         nodeTag => $$word{nodeTag},
         morphTag => $$word{morphTag}};

    # Make sure the new word is considered adjectival so it will not
    # be chosen as the NP head later.
    $$word{parent}{edgeTag} = 'AP';
    debug "reassemble city name `$$prev{word} $$word{word}'\n";
  }

  # Proper names with punctuation in them should be treated as a unit,
  # therefore we reassemble "Wham" "!" into "Wham!".
  elsif($$word{nodeTag} =~ /\$/ &&
    $$word{parent}{nodeTag} eq 'MPN' &&
     $$prev{parent} == $$word{parent}) {
    debug "reassemble punctuation in proper name: `$$prev{word} $$word{word}'";

    spliceWords $$word{parent}, $$prev{id}, $$word{id},
      {
       word => "$$prev{word} $$word{word}",
       edgeTag => $$prev{edgeTag},
       nodeTag => $$prev{nodeTag},
       morphTag => $$prev{morphTag}
      }
    }

  # Verbs ending in 's are certainly fusions opf verb + `es'.
  elsif($$word{nodeTag} =~ /FIN$/ &&
        $$word{word} =~ /\'s/) {
    debug "split verb `$$word{word}'";

    my ($verb,$pronoun) = splitWord $word, 2;
    $$verb{edgeTag} = $$word{edgeTag};
    $$verb{nodeTag} = $$word{nodeTag};
    $$verb{morphTag} = $$word{morphTag};
    $$verb{features} = $$word{features};
    $$verb{word} =~ s/\'s$//g;

    $$pronoun{edgeTag} = 'SB';
    $$pronoun{nodeTag} = 'PPER';
    $$pronoun{morphTag} = '3.Sg.Neut.Nom';
    $$pronoun{word} = "'s";
  }
}

#----------------------------------------------------------------------
# Apply correction rules that regard tokenization. See re_tokenize{}
# for details.
sub fixNegraWords {
  my $root = shift;

  for my $l(@{$$root{leaves}}) {
    re_tokenize $l;
  }

  return 1;
}

%callbacks = ( preProcess => \&fixNegraWords );
