#!/usr/bin/perl -w

#analyzes a log file of 'compare.sh' and prints some statistics about errors

$filename = $ARGV[0];
%wrongLabels = ();
%shouldBe = ();
%struct = ();
my $nrComparisons = 0;
my $nrWords = 0;

open(LOGFILE, "$filename");
while($line = <LOGFILE>){ 

  # count comparison
  if($line =~ /^INFO: Compare parses/) {
    $nrComparisons++;
  }

  # count words
  if($line =~ /^s[0-9]+ :/) {
    $nrWords++ while $line =~ /(, )|(;$)/g;
  }

  # find structure error
  if($line =~
     m|Governor of \`(.+)/(.+)\' on level \`SYN\' is \`.+/(.+)\', should be \`.+/(.+)\'|) {
    $structural++;
    $attachedWrong = "$1/$2";
    $struct{"$2/$3/$4"}++;
    next;
  }

  # find lexical error
  elsif($line =~ /^ERROR: Word with different lexical reading/) {
    $lexical++;
  }

  # find label error
  elsif(my($word,$wrong,$correct) =
        ($line =~
         /ERROR: Label of edge above node \`(.+)\'.+ is (\`[A-Z]+\'), should be (\`[A-Z0-9]+\')/)) {
    if($attachedWrong eq $word) {
      # structural error, do not count it as a label error
    } else {
      # proper label error
      $label++;
      $wrongLabels{$wrong} += 1;
      $shouldBe{$correct} += 1;
    }
  }

  # now forget about the structure error in the previous line
  $attachedWrong = '';
}
close(LOGFILE);


for(sort  keys %wrongLabels) {
  print "Label $_ was $wrongLabels{$_} times set wrong.\n";
}

print "---------------------------------------\n";

for(sort  keys %shouldBe) {
  print "Label $_ was $shouldBe{$_} times not set. \n";
}


# print some common structural errors
print "---------------------------------------\n";
for(sort {$struct{$a} <=> $struct{$b} } keys %struct) {
  print "Attachment error $_ occurred $struct{$_} times.\n"
    if $struct{$_}>10;
}


print "---------------------------------------\n";
print "$nrComparisons comparisons, $nrWords words\n";
print "$structural structure errors, $label label errors, $lexical lexical errors\n";
