/* Author: Ingo Schroeder
 * Purpose: 
 * Birth: 11/8/99
 *
 * $Id: incrementalcompletion.c,v 1.111 2004/02/19 15:27:30 engelman Exp $
 */

/* ---------------------------------------------------------------------- */
#include <configure.h>

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <blah.h>
#include "incrementalcompletion.h"
#include "constraintnet.h"
#include "netsearch.h"
#include "write.h"
#include "eval.h"
#include "statistics.h"
#include "command.h"
#include "frobbing.h"
#include "timer.h"
#include "hook.h"
#include "parse.h"
#include "set.h"

/* -- MACRO DEFINITIONS ------------------------------------------------- */
#define IC_HEUR_LASTNODE         1
#define IC_HEUR_LASTPATH         2
#define IC_HEUR_LASTNODEALL      3
#define IC_HEUR_LASTPATHALL      4
#define IC_HEUR_ALLNODE          5

#define IC_HEUR_BEGIN            1
#define IC_HEUR_END              5

/* -- TYPE DECLARATIONS ------------------------------------------------- */
typedef struct {
  int from;
  int to;
  LexiconItem le;
  String word;
} BindingNodeStruct, *BindingNode;

typedef struct {
  Boolean main_level;
  Level level;
  String label;
  BindingNode modifier;
  BindingNode modifiee;
} BindingStruct, *Binding;

typedef struct {
  List prev_bindings;
  List prev_struct;
  List prev_hotpath; /* path from hot-spot to root */
  List prev_hotspots; /* list of hot-spot nodes */
} IcIterStruct, *IcIter;
  
IcParams icParams;
String icParamsStr = NULL;
Timer icTimer = NULL;

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * build list of parses ending at the given time position
 */
List icBuildParsesList(List parses, int position)
{
  List l = NULL;
  List m = NULL;

  for (m = parses; m; m = listNext(m)) {
    Parse p = (Parse) listElement(m);
    if (p->lg->max == position) {
      l = listAppendElement(l, p);
    }
  }
  return l;
}

/* ----------------------------------------------------------------------
 * delete last added parses from the constraint net parses
 */
void icDeleteLastParses(ConstraintNet net)
{
  List l = net->parses;
  List lprev = NULL;

  while (l) {
    Parse p = (Parse) listElement(l);
    if (p->lg->max == net->lexemgraph->max) {
      if (lprev == NULL) {
	net->parses = listNext(l);
      } else {
	listSetNext(lprev, listNext(l));
      }
      l = listNext(l);
    } else {
      lprev = l;
      l = listNext(l);
    }
  }
}

/* ----------------------------------------------------------------------
 * clean list of parses of parses added before frobbing
 */
List icDeleteNonFrobbingParses(List parses)
{
  List l, lprev = NULL;
  Boolean cond = FALSE;

  for (l=parses; l; l=listNext(l)) {
    Parse p = (Parse) listElement(l);
    if (strcmp(p->searchStrategy, "frobbing") == 0) {
      cond = TRUE;
    }
  }
  
  if (!cond) {
    return parses;
  }

  lprev = NULL;
  l = parses;
  while (l) {
    Parse p = (Parse) listElement(l);
    if (strcmp(p->searchStrategy, "frobbing") != 0) {
      if (lprev == NULL) {
	parses = listNext(l);
      } else {
	listSetNext(lprev, listNext(l));
      }
      l = listNext(l);
    } else {
      lprev = l;
      l = listNext(l);
    }
  }
  return parses;
}

/* ----------------------------------------------------------------------
 * save a binding node
 */
void icSaveBindingNode(LexemNode ln, BindingNode b)
{
  b->from = ln->arc->from;
  b->to = ln->arc->to;
  b->word = ln->lexem->word;
  b->le = ln->lexem;
}

/* ----------------------------------------------------------------------
 * save the solution(s)
 */
void icSaveBinding(LevelValue lv, Binding b, Level main_level)
{
  b->main_level = (lv->level == main_level);
  b->level = lv->level;
  b->label = lv->label;
  icSaveBindingNode(lv->modifier, b->modifier);
  if (lv->modifiee == NULL) {
    b->modifiee->from = b->modifiee->to = -1;
    b->modifiee->le = NULL;
    b->modifiee->word = NULL;
  } else if (lv->modifiee == NONSPEC) {
    b->modifiee->from = b->modifiee->to = -2;
    b->modifiee->le = NONSPEC;
    b->modifiee->word = NONSPEC;
  } else {
    icSaveBindingNode(lv->modifiee, b->modifiee);
  }
}

/* ----------------------------------------------------------------------
 * icPrintStruct
 */
void icPrintStruct(List prev_struct)
{
  List m, n, p;
  for (m=prev_struct; m; m=listNext(m)) {
    for (n=(List) listElement(m); n; n=listNext(n)) {
      for (p=(List) listElement(n); p; p=listNext(p)) {
	BindingNode b = (BindingNode) listElement(p);
	cdgPrintf(CDG_INFO, "(%d %d %s)", b->from, b->to, b->le->description);
      }
      cdgPrintf(CDG_INFO, "\n");
    }
    cdgPrintf(CDG_INFO, "----\n");
  }
}

/* ----------------------------------------------------------------------
 * icFreeStruct
 */
void icFreeStruct(List prev_struct)
{
  List m, n;
  for (m=prev_struct; m; m=listNext(m)) {
    for (n=(List) listElement(m); n; n=listNext(n)) {
      listDelete(listElement(n));
    }
    listDelete(listElement(m));
  }
  listDelete(prev_struct);
}

/* ----------------------------------------------------------------------
 * icFreeBindings
 */
void icFreeBindings(List prev_bindings)
{
  List l, m;
  for (l=prev_bindings; l; l=listNext(l)) {
    for (m=(List) listElement(l); m; m=listNext(m)) {
      Binding b = (Binding) listElement(m);
      memFree(b->modifier);
      memFree(b->modifiee);
      memFree(b);
    }
    listDelete(listElement(l));
  }
  listDelete(prev_bindings);
}

/* ----------------------------------------------------------------------
 * icFreeHotPath
 */
void icFreeHotPath(List prev_hotpath)
{
  List m;
  for (m=prev_hotpath; m; m=listNext(m)) {
    listDelete(listElement(m));
  }
  listDelete(prev_hotpath);
}

/* ----------------------------------------------------------------------
 * icFreeHotSpots
 */
void icFreeHotSpots(List prev_hotspots)
{
  List l, m;
  for (l=prev_hotspots; l; l=listNext(l)) {
    for (m=(List) listElement(l); m; m=listNext(m)) {
      BindingNode b = (BindingNode) listElement(m);
      memFree(b);
    }
    listDelete(listElement(l));
  }
  listDelete(prev_hotspots);
}

/* ----------------------------------------------------------------------
 * icPrintBindings
 */
void icPrintBindings(List prev_bindings)
{
  List l, m;
  for (l=prev_bindings; l; l=listNext(l)) {
    for (m=listElement(l); m; m=listNext(m)) {
      Binding b = (Binding) listElement(m);
      if (b->modifiee->le == NULL) {
	cdgPrintf(CDG_INFO, "%s (%s): %d %d %s -- -1 -1 NIL\n", 
		  b->level->id, b->label, 
		  b->modifier->from, b->modifier->to, b->modifier->le->description);
      } else if (b->modifiee->le == NONSPEC) {
	cdgPrintf(CDG_INFO, "%s (%s): %d %d %s -- -2 -2 NONSPEC\n", 
		  b->level->id, b->label, 
		  b->modifier->from, b->modifier->to, b->modifier->le->description);
      } else {
	cdgPrintf(CDG_INFO, "%s (%s): %d %d %s -- %d %d %s\n", 
		  b->level->id, b->label, 
		  b->modifier->from, b->modifier->to, b->modifier->le->description,
		  b->modifiee->from, b->modifiee->to, b->modifiee->le->description);
      }
    }
  }
}

/* ----------------------------------------------------------------------
 * compare two integers - useful for sorting
 */
Boolean icCompareInts(int n1, int n2)
{
  return (n1 < n2);
}

/* ----------------------------------------------------------------------
 * build the sorted vector of timespan limits of wordgraph nodes
 */
Vector icBuildTimeLimits(List list)
{
  Vector t = vectorNew(0);
  for ( ; list; list = listNext(list)) {
    int number = ((Arc) listElement(list))->from;
    if (!vectorContains(t, (void *)number))
      vectorAddElement (t, (void *)number);

    number = ((Arc) listElement(list))->to;
    if (!vectorContains(t, (void *)number))
      vectorAddElement (t, (void *)number);
  }
  vectorSort(t, icCompareInts);
  return t;
}


/* ----------------------------------------------------------------------
 * get the number of timespan limits; takes a list of arcs
 */
int icGetTimeLimitsNumber(List list)
{
  Vector v = icBuildTimeLimits(list);
  int i = vectorSize(v);
  vectorDelete(v);
  return i;
}

/* ----------------------------------------------------------------------
 * returns the node on the top of the tree containing the node parameter
 */
int icGetTop(List prev_solution, int node)
{
  List n;
  int res;
  Binding oldb = NULL;
  Boolean cont;

  /* cdgPrintf(CDG_INFO, "test - %d\n", node); */
  do {
    cont = TRUE;
    for (n = prev_solution; n && cont; n=listNext(n)) {
      Binding b = (Binding) listElement(n);
      if (b->main_level) {
	if ((oldb == NULL && 
	     b->modifier->to == node) ||
	    (oldb == NONSPEC &&
	     b->modifier->to == node) ||
	    (oldb != NULL &&
	     oldb != NONSPEC &&
	     b->modifier->to == oldb->modifiee->to &&
	     b->modifier->from == oldb->modifiee->from &&
	     b->modifier->le == oldb->modifiee->le)) {
	  cont = FALSE;
	  oldb = b;
	}
      }
    }
  }
  while (!cont && 
         oldb->modifiee->to != -2 && oldb->modifiee->from != -2 &&
         oldb->modifiee->to != -1 && oldb->modifiee->from != -1);
  if (oldb == NULL) {
    res = -1;
  } else if (oldb == NONSPEC) {
    res = -2;
  } else {
    res = oldb->modifier->to;
  }
  /* cdgPrintf(CDG_INFO, "out - %d\n", res); */
  return res;
}

/* ----------------------------------------------------------------------
 * Build a path for one tree of one solution (from the node to the top)
 */
List icBuildPath(List prev_solution, int node)
{
  List n, result;
  Binding oldb = NULL;
  Boolean cont;

  result = NULL;
  do {
    cont = TRUE;
    for (n=prev_solution; n && cont; n=listNext(n)) {
      Binding b = (Binding) listElement(n);
      if (b->main_level) {
	if ((oldb == NULL && 
	     b->modifier->to == node) ||
	    (oldb == NONSPEC &&
	     b->modifier->to == node) ||
	    (oldb != NULL &&
	     oldb != NONSPEC &&
	     b->modifier->to == oldb->modifiee->to &&
	     b->modifier->from == oldb->modifiee->from &&
	     b->modifier->le == oldb->modifiee->le)) {
	  cont = FALSE;
	  oldb = b;
	  result = listAppendElement(result, b->modifier);
	}
      }
    }
  }
  while (!cont && 
         oldb->modifiee->to != -1 && oldb->modifiee->from != -1 &&
         oldb->modifiee->to != -2 && oldb->modifiee->from != -2);
  return result;
}

/* ----------------------------------------------------------------------
 * returns the arc with the maximum number of constraint violations
 */
Arc icArcMaxViolations(Parse parse, Lattice lat)
{
  int m = listSize(lat->arcs);
  Vector nr = vectorNew(m);
  List l = NULL;
  Arc a;
  LexemNode ln;
  int pos, n, i;
  
  /* getting out in case of incomplete params */
  if (parse == NULL) {
    return NULL;
  }
  if (parse->violations == NULL) {
    return NULL;
  }
  
  /* build the vector of arc frequencies in violations */
  for (l = parse->violations; l; l = listNext(l)) {
    ConstraintViolation v = (ConstraintViolation) listElement(l);
    if (v->lv1) {
      ln = (LexemNode) listElement(v->lv1->modifiers);
      a = ln->arc;
      pos = listIndex(lat->arcs, a);
      n = (int) vectorElement(nr, pos - 1);
      vectorSetElement(nr, (void *) (n + 1), pos - 1);
      if (v->lv1->modifiees != NULL && v->lv1->modifiees != NONSPEC) {
        ln = (LexemNode) listElement(v->lv1->modifiees);
        a = ln->arc;
        pos = listIndex(lat->arcs, a);
        n = (int) vectorElement(nr, pos - 1);
        vectorSetElement(nr, (void *) (n + 1), pos - 1);
      }
    }
    if (v->lv2) {
      ln = (LexemNode) listElement(v->lv2->modifiers);
      a = ln->arc;
      pos = listIndex(lat->arcs, a);
      n = (int) vectorElement(nr, pos - 1);
      vectorSetElement(nr, (void *) (n + 1), pos - 1);
      if (v->lv2->modifiees != NULL && v->lv2->modifiees != NONSPEC) {
        ln = (LexemNode) listElement(v->lv2->modifiees);
        a = ln->arc;
        pos = listIndex(lat->arcs, a);
        n = (int) vectorElement(nr, pos - 1);
        vectorSetElement(nr, (void *) (n + 1), pos - 1);
      }
    }
  }
  
  /* determine the arc with maximum number of violations */
  pos = 0;
  for (i = 1; i < m; i++) {
    if ((int) vectorElement(nr, i) > (int) vectorElement(nr, pos)) {
      pos = i;
    }
  }

  vectorDelete(nr);
  return listNthElement(lat->arcs, pos);
}

/* ----------------------------------------------------------------------
 * returns the violations score of an arc
 */
Number icArcScore(Arc arc, Parse parse, Lattice lat)
{
  LexemNode ln;
  List l;
  Number n = 1.0;

  for (l = parse->violations; l; l = listNext(l)) {
    ConstraintViolation v = (ConstraintViolation) listElement(l);

    if (v->lv1) {
      ln = (LexemNode) listElement(v->lv1->modifiers);
      if (arc->from == ln->arc->from && arc->to == ln->arc->to && arc->word == ln->arc->word) {
        n *= v->penalty;
      }
      if (v->lv1->modifiees != NULL && v->lv1->modifiees != NONSPEC) {
        ln = (LexemNode) listElement(v->lv1->modifiees);
        if (arc->from == ln->arc->from && arc->to == ln->arc->to && arc->word == ln->arc->word) {
          n *= v->penalty;
        }
      }
    }

    if (v->lv2) {
      ln = (LexemNode) listElement(v->lv2->modifiers);
      if (arc->from == ln->arc->from && arc->to == ln->arc->to && arc->word == ln->arc->word) {
        n *= v->penalty;
      }
      if (v->lv2->modifiees != NULL && v->lv2->modifiees != NONSPEC) {
        ln = (LexemNode) listElement(v->lv2->modifiees);
        if (arc->from == ln->arc->from && arc->to == ln->arc->to && arc->word == ln->arc->word) {
          n *= v->penalty;
        }
      }
    }
  }

  return n;
}

/* ----------------------------------------------------------------------
 * returns the arc with the smallest violations score
 */
Arc icArcSmallestScore(Parse parse, Lattice lat)
{
  List la;
  Arc aopt = NULL;
  Number nopt = 1.0;

  /* getting out in case of incomplete params */
  if (parse == NULL) {
    return NULL;
  }
  if (parse->violations == NULL) {
    return NULL;
  }
  
  /* external loop -- one cycle per word */
  for (la = lat->arcs; la; la = listNext(la)) {
    Arc a = (Arc) listElement(la);
    Number n = icArcScore(a, parse, lat);
    if (n < nopt) {
      nopt = n;
      aopt = a;
    }
  }

  return aopt;
}

/* ----------------------------------------------------------------------
 * returns the arc with the smallest violation penalty
 */
Arc icArcSmallestPenalty(Parse parse, Lattice lat)
{
  List la, l = NULL;
  Arc a, aopt;
  LexemNode ln;
  Number n, nopt;

  /* getting out in case of incomplete params */
  if (parse == NULL) {
    return NULL;
  }
  if (parse->violations == NULL) {
    return NULL;
  }
  
  /* external loop -- one cycle per word */
  aopt = NULL;
  nopt = 1.0;
  for (la = lat->arcs; la; la = listNext(la)) {
    a = (Arc) listElement(la);
    n = 1.0;
    
    /* compute word score */
    for (l = parse->violations; l; l = listNext(l)) {
      ConstraintViolation v = (ConstraintViolation) listElement(l);
      if (v->lv1) {
        ln = (LexemNode) listElement(v->lv1->modifiers);
        if (a->from == ln->arc->from && a->to == ln->arc->to && a->word == ln->arc->word) {
	  if (v->penalty < n) {
            n = v->penalty;
	  }
	}
        if (v->lv1->modifiees != NULL && v->lv1->modifiees != NONSPEC) {
          ln = (LexemNode) listElement(v->lv1->modifiees);
          if (a->from == ln->arc->from && a->to == ln->arc->to && a->word == ln->arc->word) {
  	    if (v->penalty < n) {
              n = v->penalty;
	    }
	  }
        }
      }
      if (v->lv2) {
        ln = (LexemNode) listElement(v->lv2->modifiers);
        if (a->from == ln->arc->from && a->to == ln->arc->to && a->word == ln->arc->word) {
	  if (v->penalty < n) {
            n = v->penalty;
	  }
	}
        if (v->lv2->modifiees != NULL && v->lv2->modifiees != NONSPEC) {
          ln = (LexemNode) listElement(v->lv2->modifiees);
          if (a->from == ln->arc->from && a->to == ln->arc->to && a->word == ln->arc->word) {
  	    if (v->penalty < n) {
              n = v->penalty;
	    }
	  }
        }
      }
    }
    
    if (n < nopt) {
      nopt = n;
      aopt = a;
    }
  }

  return aopt;
}
/* ----------------------------------------------------------------------
 * builds iteration information
 */
void icBuildIterInfo(ConstraintNet net, IcParams icParams, IcIter icIter) 
{
  Level main_level = inputGetMainlevel(inputCurrentGrammar);
  Lattice lattice = net->lexemgraph->lattice;
  Vector timelimits = icBuildTimeLimits(lattice->arcs);
  int node = vectorSize(timelimits) - 1;
  List n, l, prev_solution, prev_parses;
  List result, result_one, result_one_tree;
  int lastvalue, lastnode, top, i;
  Boolean cont;
  ByteVector trees; /* trees[i] = TRUE if i is a member of previous trees; FALSE otherwise */
  Arc worstarc;
  List result_nodes;
  List ll, result_one_nodes;
  Number worstscore;
  
  if (node == 1) {
    prev_parses = NULL;
  } else {
    prev_parses = icDeleteNonFrobbingParses(icBuildParsesList(net->parses, (int)vectorElement(timelimits, node - 1)));
  }
  for (n = prev_parses; n; n = listNext(n)) {
    List one_parse = NULL;
    Parse p = (Parse) listElement(n);
    for (i=0; i<vectorSize(p->LVs); i++) {
      LevelValue lv = (LevelValue) vectorElement(p->LVs, i);
      Binding b = (Binding) memMalloc(sizeof(BindingStruct));
      b->modifier = (BindingNode) memMalloc(sizeof(BindingNodeStruct));
      b->modifiee = (BindingNode) memMalloc(sizeof(BindingNodeStruct));
      icSaveBinding(lv, b, main_level);
      one_parse = listPrependElement(one_parse, b);
    }
    icIter->prev_bindings = listPrependElement(icIter->prev_bindings, one_parse);
  }
  
  trees = bvNew(node);

  result = NULL;
  for (l=icIter->prev_bindings; l; l=listNext(l)) {
    prev_solution = (List) listElement(l);
    result_one = NULL;
    lastvalue = node-1;
    cont = FALSE;
    bvSetAllElements(trees, FALSE);
    
    while (!cont) {
      lastnode = (int)vectorElement(timelimits, lastvalue);
      result_one_tree = icBuildPath(prev_solution, lastnode);
      result_one = listAppendElement(result_one, result_one_tree);
      
      top = icGetTop(prev_solution, lastnode);
      for (i=1; i<=node-1; i++) {
	int thistop = icGetTop(prev_solution, (int)vectorElement(timelimits, i));
	if (thistop == top || thistop == -1 || thistop == -2)
	  bvSetElement(trees, TRUE, i);
      }
      
      cont = TRUE;
      lastvalue = -1;
      for (i=node-1; i>0 && cont; i--) {
	if (!(bvElement(trees, i))) {
	  cont = FALSE;
	  lastvalue = i;
	}
      } 
    }
    result = listAppendElement(result, result_one);
  }
  icIter->prev_struct = result;
  
  result = NULL;
  result_nodes = NULL;
  for (l=icIter->prev_bindings, n=prev_parses; l; l=listNext(l), n=listNext(n)) {
    Parse prev_parse = (Parse) listElement(n);
    prev_solution = (List) listElement(l);

    worstarc = icArcSmallestScore(prev_parse, net->lexemgraph->lattice);
    if(worstarc) {
      result_one = icBuildPath(prev_solution, worstarc->to);
      result = listAppendElement(result, result_one);
    }

    worstscore = icArcScore(worstarc, prev_parse, net->lexemgraph->lattice);
    result_one_nodes = NULL;
    for (ll = net->lexemgraph->lattice->arcs; ll; ll = listNext(ll)) {
      Arc arc = (Arc) listElement(ll);
      if (icArcScore(arc, prev_parse, net->lexemgraph->lattice) - worstscore <= icParams->hotspot_val) {
        BindingNode bn = (BindingNode) memMalloc(sizeof(BindingNodeStruct));
        bn->from = arc->from;
        bn->to = arc->to;
        bn->word = arc->word;
        bn->le = NULL;
	result_one_nodes = listAppendElement(result_one_nodes, bn);
      }
    }
    result_nodes = listAppendElement(result_nodes, result_one_nodes);
  }
  icIter->prev_hotpath = result;
  icIter->prev_hotspots = result_nodes;

  bvDelete(trees);
  vectorDelete(timelimits);
  listDelete(prev_parses);
 
#if 0
  icPrintBindings(icIter->prev_bindings);
  icPrintStruct(icIter->prev_struct);
#endif
}

/* ----------------------------------------------------------------------
 * prints the name of the requested heuristics (standard 3)
 */
void icPrintHeuristicsName(int heurclass, int heurdist, String main_comment)
{
  cdgPrintf(CDG_INFO, "INFO: %s : ", main_comment);
  switch (heurclass) {
  case IC_HEUR_LASTNODE:
    cdgPrintf(CDG_INFO, "last node in rightmost tree "); 
    break;
  case IC_HEUR_LASTPATH:
    cdgPrintf(CDG_INFO, "nodes on last path in rightmost tree "); 
    break;
  case IC_HEUR_LASTNODEALL:
    cdgPrintf(CDG_INFO, "last node in all trees "); 
    break;
  case IC_HEUR_LASTPATHALL:
    cdgPrintf(CDG_INFO, "nodes on last path in all trees "); 
    break;
  case IC_HEUR_ALLNODE:
    cdgPrintf(CDG_INFO, "all nodes "); 
    break;
  default: 
    cdgPrintf(CDG_INFO, "not used");
  }
  if (heurdist != 0 && heurclass >= IC_HEUR_BEGIN && heurclass <= IC_HEUR_END)
    cdgPrintf(CDG_INFO, "within a distance of %d", heurdist);
  cdgPrintf(CDG_INFO, "\n");
}

/* ----------------------------------------------------------------------
 * icBuildIncomingArcs (return list of arcs)
 */
List icBuildIncomingArcs(Lattice lattice, int position)
{
  List m;
  List l = NULL;

  for (m = lattice->arcs; m; m = listNext(m))
    if (((Arc) listElement(m))->to == position) {
      l = listAppendElement(l, listElement(m));
      cdgPrintf(CDG_INFO, "INFO: considering arc '%s' : '%d' -> '%d'\n", 
		((Arc) listElement(m))->word,
		((Arc) listElement(m))->from, 
		((Arc) listElement(m))->to); 	
    }
  return l;
}

/* ----------------------------------------------------------------------
 * icBuildIncomingWord (return list of one arc)
 */
List icBuildIncomingWord(String word, int position)
{
  Arc arc = newArc( position-1, position, position-1,
                     strRegister(word),
                     0.0,
                     NULL, 0, 0);
  return listAppendElement(NULL, arc);
}

/* ---------------------------------------------------------------------
 * icPrintGraphemes
 */
void icPrintGraphemes(Vector graphemnodes)
{
  int i;
  for (i=0; i<vectorSize(graphemnodes); i++) {
    GraphemNode gn = (GraphemNode) vectorElement(graphemnodes, i);
    cdgPrintf(CDG_INFO, "%d %s\n", gn->no, gn->arc->word);
  }
}

/* ----------------------------------------------------------------------
 * prints the partial wordgraph
 */
void icPrintWordgraph(LexemGraph lexemgraph)
{
  int i;

  for (i=0; i < vectorSize(lexemgraph->graphemnodes); i++) {
    GraphemNode gn = (GraphemNode) vectorElement(lexemgraph->graphemnodes, i);
    if (gn->lexemes == NULL)
      continue;
    cdgPrintf(CDG_DEBUG, "%s ", ((LexemNode) (listElement(gn->lexemes)))->lexem->word);
  }
  cdgPrintf(CDG_DEBUG, "\n");
}

/* ----------------------------------------------------------------------
 * writes parse trees and information to LaTeX and XFig files
 */
void icWriteParsesToFiles(List parses, IcParams icParams, int position)
{
  Level main_level = inputGetMainlevel(inputCurrentGrammar);

  if (icParams->graph_file != NULL && icParams->info_file != NULL) {
    String g_f, i_f, sn;

    sn = strPrintf("%d", position);
    g_f = strAppend(icParams->graph_file, "-", sn, NULL);
    i_f = strAppend(icParams->info_file, "-", sn, NULL);
    writeParses(parses, main_level->id, g_f, i_f);
    cdgFreeString(g_f);
    cdgFreeString(i_f);
    cdgFreeString(sn);
  }
}

/* ----------------------------------------------------------------------
 * frees memory of icIter components
 */
void icIterFree(IcIter icIter)
{
  if (icIter->prev_struct != NULL) {
    icFreeStruct(icIter->prev_struct);
    icIter->prev_struct = NULL;
  }

  if (icIter->prev_bindings != NULL) {
    icFreeBindings(icIter->prev_bindings);
    icIter->prev_bindings = NULL;
  }

  if (icIter->prev_hotpath != NULL) {
    icFreeHotPath(icIter->prev_hotpath);
    icIter->prev_hotpath = NULL;
  }

  if (icIter->prev_hotspots != NULL) {
    icFreeHotSpots(icIter->prev_hotspots);
    icIter->prev_hotspots = NULL;
  }
}

/* ----------------------------------------------------------------------
 * final computations for the constraint net and cleans info from the
 * previous solution
 */
Boolean icCnBuildFinal(ConstraintNet net)
{
  if (net->cache) {
    scDelete(net->cache);
    net->cache = NULL;
  }
  
  if (!cnBuildFinal(net, TRUE)) {
    cdgPrintf(CDG_WARNING, "ERROR: invalid net\n");
    return FALSE;
  }
  
  cnPrintInfo(net);
  
  if (net->searchagenda) {
    agDelete(net->searchagenda);
    net->searchagenda = NULL;
  }
  net->evalUnary = 0;
  net->evalBinary = 0;
  net->statUnary = 0;
  if (net->cache) {
    net->cache->hits = 0;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 */
void icLgUpdateNodeNo(LexemGraph lg, int no)
{
  int i;
  for (i=0; i<vectorSize(lg->nodes); i++) {
    LexemNode ln = (LexemNode) vectorElement(lg->nodes, i);
    if (ln->no > no) {
      ln->no--;
    }
  }
}

/* ----------------------------------------------------------------------
 * downgrade the current lexeme graph - deletes the effect of the last word
 */
Boolean icLgDelLastWord(ConstraintNet net)
{
  List l;
  LexemGraph lg = net->lexemgraph;
  GraphemNode gn = (GraphemNode) vectorElement(lg->graphemnodes, vectorSize(lg->graphemnodes)-1);
  int i;
  ByteVector bs;
  
  for (l = gn->lexemes; l; l=listNext(l)) {
    LexemNode ln = (LexemNode) listElement(l);
    freeLexiconItem(ln->lexem);
    vectorRemoveElementAt(lg->nodes, ln->no);

    bs = bvNew(bvSize(lg->isDeletedNode)-1);
    for(i = 0; i < ln->no; i++) {
      if(bvElement(lg->isDeletedNode, i)) {
        bvSetElement(bs, TRUE, i);
      } else {
        bvSetElement(bs, FALSE, i);
      }
    }
    for(i = ln->no + 1; i < bvSize(lg->isDeletedNode); i++) {
      if(bvElement(lg->isDeletedNode, i)) {
        bvSetElement(bs, TRUE, i - 1);
      } else {
        bvSetElement(bs, FALSE, i - 1);
      }
    }
    
    bvDelete(lg->isDeletedNode);
    lg->isDeletedNode = bs;
    /* FIXME: This is obviously vastly inferior to the previous method
     of simply calling BooleanVectorRemoveElementAt(). */

    icLgUpdateNodeNo(lg, ln->no);
    memFree(ln);
  }
  listDelete(gn->lexemes);
  gn->lexemes = NULL;
  memFree(gn);
  vectorRemoveElementAt(lg->graphemnodes, vectorSize(lg->graphemnodes)-1);
  lg->max--;

  if (vectorSize(lg->graphemnodes) == 0) {
    lg->graphemnodes = NULL;
    lg->nodes = NULL;
    memFree(lg);
    net->lexemgraph = NULL;
    return FALSE;
  }

  if (lg->distance != NULL) {
    arrayDelete(lg->distance);
    lg->distance = NULL;
  }
  if (!lgNewFinal(lg)) {
    lgDelete(lg);
    net->lexemgraph = NULL;
    return FALSE;
  }
  net->lexemgraph = lg;
  
  return TRUE;
}

/* ----------------------------------------------------------------------
 * downgrade the current constraint net - deletes the effect of the last word
 */
Boolean icCnDelLastWord(ConstraintNet net)
{
  int i, j;
  int lvSize = vectorSize(net->lvTotals) - 1; /* position of the last entry */
  int lvTotals = (int)vectorElement(net->lvTotals, lvSize - 1); /* we need the one but last entry */
  GraphemNode lastgn = (GraphemNode) vectorElement(net->lexemgraph->graphemnodes, vectorSize(net->lexemgraph->graphemnodes)-1); /* last graphem node */

  /* delete net->nodes pointing to last grapheme node */
  for (i=0; i<vectorSize(net->nodes); i++) {
    ConstraintNode cn = (ConstraintNode) vectorElement(net->nodes, i);
    if (cn->gn == lastgn) {
      vectorRemoveElementAt(net->nodes, i);
      i--;
      cnDeleteNode(cn);
    }
  }

  /* delete last additions from net->values */
  for (i=0; i<vectorSize(net->values); i++) {
    LevelValue lv = (LevelValue) vectorElement(net->values, i);
    if (lv->indexWRTNet >= lvTotals) {
      vectorRemoveElementAt(net->values, i);
      i--;
      net->totalNumberOfValues--;
    }
  }
  net->totalNumberOfValues = vectorSize(net->values);
  
  /* clean also cn->values for all cn's */
  for (i=0; i<vectorSize(net->nodes); i++) {
    ConstraintNode cn = (ConstraintNode) vectorElement(net->nodes, i);
    for (j=0; j<vectorSize(cn->values); j++) {
      LevelValue lv = (LevelValue) vectorElement(cn->values, j);
      if (lv->indexWRTNet >= lvTotals) {
	vectorRemoveElementAt(cn->values, j);
	j--;
	cn->noValidValues--;
	lvDelete(lv);
      }
    }
    cn->noValidValues = vectorSize(cn->values);
    cn->totalNumberOfValues = cn->noValidValues;
    
    /* if a cn->values is empty, delete the node */ 
    /*
    if (cn->noValidValues == 0) {
      vectorRemoveElementAt(net->nodes, i);
      i--;
      cnDeleteNode(cn);
    }
    */
  }

  cnUndeleteAllLVs(net);
  icDeleteLastParses(net);
  vectorRemoveElementAt(net->lvTotals, lvSize); /* delete the last entry */
  return icCnBuildFinal(net);
}

/* ----------------------------------------------------------------------
 * downgrade the current lattice - deletes the last word
 */
Boolean icLatDelLastWord(ConstraintNet net)
{
  Lattice lat = net->lexemgraph->lattice;
  int position = net->lexemgraph->max;

  List l = NULL;
  Arc arc = NULL;

  for (l=lat->arcs; l; l=listNext(l)) {
    Arc a = (Arc) listElement(l);
    if (a->to == position) {
      arc = a;
    }
  }

  lat->arcs = listDeleteElement(lat->arcs, arc);
  if (listSize(lat->arcs) == 0) {
    listDelete(lat->arcs);
    lat->arcs = NULL;
    return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
 *  Checks that the binding node verifies the distance criteria
 */
Boolean icNodeDist(ConstraintNet net, BindingNode b, int heurdist, List listArcs)
{
  List m, l;
  List arcs = net->lexemgraph->lattice->arcs;
  Boolean cont = TRUE;
  Arc arc = NULL;
  Arc arcnew = NULL;
  LexemNode ln = NULL;
  LexemNode lnnew = NULL;
  GraphemNode gn;
  int d, darc;

  if (heurdist == 0)
    return TRUE;

  cont = TRUE;
  for (m=arcs; m && cont; m=listNext(m)) {
    arc = (Arc) listElement(m);
    cont = !(b->from == arc->from && 
	     b->to == arc->to &&
	     b->word == arc->word);
  }
  gn = cnGetGraphemNodeFromArc(net, arc);
  if (gn == NULL) {
    cdgPrintf(CDG_WARNING, "WARNING: No lexical information for '%s'\n", arc->word);
    return TRUE;
  }
  if (gn->lexemes == NULL) {
    cdgPrintf(CDG_WARNING, "WARNING: No lexical information for '%s'\n", arc->word);
    return TRUE;
  }

  l = gn->lexemes;
  ln = (LexemNode) listElement(l);
  
  d = -1;
  for (m=listArcs; m; m=listNext(m)) {
    arcnew = (Arc) listElement(m);
    gn = cnGetGraphemNodeFromArc(net, arcnew);
    if (gn == NULL) {
      cdgPrintf(CDG_WARNING, "WARNING: No lexical information for '%s'\n", arcnew->word);
      return TRUE;
    }
    if (gn->lexemes == NULL) {
      cdgPrintf(CDG_WARNING, "WARNING: No lexical information for '%s'\n", arcnew->word);
      return TRUE;
    }
    l = gn->lexemes;
    lnnew = (LexemNode) listElement(l);
    
    darc = (int) lgDistanceOfNodes(net->lexemgraph, ln, lnnew);
    if (d == -1)
      d = darc;
    else if (d > darc)
      d = darc;
  }
  return (d <= heurdist);
}

/* ----------------------------------------------------------------------
 * Checks that the binding node satisfies the heuristics positioning criteria
 */
Boolean icNodeCheck(ConstraintNet net, BindingNode l, 
		    int heurclass, int heurdist,
		    Boolean hotspot, Boolean hotpath,
		    List listArcs, IcIter icIter) 
{
  List m, n, p;
  List prev_struct     = icIter->prev_struct;
  List prev_hotpath  = icIter->prev_hotpath;
  List prev_hotspots = icIter->prev_hotspots;

  /* allow everything on all nodes with score under icParams->hotspot_val + score of hotspot */
  if (hotspot) {
    for (m=prev_hotspots; m; m=listNext(m)) {
      for (n=(List) listElement(m); n; n=listNext(n)) {
        BindingNode b = (BindingNode) listElement(n);
          if (l->from == b->from && l->to == b->to && l->word == b->word)
            if (icNodeDist(net, b, heurdist, listArcs))
              return TRUE;
      }
    }
  }
  
  /* allow everything for the nodes from hotspot to root */
  if (hotpath) {
    for (m=prev_hotpath; m; m=listNext(m)) {
      List prev_solution = (List) listElement(m);
      for (n=prev_solution; n; n=listNext(n)) {
        BindingNode b = (BindingNode) listElement(n);
        if (l->from == b->from && l->to == b->to && l->word == b->word)
          if (icNodeDist(net, b, heurdist, listArcs))
            return TRUE;
      }
    }
  }

  switch (heurclass) {
  case IC_HEUR_LASTNODE:
    for (m=prev_struct; m; m=listNext(m)) {
      List prev_solution = (List) listElement(m);
      List solution_tree = (List) listElement(prev_solution);
      BindingNode b = (BindingNode) listElement(solution_tree);
      if (l->from == b->from && l->to == b->to && l->word == b->word)
	if (icNodeDist(net, b, heurdist, listArcs))
	  return TRUE;
    }
    return FALSE;
  case IC_HEUR_LASTPATH:
    for (m=prev_struct; m; m=listNext(m)) {
      List prev_solution = (List) listElement(m);
      for (n=listElement(prev_solution); n; n=listNext(n)) {
	BindingNode b = (BindingNode) listElement(n);
	if (l->from == b->from && l->to == b->to && l->word == b->word)
	  if (icNodeDist(net, b, heurdist, listArcs))
	    return TRUE;
      }
    }
    return FALSE;
  case IC_HEUR_LASTNODEALL:
    for (m=prev_struct; m; m=listNext(m))
      for (n=(List) listElement(m); n; n=listNext(n)) {
	List solution_tree = (List) listElement(n);
	BindingNode b = (BindingNode) listElement(solution_tree);
	if (l->from == b->from && l->to == b->to && l->word == b->word)
	  if (icNodeDist(net, b, heurdist, listArcs))
	    return TRUE;
      }
    return FALSE;
  case IC_HEUR_LASTPATHALL:
    for (m=prev_struct; m; m=listNext(m))
      for (n=(List) listElement(m); n; n=listNext(n))
	for (p=(List) listElement(n); p; p=listNext(p)) {
	  BindingNode b = (BindingNode) listElement(p);
	  if (l->from == b->from && l->to == b->to && l->word == b->word)
	    if (icNodeDist(net, b, heurdist, listArcs))
	      return TRUE;
	}
    return FALSE;
  case IC_HEUR_ALLNODE:
    return icNodeDist(net, l, heurdist, listArcs);
  default:
    return FALSE;
  }
}

/* ----------------------------------------------------------------------
 * Compares the levelvalue with solution values and checks heuristics
 * for old nodes only
 */
Boolean icLevelValueSolutionCheckOld(ConstraintNet net, LevelValue lv, List listArcs, IcParams icParams, IcIter icIter)
{
  List prev_bindings = icIter->prev_bindings;

  List m, n; 
  for (m=prev_bindings; m; m=listNext(m))
    for (n=listElement(m); n; n=listNext(n)) {
      Binding b = (Binding) listElement(n);
      Boolean cond = TRUE;
      
      cond = cond && b->modifier->from == lv->modifier->arc->from;
      cond = cond && b->modifier->to == lv->modifier->arc->to;
      cond = cond && b->level == lv->level; 
      
      if (!cond)
	continue;
      
      /* check heuristic 3 (allow lexical change) for modifier */
      if (icNodeCheck(net, b->modifier, 
		      icParams->heurclass3, icParams->heurclass3dist,
		      icParams->hotspot, icParams->hotpath, 
		      listArcs, icIter))
	cond = cond && b->modifier->word == lv->modifier->lexem->word;
      else
	cond = cond && lgLexemeInLexemNodeList(b->modifier->le, lv->modifiers);
      
      /* check heuristic 4 (allow label change) for modifier */
      if (icNodeCheck(net, b->modifier,
		      icParams->heurclass4, icParams->heurclass4dist,
		      icParams->hotspot, icParams->hotpath, 
		      listArcs, icIter))
	;
      else
	cond = cond && b->label == lv->label;
      
      /* check heuristic 5 (allow governor change) for modifier */
      if (icNodeCheck(net, b->modifier,
		      icParams->heurclass5, icParams->heurclass5dist,
		      icParams->hotspot, icParams->hotpath, 
		      listArcs, icIter))
	;
      else {
	if (lv->modifiee == NULL) {
	  cond = cond && b->modifiee->to == -1;
	  cond = cond && b->modifiee->from == -1;
	  cond = cond && b->modifiee->word == NULL;
	  cond = cond && b->modifiee->le == NULL;
	} else if (lv->modifiee == NONSPEC) {
	  cond = cond && b->modifiee->to == -2;
	  cond = cond && b->modifiee->from == -2;
	  cond = cond && b->modifiee->word == NONSPEC;
	  cond = cond && b->modifiee->le == NONSPEC;
	} else {
	  cond = cond && b->modifiee->to == lv->modifiee->arc->to;
	  cond = cond && b->modifiee->from == lv->modifiee->arc->from;
	  
	  /* check heuristic 3 (allow lexical change) for modifiee */
	  if (icNodeCheck(net, b->modifiee,
			  icParams->heurclass3, icParams->heurclass3dist,
  		          icParams->hotspot, icParams->hotpath, 
			  listArcs, icIter))
	    cond = cond && b->modifiee->word == lv->modifiee->lexem->word;
	  else
	    cond = cond && lgLexemeInLexemNodeList(b->modifiee->le, lv->modifiees);
	}
      }
      if (cond)
	return TRUE;
    }
  return FALSE;
}

/* ----------------------------------------------------------------------
 * Compares the levelvalue with solution values and checks heuristics
 * for new nodes only
 */
Boolean icLevelValueSolutionCheckNew(ConstraintNet net, LevelValue lv, List listArcs, IcParams icParams, IcIter icIter)
{
  BindingNode b;
  Boolean cont = FALSE;
  b = (BindingNode) memMalloc(sizeof(BindingNodeStruct));

  /* check heuristic 1 (allow link from new node) */
  if (lv->modifiee != NULL &&
      lv->modifiee != NONSPEC)
    icSaveBindingNode(lv->modifiee, b);
  if (lv->modifiee == NULL ||
      lv->modifiee == NONSPEC ||
      icNodeCheck(net, b,
		  icParams->heurclass1, icParams->heurclass1dist,
  		  FALSE, FALSE, 
		  listArcs, icIter)) {
    List p;
    for (p=listArcs; p; p=listNext(p)) {
      Arc arc = (Arc) listElement(p);
      if (lv->modifier->arc == arc) {
	cont = TRUE;
	break;
      }
    }
  }
  
  /* check heuristic 2 (allow link to new node) */
  icSaveBindingNode(lv->modifier, b);
  if (icNodeCheck(net, b,
		  icParams->heurclass2, icParams->heurclass2dist,
  		  FALSE, FALSE, 
		  listArcs, icIter)) {
    List p;
    for (p=listArcs; p; p=listNext(p)) {
      Arc arc = (Arc) listElement(p);
      if (lv->modifiee != NULL && lv->modifiee != NONSPEC && lv->modifiee->arc == arc) {
	cont = TRUE;
	break;
      }
    }
  }
  
  memFree(b);
  return cont;
}

/* ----------------------------------------------------------------------
 * recomputes the levelvalue score
 */
void icRecomputeLvScore(LevelValue lv, ConstraintNet net)
{
  lv->score = evalUnary(lv, net, NULL, FALSE, NULL, NULL);
  if (statUseStatisticsFlag)
    lv->score *= statUnary(lv, net);
  if (lv->score == 0.0)
    lv->isDeleted = TRUE;
}

/* ----------------------------------------------------------------------
 * build constraint nodes corresponding to a graphemnode
 */
Boolean cnBuildIterHeur(ConstraintNet net, GraphemNode gn,
				       int heurclass, int heurdist,
				       Boolean hotspots, Boolean hotpath,
				       List listArcs, IcIter icIter)
{
  Level level;
  GraphemNode modifiee;
  ConstraintNode node;
  List l;
  int k, z;
  int ggs = vectorSize(net->lexemgraph->graphemnodes);
  int lgs = vectorSize(net->lexemgraph->nodes);
  int lvs = listSize(inputCurrentGrammar->levels);  

  if (gn->lexemes == NULL)
    return TRUE;

  cdgFlush();

  /* loop through all levels */
  for (l = inputCurrentGrammar->levels; l != NULL; l = listNext(l)) {
    level = (Level) listElement(l);
    if (!level->useflag)
      continue;

    /* check for ^C here */
    if (cdgCtrlCTrapped) {

      /* delete all constraint-nodes */
      for (z = 0; z < vectorSize(net->nodes); z++) {
        node = (ConstraintNode) vectorElement(net->nodes, z);
        cnDeleteNode(node);
      }
      vectorDelete(net->nodes);
      net->nodes = NULL;
      return FALSE;
    }
    cdgFlush();

    /* if an entire grapheme is deleted from the graph already, 
       do not even build the constraint node. */
    if (lgAreDeletedNodes(net->lexemgraph, gn->lexemes)) {
      continue;
    }
      
    /* allocate new constraint node */
    node = (ConstraintNode) memMalloc(sizeof (ConstraintNodeStruct));
    node->net = net;
    node->level = level;
    node->gn = gn;
    node->values = vectorNew(lgs * lvs);
    node->totalNumberOfValues = 0;
      
    /* fill it */
    for (k = 0; k <= ggs; k++) {
	BindingNode b;
      if(k == ggs) {
        modifiee = NULL;
      } else {
        modifiee = vectorElement(net->lexemgraph->graphemnodes, k);
      }
	b = (BindingNode) memMalloc(sizeof(BindingNodeStruct));
      if(modifiee) {
        icSaveBindingNode((LexemNode) listElement(modifiee->lexemes), b);
      }
	if (!modifiee ||
          icNodeCheck(net, b, heurclass, heurdist, hotspots, hotpath, listArcs, icIter)) {
	  cnBuildLevelValues(node, level, gn, modifiee);
      }
        memFree(b);
    }
    
    /* add it to the net */
    node->noValidValues = vectorSize(node->values);
    node->totalNumberOfValues = node->noValidValues;
    vectorAddElement(net->nodes, node);
  }
  return TRUE;
}


/* ----------------------------------------------------------------------
 * Compare parses between iterations and print the differences
 */
void icCompareParsesBetweenIterations(List this_parses, List prev_parses, int position)
{
  Parse p1min = NULL;
  Parse p2min = NULL;
  int difmin = -1;
  ParseCmpStatStruct stat;

  for (; prev_parses; prev_parses = listNext(prev_parses))
    for (; this_parses; this_parses = listNext(this_parses)) {
      Parse p1 = (Parse) listElement(prev_parses);
      Parse p2 = (Parse) listElement(this_parses);
      int dif = parseCompare(p2, p1, &stat, PC_SILENT, PC_NONSTRICT);

      if (difmin == -1) {
	difmin = dif;
	p1min = p1;
	p2min = p2;
      } else if (dif < difmin) {
	difmin = dif;
	p1min = p1;
	p2min = p2;
      }
    }

  if ((p1min != NULL) && (p2min != NULL)) {
    cdgPrintf(CDG_INFO,
	      "INFO: SUCC: iteration %2d, parse %8s - iteration %2d, parse %8s - differences %3d\n",
	      position - 1, p1min->id, position, p2min->id, difmin);
  } else {
    cdgPrintf(CDG_INFO,
	      "INFO: SUCC: iteration %2d, parse %8s - iteration %2d, parse %8s - differences %3d\n",
	      position - 1, "NULL", position, "NULL", -1);
  }
}

/* ----------------------------------------------------------------------
 * Compare parses to reference parse and print differences
 */
void icCompareParsesToReference(List parses, Parse refparse, Boolean totalsOnly)
{
  List m;
  ParseCmpStatStruct stat;

  if (parses == NULL) {
    return;
  }

  if (totalsOnly) { 
    /* print only totals, i.e. one line per parse */
    int dif;
    int position = icGetTimeLimitsNumber(((Parse) listElement(parses))->lg->lattice->arcs) - 1;
    for(m=parses; m; m=listNext(m)) {
      Parse p = (Parse) listElement(m);
      if (refparse) {
	dif = parseCompare(refparse, p, &stat, PC_SILENT, PC_NONSTRICT);
      } else {
	dif = -1;
      }
      cdgPrintf(CDG_INFO, "INFO: DIFFS: iteration %2d, parse %8s (%s), differences %3d, score %4.3e\n", position, p->id, p->searchStrategy, dif, p->score);
    } 
  } else {
    /* print detailed information about differences */
    if (refparse) {
      for(m=parses; m; m=listNext(m)) {
	Parse p = (Parse) listElement(m);
	parseCompare(refparse, p, &stat, PC_ALL, PC_NONSTRICT);
      } 
    }
  }
}

/* ----------------------------------------------------------------------
 * apply all heuristics to all the constraint net
 */
Boolean icApplyHeuristicsAllNet(ConstraintNet net, List listArcs, IcParams icParams, IcIter icIter)
{
  List m;
  int i, j;

  /* update old constraint nodes */
  cdgPrintf(CDG_DEBUG, "DEBUG: update old cn's\n");
  for (i=0; i<vectorSize(net->nodes); i++) {
    ConstraintNode cn = (ConstraintNode) vectorElement(net->nodes, i);
    for (j=0; j<vectorSize(cn->values); j++) {
      LevelValue lv = (LevelValue) vectorElement(cn->values, j);
      
      if (icLevelValueSolutionCheckOld(net, lv, listArcs, icParams, icIter))
	lv->isDeleted = FALSE;
	  
      icRecomputeLvScore(lv, net);
    }
  }
  
  /* create new lv's for old cn's */
  cdgPrintf(CDG_DEBUG, "DEBUG: new lv's for old cn's\n");
  /* check heuristic 2 (allow links to new node) */
  if (icParams->heurclass2 != 0)
    for (i=0; i<vectorSize(net->nodes); i++) {
      ConstraintNode cn = (ConstraintNode) vectorElement(net->nodes, i);
      Level level = cn->level;
      GraphemNode modifier = cn->gn;
      BindingNode b = (BindingNode) memMalloc(sizeof(BindingNodeStruct));
      icSaveBindingNode(listElement(cn->gn->lexemes), b);
      
      if (icNodeCheck(net, b,
		      icParams->heurclass2, icParams->heurclass2dist,
		      FALSE, FALSE,
		      listArcs, icIter)) {
#ifdef DEBUGCNBUILDNODES
	cdgPrintf(CDG_DEBUG, "\nDEBUG: generating LevelValues for node ");
	cnPrintNode(CDG_DEBUG, cn);
	cdgPrintf(CDG_DEBUG, "\n");
#endif
	for (m=listArcs; m; m=listNext(m)) {
	  GraphemNode modifiee = cnGetGraphemNodeFromArc(net, (Arc) listElement(m));
	  if (modifiee == NULL) {
	    continue;
	  }
	  cnBuildLevelValues(cn, level, modifier, modifiee);
	}
	cn->noValidValues = vectorSize(cn->values);
	cn->totalNumberOfValues = cn->noValidValues;
      }
      memFree(b);
    }
  
  /* create new cn's and new lv's for them */
  cdgPrintf(CDG_DEBUG, "DEBUG: new lv's for new cn's\n");
  /* check heuristic 1 (allow links from new node) */
  if (icParams->heurclass1 != 0) {
    for (m=listArcs; m; m=listNext(m)) {
      GraphemNode gn = cnGetGraphemNodeFromArc(net, (Arc) listElement(m));
      if (gn == NULL) {
	continue;
      }
      if (!cnBuildIterHeur(net, gn,
					     icParams->heurclass1, icParams->heurclass1dist,
					     FALSE, FALSE,
					     listArcs, icIter))
	return FALSE;
    }
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
   as cnRenew, but does not delete parses
*/
Boolean icCnRenew(ConstraintNet net)
{
  int i, j;

  /* refresh constraint nodes and LVs */
  for (i = 0; i < vectorSize(net->nodes); i++) {
    ConstraintNode node = (ConstraintNode) vectorElement(net->nodes, i);

    for (j = 0; j < vectorSize(node->values); j++) {
      LevelValue lv = vectorElement(node->values, j);

      lv->limit = lv->score;
      lv->isDeleted = FALSE;
      lv->no = -1;
    }
    node->noValidValues = vectorSize(node->values);
  }

  /* refresh individual lexemes */
  for (i = 0; i < vectorSize(net->lexemgraph->nodes); i++) {
    LexemNode ln = vectorElement(net->lexemgraph->nodes, i);
    ln->limit = 1.0;
  }

  /* undo lexem deletions */
  bvSetAllElements(net->lexemgraph->isDeletedNode, FALSE);
  lgComputeNoOfPaths(net->lexemgraph);

  /* throw away score-cache and build a new one */
  scDelete(net->cache);
  if (scUseCache) {
    net->cache = scNew(vectorSize(net->values));
  } else {
    net->cache = NULL;
  }
  
  /* re-sort the LVs by unary score */
  cnSortLVs(net);

  /* optimize the net */
  if(cnOptimizeNet(net) < 0) {
    return FALSE;
  };

  return TRUE;  
}

/* ----------------------------------------------------------------------
 * calls frobbing and does associated stuff 
 */
void icFrobbing(ConstraintNet net, IcParams icParams, Boolean lastIteration, Timer timer)
{
  int node = icGetTimeLimitsNumber(net->lexemgraph->lattice->arcs) - 1;
  int timeElapsed;
  String invocation =
    strPrintf("frobbing net=%s method=combined pressure=0.0 execute=q",
              net->id);
 
  if ((node % icParams->frob_iter != 0) && !lastIteration) {
    return;
  }

  icCnRenew(net);

  if (timer && icParams->frob_time != 0.0) {
    timeElapsed = timerElapsed(timer);
    timerSetAlarm(icParams->frob_time * timeElapsed);
  }
  frobSetPressure(icParams->frob_pressure);
  frobSetAgendaSize(icParams->frob_agendasize);
  frobSetBeamWidth(icParams->frob_beamwidth);
  commandEval(invocation);
  if (timer && icParams->frob_time != 0.0) {
    timerStopAlarm();
    timerStart(timer);
  }

  icCnRenew(net);

}

/* ----------------------------------------------------------------------
 * returns the list of parses from the most recent iteration at 
 * which parses have been produced
 */
List icGetMostRecentParses(ConstraintNet net, List net_parses)
{
  Vector timelimits = icBuildTimeLimits(net->lexemgraph->lattice->arcs);
  int node = vectorSize(timelimits) - 1;
  List prev_parses = NULL;
  int i;
  
  /* find the most recent parses */
  for (i=node-1; i>0 && !prev_parses; i--) {
    prev_parses = icDeleteNonFrobbingParses(icBuildParsesList(net_parses, (int)vectorElement(timelimits, i)));
  }

  return prev_parses;

/*
  if (prev_parses) {
    return listAppendElement(NULL, (Parse) listElement(prev_parses));
  } else {
    return NULL;
  }
*/
}

/* -------------------------------
 * icIterInit
 */
void icIterInit(IcIter icIter)
{
  icIter->prev_bindings = NULL;
  icIter->prev_struct = NULL;
  icIter->prev_hotpath = NULL;
  icIter->prev_hotspots = NULL;
}

/* -------------------------------
 * icParamsInit
 */
void icParamsInit(IcParams icParams)
{
  /* print names of all heuristics required by user */
  if (icParams->method_ns) {
    cdgPrintf(CDG_INFO, "INFO: Run netsearch unrestricted\n");
  } else {
    cdgPrintf(CDG_INFO, "INFO: Does not run netsearch\n");
  }
  if (icParams->method_fr) {
    cdgPrintf(CDG_INFO, "INFO: Run frobbing\n");
    cdgPrintf(CDG_INFO, "      every %d iterations\n", icParams->frob_iter);
    cdgPrintf(CDG_INFO, "      time limit %f\n", icParams->frob_time);
    cdgPrintf(CDG_INFO, "      pressure %f\n", icParams->frob_pressure);
    cdgPrintf(CDG_INFO, "      beam with %d\n", icParams->frob_beamwidth);
    cdgPrintf(CDG_INFO, "      agenda size %d\n", icParams->frob_agendasize);
  } else {
    cdgPrintf(CDG_INFO, "INFO: Does not run frobbing\n");
  }
  icPrintHeuristicsName(icParams->heurclass1, icParams->heurclass1dist, "Allow links from new node     ");
  icPrintHeuristicsName(icParams->heurclass2, icParams->heurclass2dist, "Allow links to new node       ");
  icPrintHeuristicsName(icParams->heurclass3, icParams->heurclass3dist, "Allow changing lexical reading");
  icPrintHeuristicsName(icParams->heurclass4, icParams->heurclass4dist, "Allow changing label          ");
  icPrintHeuristicsName(icParams->heurclass5, icParams->heurclass5dist, "Allow changing governor       ");
  if (icParams->hotspot) {
    cdgPrintf(CDG_INFO, "INFO: Allow all changes for nodes with scores smaller than hotspot's + %f\n", icParams->hotspot_val);
  }
  if (icParams->hotpath) {
    cdgPrintf(CDG_INFO, "INFO: Allow all changes for nodes from the first hotspot to the root node\n");
  }
}

/* -------------------------------
 * icCnInit
 */
void icCnInit(ConstraintNet net, Lattice lattice)
{
  Lattice lat = (Lattice)memMalloc( sizeof(LatticeStruct) );

  /* add structures for tracing heuristics effect */

  /* initialize local data structure */
  if (lattice == NULL) {
    cdgPrintf(CDG_INFO, "No reference parse for dynamic wordgraphs\n");
  } else {
    Parse ref_parse = parseFind(lattice->id);
    if (ref_parse == NULL) {
      cdgPrintf(CDG_INFO, "Reference parse %s not found\n", lattice->id);
    }
  }

  /* initialize the partial wordgraph */
  if (lattice != NULL) {
    lat->id = strRegister(lattice->id);
    lat->arcs = NULL;
    lat->filename = lattice->filename;
    lat->lineNo = lattice->lineNo;
  } else {
    lat->id = strRegister("interactive");
    lat->arcs = NULL;
    lat->filename = NULL;
    lat->lineNo = 0;
  }

  /* initialize the constraint net and the lexemgraph */
  net->lexemgraph = lgNewInit();
  net->lexemgraph->lattice = lat;
  net->totalNumberOfValues = 0;
  net->nodes = vectorNew(0);
  net->values = vectorNew(0);
  net->edges = NULL;
  net->parses = (List) NULL;
  net->searchagenda = (Agenda) NULL;
  net->evalUnary = 0;
  net->evalBinary = 0;
  net->statUnary = 0;
  net->cache = NULL;
  net->lvTotals = vectorNew(0);

  /* start the timer */
  if (icTimer == NULL) {
    icTimer = timerNew(icTimer);
  } else {
    timerStart(icTimer);
  }
}

/* ----------------------------------------------------------------------
 * icCnFree
 */
void icCnFree(ConstraintNet net)
{
  int i;
  ConstraintNode node;
  ConstraintEdge edge;
  
  if (net == NULL) {
    cdgPrintf(CDG_WARNING, "WARNING: cnDelete: net is NULL\n");
    return;
  }

  if (net->lexemgraph != NULL) {
    lgDelete(net->lexemgraph);
    net->lexemgraph = NULL;
  }

  if (net->values != NULL) {
    vectorDelete(net->values);
    net->values = NULL;
  }
  
  if (net->nodes != NULL) {
    for (i = 0; i < vectorSize(net->nodes); i++) {
      node = (ConstraintNode) vectorElement(net->nodes, i);
      cnDeleteNode(node);
    }
    vectorDelete(net->nodes);
    net->nodes = NULL;
  }
  
  if (net->edges != NULL) {
    for (i = 0; i < vectorSize(net->edges); i++) {
      edge = (ConstraintEdge) vectorElement(net->edges, i);
      cnDeleteEdge(edge);
    }
    vectorDelete(net->edges);
    net->edges = NULL;
  }
  
  if (net->searchagenda != NULL) {
    agDelete(net->searchagenda);
    net->searchagenda = NULL;
  }
  
  if (net->cache != NULL) {
    scDelete(net->cache);
    net->cache = NULL;
  }
  
  if (net->lvTotals != NULL) {
    vectorDelete(net->lvTotals);
    net->lvTotals = NULL;
  }

  /* destroy the timer */
  if (icTimer != NULL) {
    timerFree(icTimer);
    icTimer = NULL;
  }
}

/* ----------------------------------------------------------------------
 * icInitializeNet
 */
ConstraintNet icInitializeNet(Lattice lattice, IcParams icParams)
{
  ConstraintNet net = cnBuildInit();
  icParamsInit(icParams);
  icCnInit(net, lattice);

  return net;
}

/* ----------------------------------------------------------------------
 * icAddArcsToNet
 */
Boolean icAddArcsToNet(ConstraintNet net, List listArcs, IcParams icParams)
{
  int i, node;

  IcIter icIter = (IcIter) memMalloc(sizeof(IcIterStruct));
  icIterInit(icIter);

  /* update the partial wordgraph */
  net->lexemgraph->lattice->arcs = listAppendList(net->lexemgraph->lattice->arcs, listClone(listArcs));
  /* mark all lv's as undeleted */
  cnUndeleteAllLVs(net);
  /* update the lexemgraph */
  if (!lgUpdateArcs(net->lexemgraph, net->lexemgraph->lattice, listArcs)) {
    cdgPrintf(CDG_WARNING, "ERROR: invalid lexemgraph\n");
    return FALSE;
  }

  /* DEBUG: print the wordgraph being considered at this step */
  icPrintWordgraph(net->lexemgraph);
  
  /* prepare helping structures based on previous iteration parses */
  icBuildIterInfo(net, icParams, icIter);
  
  /* update the constraint net */
  if (!cnBuildUpdateArcs(net, listArcs)) {
    cdgPrintf(CDG_WARNING, "WARNING: invalid net, net contains no nodes\n");
    return FALSE;
  }

  /* 
   * mark those values as deleted that aren't valid extensions 
   * of the previous solution
   */
  if (icParams->method_ns) {
    for (i=0; i < vectorSize(net->values); i++) {
      LevelValue lv = (LevelValue)vectorElement(net->values, i);
      lv->isDeleted = TRUE;
    
      if (!lvTestAllDeleted(lv, net->lexemgraph->isDeletedNode)) {
        continue;
      }
    
      if (icLevelValueSolutionCheckOld(net, lv, listArcs, icParams, icIter)) {
        lv->isDeleted = FALSE;
      } else if (icLevelValueSolutionCheckNew(net, lv, listArcs, icParams, icIter)) {
        lv->isDeleted = FALSE;
      }
      
      icRecomputeLvScore(lv, net);
    }
  } else {
    cnUndeleteAllLVs(net);
  }   

  /*
   * final computations for the constraint net and
   * cleaning the information from the previous solution
   */
  if (!icCnBuildFinal(net)) {
    return FALSE;
  }
  
  node = icGetTimeLimitsNumber(net->lexemgraph->lattice->arcs) - 1;
  vectorSetElement(net->lvTotals, (void *)net->totalNumberOfValues, node);

  /* DEBUG: print active levelvalues */
  cnPrintActiveLVs(net);
  
  /* free iteration structures */
  icIterFree(icIter);

  /* cnPrint(CDG_INFO, net); */

  return TRUE;
}

/* ----------------------------------------------------------------------
 * icAddArcsToNetSmaller
 */
Boolean icAddArcsToNetSmaller(ConstraintNet net, List listArcs, IcParams icParams)
{
  int node;

  IcIter icIter = (IcIter) memMalloc(sizeof(IcIterStruct));
  icIterInit(icIter);

  /* build list of incoming arcs */
  net->lexemgraph->lattice->arcs = listAppendList(net->lexemgraph->lattice->arcs, listClone(listArcs));
  /* update the lexemgraph */
  if (!lgUpdateArcs(net->lexemgraph, net->lexemgraph->lattice, listArcs)) {
    cdgPrintf(CDG_WARNING, "ERROR: invalid lexemgraph\n");
    return FALSE;
  }
  
  /* DEBUG: print the wordgraph being considered at this step */
  icPrintWordgraph(net->lexemgraph);
  
  /* prepare helping structures based on previous iteration parses */
  icBuildIterInfo(net, icParams, icIter);

  /* apply all heuristics to all the net*/
  cnDeleteAllLVs(net);
  if (!icApplyHeuristicsAllNet(net, listArcs, icParams, icIter)) {
    cdgPrintf(CDG_WARNING, "WARNING: invalid net, net contains no nodes\n");
    return FALSE;
  }
  
  /*
   * final computations for the constraint net and
   * cleaning the information from the previous solution
   */
  if (!icCnBuildFinal(net)) {
   return FALSE;
  }
  
  node = icGetTimeLimitsNumber(net->lexemgraph->lattice->arcs) - 1;
  vectorSetElement(net->lvTotals, (void *)net->totalNumberOfValues, node);

  /* DEBUG: print active levelvalues */
  cnPrintActiveLVs(net);
  
  /* free iteration structures */
  icIterFree(icIter);
  return TRUE;
}
    
/* -------------------------------
 * icSearchNet
 */
Boolean icSearchNet(ConstraintNet net, IcParams icParams, Boolean lastIteration, int no, char **args)
{
  List parses = NULL;
  List net_parses = NULL;
  Parse p;
  
  parses = net->parses;
  net->parses = NULL;
  
  if (icParams->method_ns) {
    /* do the search on this adjusted constraints net */
    net_parses = nsBranchBound(net, no, args);
    if (!net_parses) {
      cdgPrintf(CDG_WARNING, "WARNING: search failed\n");
      return FALSE;
    }
    nsRegisterParses(net, net_parses);
  } else {
    /* else, pass the most recent parses to frobbing */
    net->parses = icGetMostRecentParses(net, parses);
    net_parses = net->parses;
  }

  /* call frobbing and do associated stuff */
  if (icParams->method_fr) {
    icFrobbing(net, icParams, lastIteration, icTimer);
  }

  if (! icParams->method_ns) {
    if (net->parses && net->parses != net_parses) {
      listDelete(listNext(net->parses));
      listSetNext(net->parses, NULL);
    }
  }

  if (net->parses != NULL) {
    p = parseClone(listElement(net->parses));
    cdgFreeString(p->searchStrategy);
    p->searchStrategy = strRegister("IC");
    parseRegister(p);
    cdgExecHook(HOOK_PARTIALRESULT, p);
  }

  net->parses = listAppendList(net->parses, parses);

  return TRUE;
}

/* -------------------------------
 * icPrintSolutions
 */
void icPrintSolutions(ConstraintNet net, IcParams icParams)
{
  Vector timelimits = icBuildTimeLimits(net->lexemgraph->lattice->arcs);
  int node = vectorSize(timelimits) - 1;
  Parse ref_parse = parseFind(net->lexemgraph->lattice->id);
  List this_parses = icBuildParsesList(net->parses, net->lexemgraph->max);
  List prev_parses = NULL;

  if (hkVerbosity & CDG_DEBUG) {
    if (ref_parse) {
      icCompareParsesToReference(this_parses, ref_parse, TRUE);
    }
    this_parses = icDeleteNonFrobbingParses(this_parses);
    if (node > 1) {
      prev_parses = icDeleteNonFrobbingParses(icBuildParsesList(net->parses, (int)vectorElement(timelimits, node - 1))); 
      icCompareParsesBetweenIterations(this_parses, prev_parses, node);
    }
    if (ref_parse) {
      icCompareParsesToReference(this_parses, ref_parse, FALSE);
    }
  }

  if (hkVerbosity & CDG_SEARCHRESULT) {
    List l;
    for(l = this_parses; l; l = listNext(l)) {
      parsePrint(listElement(l));
    }
  }

  icWriteParsesToFiles(this_parses, icParams, node);

  listDelete(this_parses);
  listDelete(prev_parses);
}

/* -------------------------------
 * icPrintSolutionsFinal
 */
void icPrintSolutionsFinal(ConstraintNet net)
{
  int n;
  List l;
  List this_parses = icDeleteNonFrobbingParses(icBuildParsesList(net->parses, net->lexemgraph->max));
  Parse ref_parse = parseFind(net->lexemgraph->lattice->id);
  ParseCmpStatStruct stat;

  n = listSize(this_parses);
  cdgPrintf(CDG_INFO, "\nINFO: %d solution(s) found\n", n);

    for(l = this_parses; l; l = listNext(l)) {
      parsePrint(listElement(l));
      if (ref_parse) {
        cdgPrintf(CDG_INFO, "\n");
        parseCompare(listElement(l), ref_parse, &stat, PC_MEASURES, PC_STRICT);
      }
    }
    cdgPrintf(CDG_INFO, "\n");
  
  listDelete(this_parses);
}

/* -------------------------------
 * icDeleteWord
 */
Boolean icDeleteWord(ConstraintNet net)
{
  int node = listSize(net->lexemgraph->lattice->arcs);

  cdgPrintf(CDG_INFO, "\n");
  if (node == 1) {
    cdgPrintf(CDG_INFO, "INFO: Reinitializing the data structures\n");
    icCnFree(net);
    icCnInit(net, NULL);
  } else {
    /* downgrade the lattice */
    cdgPrintf(CDG_INFO, "INFO: Downgrading the wordgraph\n");
    if(!icLatDelLastWord(net)) {
      return FALSE;
    }
    
    /* downgrade the constraint net */
    cdgPrintf(CDG_INFO, "INFO: Downgrading the constraint net\n");
    if (!icCnDelLastWord(net)) {
      return FALSE;
    }
    
    /* downgrade the lexeme graph */
    cdgPrintf(CDG_INFO, "INFO: Downgrading the lexemgraph\n");
    if (!icLgDelLastWord(net)) {
      cdgPrintf(CDG_WARNING, "ERROR: invalid lexemgraph\n");
      return FALSE;
    }
  }

  /* cnPrint(CDG_INFO, net); */
  cnPrintInfo(net);
  cnPrintActiveLVs(net);

  return TRUE;
}
/* -------------------------------
 * icAddWord
 */
Boolean icAddWord(ConstraintNet net, String word, IcParams icParams, int no, char **args)
{
  List list;

  cdgPrintf(CDG_INFO, "\n");
  list = icBuildIncomingWord(word, net->lexemgraph->max + 1);

  if (!icAddArcsToNet(net, list, icParams)) {
    return FALSE;
  }
  if (!icSearchNet(net, icParams, TRUE, no, args)) {
    return FALSE;
  }
  /* icPrintSolutions(net, icParams); */
  
  return TRUE;
}

/* ----------------------------------------------------------------------
 * icInc (new, more detailed heuristics, ns with or without fr, cn updated)
 */
void icInc(Lattice lattice, IcParams icParams, int no, char **args)
{
  Vector timelimits = icBuildTimeLimits(lattice->arcs);
  int size = vectorSize(timelimits);
  int node = 0;
  List list = NULL;
  ConstraintNet net;

  net = icInitializeNet(lattice, icParams);

  /* this is the iterative process */
  for (node = 1; node < size; node++) {
    cdgPrintf(CDG_INFO, "\n");
    list = icBuildIncomingArcs(lattice, (int)vectorElement(timelimits, node));

    if (!icAddArcsToNet(net, list, icParams)) {
      break;
    }
    if (!icSearchNet(net, icParams, (node+1==size), no, args)) {
      break;
    }
    icPrintSolutions(net, icParams);

    listDelete(list);
  }

  icPrintSolutionsFinal(net);
  
  /* free the memory */
  cnDelete(net);
  vectorDelete(timelimits);
}

/* ----------------------------------------------------------------------
 * icIncInteractive (new heuristics + cn update + interactive)
 */
void icIncInteractive(IcParams icParams, int no, char **args)
{
   ConstraintNet net = icInitializeNet(NULL, icParams);

#if 1
   char word[1024];
   Boolean cond = TRUE; 
   while (cond) {
    cdgExecHook(HOOK_ICINTERACTION, word);
    
    /* end of analysis */
    if (strcmp(word, "") == 0) {
      cond = FALSE;
      
    /* delete one word */
    } else if (strcmp(word, "$#$") == 0) {
      if (!icDeleteWord(net))
        cond = FALSE;
      
    /* add one word */
    } else {
      if (!icAddWord(net, word, icParams, no, args))
        cond = FALSE;
    }
  }
#endif

#if 0
  icAddWord(net, "an", icParams, no, args);
  icAddWord(net, "dem", icParams, no, args);
  icAddWord(net, "Dienstag", icParams, no, args);
  icAddWord(net, "den", icParams, no, args);
  icAddWord(net, "sechste", icParams, no, args);
  icAddWord(net, "April", icParams, no, args);
  icAddWord(net, "habe", icParams, no, args);
  icAddWord(net, "ich", icParams, no, args);
  icAddWord(net, "noch", icParams, no, args);
  icAddWord(net, "einen", icParams, no, args);
  icAddWord(net, "Termin", icParams, no, args);
  icDeleteWord(net);
  //icDeleteWord(net);
  //icDeleteWord(net);
  //icDeleteWord(net);
  //icDeleteWord(net);
  //icAddWord(net, "an", icParams, no, args);
  //icAddWord(net, "dem", icParams, no, args);
  //icAddWord(net, "Dienstag", icParams, no, args);
  //icAddWord(net, "den", icParams, no, args);
  //icAddWord(net, "sechste", icParams, no, args);
  //icAddWord(net, "xyz", icParams, no, args);
#endif

  /* free the memory */
  cnDelete(net);
}

/* ----------------------------------------------------------------------
 * icIncrementalCompletion
 */
int icIncrementalCompletion(Lattice lattice, IcParams icParams, int no, char **args)
{
  /* put agenda size into arguments list */
  if (icParams->agendasize != NULL) {
    if (no == 0) {
      no++;
      args[0] = icParams->agendasize;
    } else {
      no++;
      args--;
      args[0] = icParams->agendasize;
    }
  }

  if (lattice == NULL) {
    icIncInteractive(icParams, no, args);
  } else {
    icInc(lattice, icParams, no, args);
  }

  return 0;
}

/* ----------------------------------------------------------------------
 * icSetParams
 */
IcParams icSetParams(String buffer)
{
  int no = 0;
  int size = 100;
  char **args = memMalloc(size);
  int i;
  char string[strlen(buffer)+1];
  char *bufptr = string;
  char *token;
  IcParams icParams = (IcParams) memMalloc(sizeof(IcParamsStruct));

  icParams->agendasize = NULL;
  icParams->graph_file = NULL;
  icParams->info_file  = NULL;

  icParams->heurclass1     = 0;
  icParams->heurclass1dist = 0;
  icParams->heurclass2     = 0;
  icParams->heurclass2dist = 0;
  icParams->heurclass3     = 0;
  icParams->heurclass3dist = 0;
  icParams->heurclass4     = 0;
  icParams->heurclass4dist = 0;
  icParams->heurclass5     = 0;
  icParams->heurclass5dist = 0;

  icParams->frob_time       = 0.0;
  icParams->frob_pressure   = 0.0;
  icParams->frob_agendasize = 3000;
  icParams->frob_beamwidth  = INT_MAX;
  icParams->frob_iter       = 2;

  icParams->hotspot_val = 0.0;
  icParams->hotspot     = FALSE;
  icParams->hotpath     = FALSE;

  icParams->method_fr = FALSE;
  icParams->method_ns = FALSE;

  strcpy(string, buffer); 
  token=strtok_r(string, " ", &bufptr);
  while (token && no < size) {
    args[no] = strRegister(token);
    no++;
    token = strtok_r(NULL, " ", &bufptr);
  }

  if (no == 0 || strcmp(args[0], "help") == 0) {
    cdgPrintf(CDG_ERROR, "ERROR: set icparams '[-a agenda_size] [-f graph_file info_file]\n");
    cdgPrintf(CDG_ERROR, "       [-1 m [n]] [-2 m [n]] [-3 m [n]] [-4 m [n]] [-5 m [n]]\n");
    cdgPrintf(CDG_ERROR, "       [-fi frob_iter] [-ft frob_time] [-fp frob_pressure]\n");
    cdgPrintf(CDG_ERROR, "       [-fb frob_beamwidth] [-fa frob_agendasize]\n");
    cdgPrintf(CDG_ERROR, "       [-hv hotspot_val] [-hs] [-hp] [-ns] [-fr]'\n");
    cdgPrintf(CDG_ERROR, "       you need to produce ONE string argument with everything enclosed\n");
    memFree(icParams);
    memFree(args);
    return NULL;
  }

  i = 0;
  while (i != no) {
    if (strcmp(args[i], "-f") == 0) {
      if (no > i+2) {
	icParams->graph_file = args[i+1];
	icParams->info_file  = args[i+2];
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -f argument(s)\n");
	memFree(args);
	memFree(icParams);
	return NULL;
      }
      i += 3;
    } else if (strcmp(args[i], "-a") == 0) {
      if (no > i+1) {
	icParams->agendasize = strRegister(args[i+1]);
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -a argument\n");
	memFree(args);
	memFree(icParams);
	return NULL;
      }
      i += 2;
    } else if (strcmp(args[i], "-1") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass1 = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -1 argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass1dist = atoi(args[i]);
	i++;
      }
    } else if (strcmp(args[i], "-2") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass2 = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -2 argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass2dist = atoi(args[i]);
	i++;
      }
    } else if (strcmp(args[i], "-3") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass3 = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -3 argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass3dist = atoi(args[i]);
	i++;
      }
    } else if (strcmp(args[i], "-4") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass4 = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -4 argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass4dist = atoi(args[i]);
	i++;
      }
    } else if (strcmp(args[i], "-5") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass5 = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -5 argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->heurclass5dist = atoi(args[i]);
	i++;
      }
    } else if (strcmp(args[i], "-fi") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->frob_iter = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -ft argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
    } else if (strcmp(args[i], "-ft") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->frob_time = atof(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -ft argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
    } else if (strcmp(args[i], "-fp") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->frob_pressure = atof(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -fp argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
    } else if (strcmp(args[i], "-fb") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->frob_beamwidth = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -fb argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
    } else if (strcmp(args[i], "-fa") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->frob_agendasize = atoi(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -fa argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
    } else if (strcmp(args[i], "-hv") == 0) {
      i++;
      if (no > i && strchr(args[i], '-') == NULL) {
	icParams->hotspot_val = atof(args[i]);
	i++;
      } else {
	cdgPrintf(CDG_ERROR, "ERROR: `set icparams': missing -hv argument\n");
	memFree(icParams);
	memFree(args);
	return NULL;
      }
    } else if (strcmp(args[i], "-hs") == 0) {
      i++;
      icParams->hotspot = TRUE;
    } else if (strcmp(args[i], "-hp") == 0) {
      i++;
      icParams->hotpath = TRUE;
    } else if (strcmp(args[i], "-fr") == 0) {
      i++;
      icParams->method_fr = TRUE;
    } else if (strcmp(args[i], "-ns") == 0) {
      i++;
      icParams->method_ns = TRUE;
    } else {
      cdgPrintf(CDG_ERROR, "ERROR: `set icparams': wrong %s argument\n", args[i]);
      i++;
      memFree(icParams);
      memFree(args);
      return NULL;
    }
  }

  memFree(args);
  return icParams;
}

/* ----------------------------------------------------------------------
 *
 */
Boolean icValidateIcParams(String name, String value, String *var)
{
  IcParams icParamsRead = NULL;

  if (strcmp(name, "icparams") != 0) {
    return FALSE;
  }
  
  icParamsRead = icSetParams(*var);
  if (icParamsRead != NULL) {
    memFree(icParams);
    icParams = icParamsRead;
    return TRUE;
  } else {
    return FALSE;
  }
}

/* ----------------------------------------------------------------------
 * icInitialize
 */
void icInitialize(void) {
  icParamsStr = strRegister("-a 10000 -ns -fr -1 5 -2 5 -3 5 -4 1 -5 1 -fi 2 -ft 1 -fp 0.3");
  icParams    = icSetParams(icParamsStr);

  setRegister("icparams", SET_STRING, &icParamsStr, NULL, icValidateIcParams, NULL, NULL);
}

/* ----------------------------------------------------------------------
 * icFinalize
 */
void icFinalize(void) {
  cdgFreeString(icParamsStr);
  memFree(icParams);
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
