/*
 * File: testing.c
 * Author: Ingo Schroeder
 *
 * $Id: testing.c,v 1.40 2004/03/31 14:14:53 foth Exp $
 */

/* ---------------------------------------------------------------------- */
#include <configure.h>

#include <sys/stat.h>
#include <sys/wait.h>
#include <stdio.h>
#include <limits.h>

#include "blah.h"
#include "input.h" /* for Input */
#include "parsing.h" /* for parseResult */
#include <string.h>
#include "hook.h" /* for cdgPrintf() */
#include "input.h" /* for data types */
#include "lexemgraph.h" /* for LexemGraph & LexemNode */
#include "levelvalue.h"
#include "eval.h"
#include "parse.h"
#include "tagger.h"
#include "chunker.h"
#include "write.h"
#include "command.h"
#include "eval.h"

/* ----------------------------------------------------------------------
   Undocumented command for testing
   1: determine base line for accuracy
*/
void tst1(int no, char **args)
{

  HashIterator hi;
  int all=0, base=0;
  char *level=no>=1 ? args[0] : NULL;
  char *label=no>=2 ? args[1] : "";
  char *use_pos=no>=3 ? args[2] : "yes";
  AnnoEntry ae;
  
  hi = hashIteratorNew(inputCurrentGrammar->annotations);
  while((ae = hashIteratorNextValue(hi))) {
    List m;
    for (m=ae->annos; m; m=listNext(m)) {
      List n;
      Annotation a=listElement(m);
      
      
      for (n=a->specs; n; n=listNext(n)) {
        Specification s=listElement(n);
        if (s->type!=STDep) { continue; }
        if (level && strcmp(level, s->kind)!=0) { continue; }
        all++;
        if (strcmp(label, s->name)!=0) { continue; }
        if (strcmp(use_pos, "yes")==0 && s->position>0) { continue; }
        base++;
      }
    }
  }
  hashIteratorDelete(hi);
  
  cdgPrintf(CDG_INFO, "TESTING: %d out of %d edges ", base, all);
  if (level) { cdgPrintf(CDG_INFO, "on level `%s' ", level); }
  cdgPrintf(CDG_INFO, "(%5.2f%%) in %d annotations are trivial.\n",
	    all>0 ? 100.0*(float)base/(float)all : 0.0,
	    hashSize(inputCurrentGrammar->annotations));
}

/* ----------------------------------------------------------------------
   Undocumented command for testing
   determine base line for accuracy
*/
void tst2(int no, char **args)
{
#if 0

  This used the "id" spec in the annotation to compare lexical entries, which
  is deprecated. Rewrite it to compare against Parses on a rainy day.
  
  List l;
  char *level=no>=1 ? args[0] : NULL;
  char *label=no>=2 ? args[1] : "";
  char *use_pos=no>=3 ? args[2] : "yes";
  double total_ratio=0.0;
  
  for (l=inputCurrentGrammar->annotations; l; l=listNext(l))
    {
      List m;
      AnnoEntry ae=listElement(l);
      Lattice lat = NULL; /* TODO this doesn't fix a possible coredump in lgNew
			   *      when there are no lattices
			   */
      LexemGraph lg;
      int i;
      double ratio=0.0;
      
      /* find word graph w for ae */
      for (m=inputCurrentGrammar->lattices; m; m=listNext(m))
	{
	  lat=listElement(m);
	  if (!strcmp(lat->id, ae->id)) { break; }
	  if (!strcmp(lat->annoId, ae->id)) { break; }
	}
      if (!m)
	{
	  cdgPrintf(CDG_INFO,
		    "Can't find wordgraph for annotation `%s', skipping\n",
		    ae->id);
	  continue;
	}
      lg=lgNew(lat, inputCurrentGrammar->lexicon);
      if (!lg)
	{
	  cdgPrintf(CDG_INFO,
		    "Can't create lexemgraph for wordgraph `%s', skipping\n",
		    lat->id);
	  continue;
	}

      /* run the simulation multiple times */
#define NO_ITER 100
      for (i=0; i<NO_ITER; i++)
	{
	  int j;
	  int all=0, base=0;
	  /* find random path p through lg */
	  for (j=lg->min; j<lg->max; /* NOT HERE */)
	    {
	      int count=0;
	      int choose;
	      int k;

	      for (k=0; k<vectorSize(lg->nodes); k++)
		{
		  LexemNode ln=vectorElement(lg->nodes, k);
		  if (ln->arc->from==j) { count++; }
		}
	      if (count<=0)
		{
		  cdgPrintf(CDG_INFO,
			    "Can't find lexem starting at %d in `%s', skipping\n",
			    j, lat->id);
		  j++;
		  continue;
		}
	      count--;
	      choose=(int)(drand48()*count+.5)+1;
	      for (k=0; k<vectorSize(lg->nodes); k++)
		{
		  LexemNode ln=vectorElement(lg->nodes, k);
		  if (ln->arc->from==j && --choose==0)
		    {
		      List n;
		      /* We've chosen ln as the next arc. */
		      j=ln->arc->to;

		      /* check for each arc a on path p whether a is in ae */
		      for (n=ae->annos; n; n=listNext(n))
			{
			  List o;
			  Annotation a=listElement(n);
			  char *id=NULL;
			  if (a->from!=ln->arc->from) { continue; }
			  if (a->to!=ln->arc->to) { continue; }
			  if (strcmp(a->word, ln->arc->word)) { continue; }
			  /* a is the annotation for ln */			  
			  for (o=a->specs; o; o=listNext(o))
			    {
			      Specification s=listElement(o);
			      if (s->type!=STTag) { continue; }
			      if (strcmp(s->kind, "id")!=0) { continue; }
			      id=s->name;
			    }
			  for (o=a->specs; o; o=listNext(o))
			    {
			      Specification s=listElement(o);
			      if (s->type!=STDep) { continue; }
			      if (level && strcmp(level, s->kind)!=0) { continue; }
			      all++;
			      if (strcmp(label, s->name)!=0) { continue; }
			      if (strcmp(use_pos, "yes")==0 && s->position>0) { continue; }
			      if (id && ln->lexem->description && strcmp(id, ln->lexem->description)!=0) { continue; }
			      base++;
			    }
			}		  
		      break;
		    }
		}
	    }
	  /*
	    cdgPrintf(CDG_INFO, "TESTING: all %4d base %4d `%s'\n",
		    all, base, lg->lattice->id);
	  */
	  ratio+=all>0 ? (double)base/(double)all : 0.0;
	  
	}
      cdgPrintf(CDG_INFO, "TESTING: ratio %10.4f average %8.5f%% `%s'\n",
		ratio, ratio/(double)NO_ITER*100.0, lg->lattice->id);
      lgDelete(lg);
      total_ratio+=ratio;
    }
  cdgPrintf(CDG_INFO, "\nTESTING: total ratio %10.4f average %8.5f%%",
		total_ratio, total_ratio*100.0/(NO_ITER*listSize(inputCurrentGrammar->annotations)));
  if (level)
    {
      cdgPrintf(CDG_INFO, " on level `%s'\n", level);
    } else {
      cdgPrintf(CDG_INFO, "\n", level);      
    }
#endif
}


/* undocumented: show where constraints apply */

/* forward declaration -- evalFormula not normally exported */
Boolean evalFormula(Formula f, LexemGraph lg, Vector context);

void tst5(int no, char **args) {

  List l, m, n;
  List parses = NULL;
  
  /* build decorated parses from all annotations */
  commandEval("anno2parse");

  for(l = m = hashListOfKeys(cdgParses); l != NULL; l = listNext(l)) {
    String id = listElement(l);
    Parse p = hashGet(cdgParses, id);
    if(!parseIsAbstract(p)) {
      parses = listPrependElement(parses, p);
    }
  }
  
  /* check each constraint */
  for(l = m = hashListOfKeys(inputCurrentGrammar->constraints);
      l != NULL; l = listNext(l)) {
    String id = listElement(l);
    Constraint c = hashGet(inputCurrentGrammar->constraints, id);
    int apply = 0, fail = 0;
    
    /* don't regard variable-strength constraints */
    if(c->penaltyTerm) {
      continue;
    }

    /* try unary constraints on each edge */
    if(listSize(c->vars) == 1) {
      for(n = parses; n != NULL; n = listNext(n)) {
        Parse p = listElement(n);
        int i;
        for(i = 0; i < vectorSize(p->LVs); i++) {
          LevelValue lv = vectorElement(p->LVs, i);
          if(!lv) continue;

          if(matchUnarySignature(lv, c->sig)) {
            Formula premise;
            Formula conclusion;

            /* instantiate */
            VarInfo vi = listElement(c->vars);
            vi->levelvalue = lv;
            evalCurrentConstraint = c;
            
            /* test */
            if(FTImplication == c->formula->type) {
              premise = c->formula->data.junction.arg1;
              conclusion = c->formula->data.junction.arg2;
            } else {
              premise = NULL;
              conclusion = c->formula;
            }
            if(!premise || evalFormula(premise, p->lg, p->LVs)) {
              apply++;
              if(!evalFormula(conclusion, p->lg, p->LVs)) {
                fail++;
              }
            }
          }
        }
      }
      cdgPrintf(CDG_INFO,
                "%s: applied %d, held %d, failed %d\n",
                c->id, apply, apply-fail, fail);
    }
    /* try binary constraints on each edge pair */
    else {
      for(n = parses; n != NULL; n = listNext(n)) {
        Parse p = listElement(n);
        int i;
        for(i = 0; i < vectorSize(p->LVs); i++) {
          int j;
          LevelValue lvi = vectorElement(p->LVs, i);
          if(!lvi) continue;
          for(j = 0; j < vectorSize(p->LVs); j++) {
            LevelValue lvj = vectorElement(p->LVs, j);
            int k;
            if(!lvj) continue;
            if(lvi == lvj) continue;
            for(k = 0; k < 2; k++) {
              LevelValue lva, lvb;
              if(k == 0) {
                lva = lvi;
                lvb = lvj;
              } else {
                lva = lvj;
                lvb = lvi;
              }
              
              if(matchBinarySignature(lva, lvb, c->sig)) {
                Formula premise;
                Formula conclusion;
            
                /* instantiate */
                VarInfo vi = listElement(c->vars);
                vi->levelvalue = lva;
                vi = listElement(listNext(c->vars));
                vi->levelvalue = lvb;
                evalCurrentConstraint = c;

                /* test */
                if(FTImplication == c->formula->type) {
                  premise = c->formula->data.junction.arg1;
                  conclusion = c->formula->data.junction.arg2;
                } else {
                  premise = NULL;
                  conclusion = c->formula;
                }

                if(!premise || evalFormula(premise, p->lg, p->LVs)) {
                  apply++;
                  if(!evalFormula(conclusion, p->lg, p->LVs)) {
                    fail++;
                  }
                }
              }
            }
          }
        }
      }

      cdgPrintf(CDG_INFO,
                "%s: applied %d, held %d, failed %d\n",
                c->id, apply, apply-fail, fail);
    }
  }
}


/* show what constraints are violated
   by breaking the sentence at specific points. */
void tst7(int no, char **args)
{
  AnnoEntry ae = NULL;
  Parse p;
  List l;
  int i = 0;
  List boundaries = NULL;
  LexemGraph lg;
  Lattice lat = NULL;

  if (no < 1) {
    cdgPrintf(CDG_INFO, " usage: testing 7 <annoid>\n");
    return;
  }

  lat = findLattice(strRegister(args[0]));
  ae = findAnnoForLattice(lat, TRUE);
  if (!ae || !lat) {
    cdgPrintf(CDG_ERROR, "ERROR: no annotation `%d'.\n", args[0]);
    return;
  }

  lg = lgNew(lat);
  p = parseFromAnno(ae);
  if (!p || !parseDecorate(p, lg, ae)) {
    cdgPrintf(CDG_INFO, "ERROR: could not decorate this parse.\n");
    lgDelete(lg);
    return;
  }
  parseEval(p);
  
  i = 1;
  while (i < no) {
    int j;

    sscanf(args[i], "%d", &j);
    boundaries = listAppendElement(boundaries, (Pointer) j);
    i++;
  }

  /* direct all edges to NIL if they cross boundaries */
  for (i = 0; i < vectorSize(p->LVs); i++) {
    LevelValue lv = vectorElement(p->LVs, i);
    int index = lvIndex(lv);
    String label;

    if (!lv->modifiee) {
      continue;
    }
    if (lv->level == inputGetMainlevel(inputCurrentGrammar)) {
      label = strRegister("ROOT");
    } else {
      label = strRegister("");
    }
    for (l = boundaries; l; l = listNext(l)) {
      int j = (int)listElement(l);

      if ((lv->modifiee->arc->from > j && lv->modifier->arc->from <= j)
	  || (lv->modifier->arc->from > j && lv->modifiee->arc->from <= j)) {
	parseShiftEdge(p, index, -1);
	parseSwapLabel(p, index, label);
      }
      break;
    }
  }
  for (l = p->violations; l != NULL; l = listNext(l)) {
    ConstraintViolation cv = listElement(l);
    if (cv->penalty < 0.2) {
      cvPrint(CDG_INFO, cv);
      cdgPrintf(CDG_INFO, "\n");
    }
  }

}
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
