/* Copyright (C) 1997-2002 The BLAH Team <blah@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * The functions tested in TreeAgenda Module
 *
 * taNew
 * taSize
 * taMaxSize
 * taMaxSizeSoFar
 * taIsEmpty
 * taIsTruncated
 * taInsert
 * taDelete
 * taBest
 * taRemoveBest
 * taIteratorNew
 * taIteratorNextElement
 * taIteratorDelete
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <blah.h>
#include <assert.h>

int main(int argc, char **argv)
{
  char *a[] =
    { "Mike", "Killian", "Ajay", "Sofiya", "Aarthi", "Krish", "Adarsh",
      "Sebastian", "Louis", "Kathrine" };
  TreeAgenda ta;
  int maxs, i;
  Boolean ins, trnc;
  TreeAgendaIterator ti;
  String valstr, getBestStr, remBestStr;

  void func(Pointer p) {
    free(p);
  }

  blahInitialize();

  /* 
   * Creating a new treeagenda,taSize
   */

  ta = taNew(10, func);

  /* Test for taSize, taMaxSize */

  assert(taSize(ta) == 0);
  maxs = taMaxSize(ta);
  assert(maxs == 10);

  /* Test taInsert, taIsEmpty, Inserting 9 elements ( 0 to 8 ) */

  for (i = 0; i < maxs - 1; i++) {     /* Scores are same as their position
				        * in a[ ]) */
    ins = taInsert(ta, i, a[i]);
  }
  assert(ins == FALSE);
  assert(taIsEmpty(ta) == FALSE);

  /* Test for taMaxSizeSoFar */

  assert(taMaxSizeSoFar(ta) == 9);
  assert(taMaxSizeSoFar(ta) == taMaxSize(ta) - 1);

  /* Testing taIsTruncated */
  for (i = 9; i < 10; i++) {
    ins = taInsert(ta, i, a[i]);
  }
  trnc = taIsTruncated(ta);
  assert(trnc == FALSE);

  /* Testing for taIteratorNextElement */

  ti = taIteratorNew(ta);

  for (i = 9; i > -1; i--) {	       /* first retrieved is with max score */
    valstr = taIteratorNextElement(ti);
    assert(valstr == a[i]);
  }

  /* Test for taBest,taRemoveBest */

  for (i = 9; i > -1; i--) {
    getBestStr = (String) taBest(ta);	/* Best is the last in a [ ] */
    assert(getBestStr == a[i]);

    remBestStr = (String) taRemoveBest(ta);
    assert(remBestStr == a[i]);
  }

  /* taIteratorDelete */

  taIteratorDelete(ti);

  blahFinalize();

  return 0;
}
