eval '(exit $?0)' && eval 'exec perl -w -S $0 ${1+"$@"}'
 & eval 'exec perl -w -S $0 $argv:q'
 if 0;

# Horia F. Pop
#
# Purpose: genetic algorithm wrapper for cdgp, to find best constraint weights

# Files needed to run this script:
# o cdgp and stellingen.cdg
# o ga-cdg.pl and schedule.pl
# o UTTERANCES, VALUES and WEIGHTS (for ga-cdg.pl)
# o HOSTS (for schedule.pl)
# o JOBS (for schedule.pl) will be created by ga-cdg.pl

use strict;

require "getopts.pl";

&Getopts('a:b:c:dfh:k:m:q:rvz:xy:');
# -a 0.02  probability of absolute mutation
# -b 0.02  probability of relative mutation
# -c 0.5   probability of cross-over
# -d       also mutate those individuals that took part in a cross-over
# -f       force manual hard constraints to stay zero during reload
# -h 1.0   probability that hard weights stay hard under mutation (1.0: don't mutate hard constraints)
# -k 1.0   probability that under random initial assignments zero constraints stay zero
# -m 0.5   probability of multi-point cross-over under cross-over
# -q 1.01  quality threshold
# -r       random initial assignments of weights (manual grammar NOT part of the population)
# -v       be verbose
# -z 0.02  probability of mutation to zero
# -x       debug mode
# -y 30    force reinitialization at every Y generations (i.e. right before generation Y, 2Y, ...) 

$::opt_a=0.002 unless defined($::opt_a);
$::opt_b=0.02 unless defined($::opt_b);
$::opt_c=0.5 unless defined($::opt_c);
$::opt_d=defined($::opt_d) ? 1 : 0;
$::opt_f=defined($::opt_f) ? 1 : 0;
$::opt_h=1.0 unless defined($::opt_h);
$::opt_k=1.0 unless defined($::opt_k);
$::opt_m=0.5 unless defined($::opt_m);
$::opt_q=1.01 unless defined($::opt_q);
$::opt_r=defined($::opt_r) ? 1 : 0;
$::opt_v=defined($::opt_v) ? 1 : 0;
$::opt_z=0.0 unless defined($::opt_z);
$::opt_x=defined($::opt_x) ? 1 : 0;
$::opt_y=30 unless defined($::opt_y);

my $pwd = `pwd`;
chop($pwd);

open(LOG, ">GA.LOG");
my $oldh = select(LOG);
$| = 1;

printf "a=%4.3f b=%4.3f c=%4.3f d=%d f=%d h=%4.3f k=%4.3f m=%4.3f q=%4.3f r=%d v=%d x=%d y=%d z=%4.3f\n",
  $::opt_a, $::opt_b, $::opt_c, $::opt_d, $::opt_f, $::opt_h, $::opt_k,
  $::opt_m, $::opt_q, $::opt_r, $::opt_v, $::opt_x, $::opt_y, $::opt_z
  if $::opt_v;

# makes sure that there is a subdirectory named 'data'
# all the data directories are compressed and moved there
system("mkdir -p data");

Main();

close(LOG);
select($oldh);

#-----------------------------------------------------------------------------
# main function; runs everything
#
sub Main {
print "in  Main\n" if ($::opt_v);

  my @pop;			# population array
  my @used;			# used[i] = 1 if used by crossover
  my @fit;			# fitness array
  my @utterances = ReadFromFile("UTTERANCES"); # lexem graph ids
  my @weights = ReadFromFile("WEIGHTS"); # constraint names
  my @values = ReadFromFile("VALUES");  # manual constraint weights

  my $g  = 100;			# number of generations
  my $n  = 30;			# population size

  my $s  = $#{@weights} + 1;    # number of weights
  my $pc = $::opt_c;            # probability of crossover
  my $pa = 1-$::opt_m;          # probability of arithmetic crossover under cross over
  # Note that $prm+$pam+$pz < 1.0 !
  my $prm = $::opt_b;           # probability of relative mutation
  my $pam = $::opt_a;           # probability of absolute mutation
  my $pz = $::opt_z;            # probability of mutation to zero
  my $q  = $::opt_q;            # quality threshold

  my $y  = $::opt_y;            # reinitialization interval
  
  my $time_per_utterance = 60;	# seconds
  my $memory_per_job = 128;	# MB
  my $utterances_per_job = 10;

  srand( time() ^ ($$ + ($$ << 15)) );

  my $iter = 0;
  if ( -e ".ga-cdg-rc" ) {
    # if we have a recovery file, load it
    ReadConfig(\@pop, \@fit, \@used, $n, $s, \$iter);
    print "Restoring population saved at generation $iter\n\n";
    for my $j ( 0 .. $s-1 ) {
      # printf "-- %5.4f %5.4f\n", ${$pop[0]}[$j], $values[$j] if $::opt_v;
      if ($::opt_f && $values[$j] == 0.0) {
	printf "  Forcing %d to zero\n", $j if $::opt_v;
        for my $i ( 0 .. $n-1 ) {
	  ${$pop[$i]}[$j] = 0.0;
	}
      }
    }

    WriteData(\@pop, \@fit, $s, "");
  } else {
    # we are at the beginning of a fresh process
    # initialize population
    InitData(\@pop, \@fit, \@used, \@values, $n, $s);

    print "Generation $iter\n\n";

    # determine the fitness of the population
    if ($::opt_x) {
      FitnessDebug(\@pop, \@fit, $s);
    } else {
      Fitness(\@pop, \@fit, $s, \@utterances, \@weights,
	      $time_per_utterance, $memory_per_job, $utterances_per_job, $iter);
    }
    WriteData(\@pop, \@fit, $s, "After fitness");

    # select the population for the genetic operators
    if ($iter % $y == ($y-1)) {
      SelectPopInit(\@pop, \@fit, \@used, $n, $s);
    } else {
      SelectPop(\@pop, \@fit, \@used, $n, $s);
    }
    WriteData(\@pop, \@fit, $s, "After select");

    # save recovery information
    SaveConfig(\@pop, \@fit, $n, $s, $iter);
  }

  while (Terminate(\@fit, $iter, $g, $q) == 0) {
    # run the genetic operators
    Cross (\@pop, \@fit, \@used, $pc, $pa, $s);
    Mutate(\@pop, \@fit, \@used, $prm, $pam, $pz, $s);
    print "\n";

    $iter++;
    print "Generation $iter\n\n";

    # determine the fitness of the population
    if ($::opt_x) {
      FitnessDebug(\@pop, \@fit, $s);
    } else {
      Fitness(\@pop, \@fit, $s, \@utterances, \@weights,
	      $time_per_utterance, $memory_per_job, $utterances_per_job, $iter);
    }
    WriteData(\@pop, \@fit, $s, "After fitness");

    # select the population for the genetic operators
    if ($iter % $y == ($y-1)) {
      SelectPopInit(\@pop, \@fit, \@used, $n, $s);
    } else {
      SelectPop(\@pop, \@fit, \@used, $n, $s);
    }
    WriteData(\@pop, \@fit, $s, "After select");

    # save recovery information
    SaveConfig(\@pop, \@fit, $n, $s, $iter);
  }

print "out Main\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# save population information for recovery purposes
#
# o @$a   = population array
# o @$f   = fitness array
# o $n    = population size
# o $s    = number of genes
# o $iter = iteration number
#
sub SaveConfig {
print "in  SaveConfig\n" if ($::opt_v);

  my ($a, $f, $n, $s, $iter) = @_;

  if ( -e ".ga-cdg-rc" ) {
    system ("mv .ga-cdg-rc .ga-cdg-rc~");
  }

  open (FILE, ">.ga-cdg-rc");
  print FILE "$iter\n";
  for my $i ( 0 .. $n-1 ) {
    print FILE "@$f[$i] ";
    for my $j ( 0 .. $s-1 ) {
      print FILE "${@{$a}}[$i][$j] ";
    }
    print FILE "\n";
  }
  close (FILE);

print "out SaveConfig\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# reading population information for recovery purposes
#
# o @$a    = population array
# o @$f    = fitness array
# o @$u    = array to mark individuals for crossover
# o $n     = population size
# o $s     = number of genes
# o $$iter = iteration number
#
sub ReadConfig {
print "in  ReadConfig\n" if ($::opt_v);

  my ($a, $f, $u, $n, $s, $iter) = @_;

  open (FILE, "<.ga-cdg-rc");

  my $line = <FILE>;
  chomp($line);
  my @line = split(/ /, $line);
  $$iter = $line[0];

  for my $i ( 0 .. $n-1 ) {
    $line = <FILE>;
    chomp($line);
    @line = split(' ', $line);

    @$f[$i] = $line[0];
    for my $j ( 0 .. $s-1 ) {
      ${@{$a}}[$i][$j] = $line[$j+1];
    }
    @$u[$i] = 0;
  }
  close (FILE);

print "out ReadConfig\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# initialize the population -- random
#
# o @$a = population array
# o @$f = fitness array
# o @$u = array to mark individuals for crossover
# o @$v = array with weight values of our current grammar
# o $n  = population size
# o $s  = number of genes
#
sub InitDataRandom {
print "in  InitDataRandom\n" if ($::opt_v);

  my ($a, $f, $u, $v, $n, $s) = @_;

  printf "== Starting random assigment with opt_k==%4.3f\n", $::opt_k;
  # random initial population
  for my $i ( 0 .. $n-1 ) {
    for my $j ( 0 .. $s-1 ) {
      my $val = @$v[$j];

      if ($val == 0 && rand()<=$::opt_k) {
	# manual zero means zero here
	${@{$a}}[$i][$j]=0;
      } else {
        ${@{$a}}[$i][$j]=rand();
      }
    }
  }

  for my $i ( 0 .. $n-1 ) {
    @$u[$i] = 0;
  }
  for my $i ( 0 .. $n-1 ) {
    @$f[$i] = -1;
  }

print "out InitDataRandom\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# initialize the population -- manual
#
# o @$a = population array
# o @$f = fitness array
# o @$u = array to mark individuals for crossover
# o @$v = array with weight values of our current grammar
# o $n_ini = individual to start
# o $n_end = individual to end
# o $s  = number of genes
#
sub InitDataManual {
print "in  InitDataManual\n" if ($::opt_v);

  my ($a, $f, $u, $v, $n_ini, $n_end, $s) = @_;

  # a[0] is the original grammar
  for my $j ( 0 .. $s-1 ) {
    my $val = @$v[$j];
    ${@{$a}}[$n_ini][$j] = $val;
  }

  # a[i] is initialized as usual
  for my $i ( $n_ini+1 .. $n_end ) {
    for my $j ( 0 .. $s-1 ) {
      my $val = @$v[$j];
      my $r = rand();
      if ($r <= 0.1) {
        # decrease the original weights
	${@{$a}}[$i][$j] = $val - $val * rand();
      } else {
        if ($r <= 0.2) {
	  # increase the original weights
          ${@{$a}}[$i][$j] = $val + (1 - $val) * rand();
        } else {
          # let the weights unchanged
          ${@{$a}}[$i][$j] = $val;
        }
      }
    }
  }

  for my $i ( $n_ini .. $n_end ) {
    @$u[$i] = 0;
  }
  for my $i ( $n_ini .. $n_end ) {
    @$f[$i] = -1;
  }

print "out InitDataManual\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# initialize the population
#
# o @$a = population array
# o @$f = fitness array
# o @$u = array to mark individuals for crossover
# o @$v = array with weight values of our current grammar
# o $n  = population size
# o $s  = number of genes
#
sub InitData {
print "in  InitData\n" if ($::opt_v);

  my ($a, $f, $u, $v, $n, $s) = @_;

  if ($::opt_r) {
    InitDataRandom(\@$a, \@$f, \@$u, \@$v, $n, $s);
  } else {
    InitDataManual(\@$a, \@$f, \@$u, \@$v, 0, $n-1, $s);
  }

print "out InitData\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# random assignment of population fitness -- only for debug
#
# o @$a = population array
# o @$f = fitness array; compute only for f[i] == -1
# o $s  = number of genes
#
sub FitnessDebug {
print "in  FitnessDebug\n" if ($::opt_v);

  my ($a, $f, $s) = @_;

  for my $i ( 0 .. $#{@$a} ) {
    if (@$f[$i] == -1) {
      @$f[$i] = rand();
    }
  }

print "out FitnessDebug\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# compute the population fitness
#
# o @$a = population array
# o @$f = fitness array; compute only for f[i] == -1
# o $s  = number of genes
# o @$utterances = array with names of utterances
# o @$weights    = array with names of weights
# o $t    = time per utterance, in seconds
# o $m    = memory per job, in MB
# o $nru  = number of utterances per job
# o $iter = index of current generation
#
sub Fitness {
print "in  Fitness\n" if ($::opt_v);

  my ($a, $f, $s, $utterances, $weights, $t, $m, $nru, $iter) = @_;

  PrepareFitness(\@$a, \@$f, $s, \@$utterances, \@$weights, $t, $m, $nru, $iter);

  for my $i ( 0 .. $#{@$a} ) {
    if (@$f[$i] == -1) {
      @$f[$i] = GetFitness($i, $iter, $#{@$utterances} + 1, $t);
    }
  }

print "out Fitness\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# prepare the scripts and run the scheduler
#
# o @$a = population array
# o @$f = fitness array
# o $s  = number of genes
# o @$utterances = array with names of utterances
# o @$weights    = array with names of weights
# o $t    = time per utterance, in seconds
# o $m    = memory per job, in MB
# o $nru  = number of utterances per job
# o $iter = index of current generation
#
sub PrepareFitness {
print "in  PrepareFitness\n" if ($::opt_v);

  my ($a, $f, $s, $utterances, $weights, $t, $m, $nru, $iter) = @_;

  open(JOBS, ">JOBS");
  for my $ind ( 0 .. $#{@$a} ) {
    if (@$f[$ind] == -1) {

      # write .cdgrc.iter.ind.ii files
      for my $ii ( 0 .. $#{@$utterances} / $nru ) {
	open(CDG, ">.cdgrc.$iter.$ind.$ii");
	printf CDG "set timelimit %d\n", $t * 1000;
	print CDG "set debug off\n";
	print CDG "set profile on\n";
	print CDG "set cache off\n";
	print CDG "load $pwd/stellingen.cdg\n";
	for my $j ( 0 .. $#{@$weights} ) {
	  print CDG "weight @$weights[$j] ${@{$a}}[$ind][$j]\n";
	}
	my $number = min(($ii+1)*$nru-1, $#{@$utterances});
	for my $i ( $ii*$nru .. $number ) {
	  my $iu = $i % $nru;
	  print CDG "newnet @$utterances[$i]\n";
	  print CDG "frobbing net$iu combined 0.3 batch\n";
	  print CDG "compareparses @$utterances[$i]\n";
	  print CDG "netdelete net$iu\n";
	}
	print CDG "quit\n";
	close(CDG);

	# write sh.iter.ind.ii files
	open(SH, ">sh.$iter.$ind.$ii");
	print SH "#! /bin/bash\n";
	printf SH "nice -19 time -p $pwd/cdgp -i $pwd/.cdgrc.$iter.$ind.$ii -c %d -b %d\n", 
	  $t * ($number % $nru + 1) * 1.5, 
	    $m * 1024 * 1024;
	print SH "touch FINISHED\n";
	close(SH);

	system("chmod u+x sh.$iter.$ind.$ii");

	# write the JOBS file
	print JOBS "$pwd/sh.$iter.$ind.$ii\n";
      }
    }
  }

  close(JOBS);

  # run the scheduler
  system("./scheduler.pl -j JOBS -l HOSTS > SCHEDULER.LOG");

  for my $ind ( 0 .. $#{@$a} ) {
    if (@$f[$ind] == -1) {
      system("rm -f .cdgrc.$iter.$ind.*");
      system("rm -f sh.$iter.$ind.*");
    }
  }

print "out PrepareFitness\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# compute the fitness of the given individual
#
# o $ind             = index of current individual
# o $iter            = index of current generation
# o $totalutterances = total number of utterances
# o $t               = time per utterance, in seconds
#
sub GetFitness {
print "in  GetFitness\n" if ($::opt_v);

  my ($ind, $iter, $totalutterances, $t) = @_;

  # collect the results
  my $recall   = 0.0;
  my $accuracy = 0.0;
  my $time     = 0.0;
  my $usertime = 0.0;
  my $systime  = 0.0;
  my $frobtime = 0.0;

  open (S, "<SCHEDULER.LOG");
  while (<S>) {
    if ( m/^Finished\s+(.*)\s+(.*)\s+in\s+(.*)$/ ) {
      my $dir = $3;
      my $script = $2;

      if ( $script =~ /sh.$iter.$ind\./ ) {

	open(FILE, "<$dir/STDOUT");
	while (<FILE>) {
	  # new cdgp log format
	  if ( m/^INFO: Total\s+Accuracy\s+=\s+([\d\.]+)\s+Recall\s+=\s+([\d\.]+).*$/ ) {
	    $accuracy += $1;
	    $recall   += $2;
	  }

	  # old cdgp log format
	  # if ( m/^INFO: Recall\s+=\s+([\d\.]+)\s*$/ ) {
	  #   $recall += $1;
	  # }
	  # if ( m/^INFO: Accuracy\s+=\s+([\d\.]+)\s*$/ ) {
	  #   $accuracy += $1;
	  # }

	  # frobbing time
	  if ( m/^PROFILE:\s+([\d\.]+)s spent frobbing.*$/ ) {
	    $frobtime += min($1, $t);
	  }
	}
	close(FILE);

	# user + system time
	open(FILE, "<$dir/STDERR");
	while (<FILE>) {
	  if ( m/^user (.*)[,.](.*)$/ ) {
	    $usertime += $1 + $2/100;
	  }
	  if ( m/^sys (.*)[,.](.*)$/ ) {
	    $systime += $1 + $2/100;
	  }
	}
	close(FILE);

	system("zip -r -m data/$iter.$ind.zip $dir");
      }
    }
  }
  close(S);

  $recall   /= $totalutterances;
  $accuracy /= $totalutterances;
  # $time      = 1 - ($usertime + $systime) / ($t * $totalutterances * 1.5);
  $time      = 1 - $frobtime / ($t * $totalutterances);

  my $recall_weight = 0.5;	# recall weight inside f-measure
  my $time_weight   = 0.2;	# time weight inside fitness

  my $f_measure = 0.0;
  # this test is to avoid division-by-zero errors
  if ($recall * $accuracy != 0) {
    $f_measure = 1 / ($recall_weight / $recall + (1 - $recall_weight) / $accuracy);
  }

print "out GetFitness\n" if ($::opt_v);

  return $time_weight * $time + (1 - $time_weight) * $f_measure;
}

#-----------------------------------------------------------------------------
# test the termination criteria:
#
# o the fitness of an individual is larger than the threshold
# o the given number of generations has been reached
#
# o @$f   = fitness array
# o $iter = number of current generation
# o $g    = total number of allowed generations
# o $q    = fitness threshold
#
sub Terminate {
print "in  Terminate\n" if ($::opt_v);

  my ($f, $iter, $g, $q) = @_;

  if ($iter >= $g-1) {
    print "\nTerminate: maxiter >= $g - 1\n";
    return 1;
  } else {
    my $maxfit = 0;
    my $maxwho = -1;
    for my $i ( 0 .. $#{@$f} ) {
      if ($maxfit < @$f[$i]) {
	$maxfit = @$f[$i];
	$maxwho = $i;
      }
    }
    if ($maxfit > $q) {
      print "\n";
      print "Terminate: maxfit ($maxwho) = $maxfit\n";
      return 1;
    }
  }

print "out Terminate\n" if ($::opt_v);

  return 0;
}

#-----------------------------------------------------------------------------
# select the population for the next generation; reinitialize with three best
#
# o @$a = population array
# o @$f = fitness array
# o @$u = array to mark the individuals used for crossover, so that they are
#         ignored at mutation
# o $n  = number of individuals in the new population
# o $s  = number of genes (to avoid an extra computation)
#
sub SelectPopInit {
print "in  SelectPopInit\n" if ($::opt_v);

  my ($a, $f, $u, $n, $s) = @_;

  my @newa;
  my @newf;
  my @newu;
  my @sort;
  my @besta0;
  my @besta1;
  my @besta2;
  my @bestf;
  my $n1 = int($n/3);
  my $n2 = int($n*2/3);

  # sort the fitness values
  for my $i ( 0 .. $#{@$a} ) {
    $sort[$i] = $i;
  }
  for my $i (0 .. $#{@$a}-1 ) {
    for my $j ($i+1 .. $#{@$a} ) {
      if (@$f[$sort[$i]] < @$f[$sort[$j]]) {
	my $tmp = $sort[$i];
	$sort[$i] = $sort[$j];
	$sort[$j] = $tmp;
      }
    }
  }

  for my $i ( 0 .. $s-1 ) {
    $besta0[$i] = ${@{$a}}[$sort[0]][$i];
    $besta1[$i] = ${@{$a}}[$sort[1]][$i];
    $besta2[$i] = ${@{$a}}[$sort[2]][$i];
  }
  $bestf[0] = @$f[$sort[0]];
  $bestf[1] = @$f[$sort[1]];
  $bestf[2] = @$f[$sort[2]];

  InitDataManual(\@newa, \@newf, \@newu, \@besta0,   0, $n1-1, $s);
  InitDataManual(\@newa, \@newf, \@newu, \@besta1, $n1, $n2-1, $s);
  InitDataManual(\@newa, \@newf, \@newu, \@besta2, $n2, $n-1,  $s);
  $newf[0]   = $bestf[0];
  $newf[$n1] = $bestf[1];
  $newf[$n2] = $bestf[2];

  @$a = @newa;
  @$f = @newf;
  @$u = @newu;

print "out SelectPopInit\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# select the population for the next generation
#
# o set the fitness of new individuals to -1, so that these individuals
#   do not get their fitness computed again
#
# o @$a = population array
# o @$f = fitness array
# o @$u = array to mark the individuals used for crossover, so that they are
#         ignored at mutation
# o $n  = number of individuals in the new population
# o $s  = number of genes (to avoid an extra computation)
#
sub SelectPop {
print "in  SelectPop\n" if ($::opt_v);

  my ($a, $f, $u, $n, $s) = @_;

  my @newa;
  my @newf;
  my @newu;
  my @sort;
  my $ftot = 0;
  my $fpar = 0;
  my @p;
  my $first = 5;
  my $last = 5;

  # sort the fitness values
  for my $i ( 0 .. $#{@$a} ) {
    $sort[$i] = $i;
  }
  for my $i (0 .. $#{@$a}-1 ) {
    for my $j ($i+1 .. $#{@$a} ) {
      if (@$f[$sort[$i]] < @$f[$sort[$j]]) {
	my $tmp = $sort[$i];
	$sort[$i] = $sort[$j];
	$sort[$j] = $tmp;
      }
    }
  }

  # compute the total of fitness values
  for my $i ( $first .. $#{@$a}-$last ) {
    $ftot += @$f[$sort[$i]];
  }

  # this cannot normally happen
  if ($ftot == 0) {
    # avoid division by zero error
    $ftot = 1;
  }

  # compute the selection probabilities
  $p[$first] = 0;
  for my $i ( $first .. $#{@$a}-$last ) {
    $fpar += @$f[$sort[$i]];
     $p[$i+1] = $fpar / $ftot;
  }
  # rand() may give you a zero but no one.
  # Actual bug raised its head!
  # just to be sure...
  $p[$#{@$a}-$last+1]=1;

  # select the first individuals
  for my $i ( 0 .. $first-1 ) {
    my @tmp;
    for my $k ( 0 .. $s-1 ) {
      $tmp[$k] = ${@{$a}}[$sort[$i]][$k];
    }
    push(@newa, [@tmp]);
    push(@newf, @$f[$sort[$i]]);
    push(@newu, 0);
  }

  # apply roulette wheel for the other individuals
  for my $j ( $first .. $n-1 ) {
    my $r = rand();
    my @tmp;
    for my $i ( $first .. $#{@$a}-$last ) {
      # rand() may give you a zero but no one.
      # Actual bug raised its head!
      if ($p[$i] <= $r and $r < $p[$i+1]) {
	for my $k ( 0 .. $s-1 ) {
	  $tmp[$k] = ${@{$a}}[$sort[$i]][$k];
        }
        push(@newa, [@tmp]);
        push(@newf, @$f[$sort[$i]]);
        push(@newu, 0);
        last;
      }
    }
  }

  @$a = @newa;
  @$f = @newf;
  @$u = @newu;

print "out SelectPop\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# apply crossover operators
#
# o @$a = population array
# o @$f = fitness array
# o @$u = array to mark the individuals used for crossover
# o $pc = probability of crossover
# o $pa = probability of arithmetic crossover inside crossover itself
# o $s  = number of genes
#
sub Cross {
print "in  Cross\n" if ($::opt_v);

  my ($a, $f, $u, $pc, $pa, $s) = @_;

  my $selected = -1;
  my $total = $#{@$a};

  for my $i ( 0 .. $total ) {
    if (rand() < $pc) {
      if ($selected == -1) {
	$selected = $i;
      } else {
	if (rand() < $pa) {
	  ArithmeticCross(\@$a, \@$u, $selected, $i, $s);
	} else {
	  MultipointCross(\@$a, \@$u, $selected, $i, $s);
	}
	push(@$f, (-1, -1));
	${@{$u}}[$selected] = 1;
	${@{$u}}[$i] = 1;
        $selected = -1;
      }
    }
  }
  if ($selected != -1) {
    if (rand() < 0.5) {
      my $newselected = $selected + 1 + int(rand($total - $selected - 1));
      if (rand() < $pa) {
        ArithmeticCross(\@$a, \@$u, $selected, $newselected, $s);
      } else {
	MultipointCross(\@$a, \@$u, $selected, $newselected, $s);
      }
      push(@$f, (-1, -1));
      ${@{$u}}[$selected] = 1;
      ${@{$u}}[$newselected] = 1;
    }
  }

print "out Cross\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# apply arithmetic crossover
#
# o @$a = population array
# o @$u = array to mark individuals used for crossover
# o $k1, $k2 = indices of individuals selected for crossover
# o $s  = number of genes
#
sub ArithmeticCross {
print "in  ArithmeticCross\n" if ($::opt_v);

  my ($a, $u, $k1, $k2, $s) = @_;

  my @tmp;

  print "Arithmetic crossover $k1 $k2\n";
  for my $i ( 0 .. $s-1 ) {
    my $r = rand();
     $tmp[$i] = $r * ${$a}[$k1][$i] + (1 - $r) * ${$a}[$k2][$i];
  }

  push(@$a, [@tmp]);
  push(@$u, 0);

print "out ArithmeticCross\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# apply multipoint crossover
#
# o @$a = population array
# o @$u = array to mark individuals used for crossover
# o $k1, $k2 = indices of individuals selected for crossover
# o $s  = number of genes
#
sub MultipointCross {
print "in  MultipointCross\n" if ($::opt_v);

  my ($a, $u, $k1, $k2, $s) = @_;

  my $nrpos = int(rand($s-1));
  my @tmp1;
  my @tmp2;
  my @pos;
  my $how = 0;

  print "Multipoint crossover $k1 $k2\n";
  for my $i ( 0 .. $nrpos ) {
    $pos[$i] = int(rand($s-1));
  }
  push @pos, $s;
  @pos = sort(@pos);

  my $j = 0;
  for my $i ( 0 .. $s-1 ) {
    if ($i == $pos[$j]) {
      $j++;
      $how = 1 - $how;
    }
    if ($how == 0) {
        $tmp1[$i] = ${@{$a}}[$k1][$i];
      $tmp2[$i] = ${@{$a}}[$k2][$i];
    } else {
       $tmp1[$i] = ${@{$a}}[$k2][$i];
       $tmp2[$i] = ${@{$a}}[$k1][$i];
    }
  }

  push(@$a, [@tmp1]);
  push(@$a, [@tmp2]);
  push(@$u, 0);
  push(@$u, 0);

print "out MultipointCross\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# apply mutation operator
#
# o @$a = population array
# o @$f = fitness array
# o @$u = array to mark individuals used for crossover
# o $prm = probability of relative mutation
# o $pam = probability of absolute mutation
# o $pz = probability of mutation to zero
# o $s  = number of genes
#
sub Mutate {
print "in  Mutate\n" if ($::opt_v);

  my ($a, $f, $u, $prm, $pam, $pz, $s) = @_;
  print "Mutation ";

  my $la = $#{@$a};
  for my $i ( 0 .. $la ) {
    if ($::opt_d || ${@$u}[$i] == 0) {
      my @tmp;
      my $mut = 0;

      for my $j ( 0 .. $s-1 ) {
	my $r = rand();
	if (${@{$a}}[$i][$j]==0 && rand()<=$::opt_h) {
	  $tmp[$j] = ${@{$a}}[$i][$j];
	} elsif ($r < $pam) {
	  $mut = 1;
          $tmp[$j] = rand();
  	  print "$i($j)a ";
	} elsif ($r < $prm+$pam) {
	  $mut = 1;
	  my $x=${@{$a}}[$i][$j];
	  if (rand()<0.5) {
	     $tmp[$j] = $x - $x * rand();
	     $tmp[$j] = $tmp[$j]<0 ? 0 : $tmp[$j];
	    print "$i($j)- ";
	  } else {
	     $tmp[$j] = $x + (1-$x) * rand();
	     $tmp[$j] = $tmp[$j]>1 ? 1 : $tmp[$j];
	    print "$i($j)+ ";
	  }
        } elsif ($r < $prm+$pam+$pz) {
	  $mut = 1;
	  $tmp[$j] = 0;
	  print "$i($j)0 ";
        } else {
	  $tmp[$j] = ${@{$a}}[$i][$j];
        }
      }
      if ($mut == 1) {
	push(@$a, [@tmp]);
	push(@$f, -1);
	push(@$u, 0);
      }
    }
  }
  print "\n";

print "out Mutate\n" if ($::opt_v);
}

#-----------------------------------------------------------------------------
# write population information
#
# o @$a     = population array
# o @$f     = fitness array
# o $s      = number of genes
# o $string = string comment to display
#
sub WriteData {
print "in  WriteData\n" if ($::opt_v);

  my ($a, $f, $s, $string) = @_;

  if ($string ne "") {
    print "$string\n\n";
  }
  for my $i ( 0 .. $#{@$a} ) {
    printf "%2d : %7.5f = ", $i, @$f[$i];
    for my $j ( 0 .. $s-1 ) {
      printf "%7.5f ", ${@{$a}}[$i][$j];
    }
    print "\n";
  }
  print "\n";

print "out WriteData\n" if ($::opt_v);
}

# ----------------------------------------------------------------------------
# function to read data from a file
#
sub ReadFromFile {
  my ($name) = @_;
  my @list;

  open(H, "<$name") || die "can't read from file \"$name\"";
  while (<H>) {
    chop($_); push @list, $_;
  }
  close(H);
  
  return @list;
}

# ----------------------------------------------------------------------------
# min
#
sub min {
  my ($a, $b) = @_;

  if ($a < $b) {
    return $a;
  } else {
    return $b;
  }
}
