#!/usr/bin/env perl
# cdg/utils/splice-annotations.pl

# Transform CDG annotations so that two words become one.
#
# Which words are splices is determined by command line arguments;
# to splice successive arabic numbers, invoke me like this:
#
#  splice-annotations.pl '\d+' '\d+' foo.cdg
use warnings;
use strict;

BEGIN {
 if(!defined $ENV{CDGHOME}) {
   die "Please set CDGHOME.\n";
 }
 $ENV{CDGPERL} = "$ENV{CDGHOME}/bindings";
}
use lib "$ENV{CDGPERL}";
use lib "$ENV{CDGPERL}/.libs";
use CDG;
use CDG::Utils qw(:statistics :functions);

our ($p1,$p2);

# Transform one perl annotation
sub condense {
  my $a = shift;
  my $id = $a->{id};
  my $nrwords = $#{$$a{words}};
  my @splices;

  # find instances of words to be spliced
  for (my $i = 0; $i < $nrwords; $i++) {
    my $x = $$a{words}[$i];
    my $y = $$a{words}[$i+1];
    if ($$x{word} =~ $p1 && $$y{word} =~ $p2) {
      print "$$a{id}: $$x{word} $$y{word}\n";
      push @splices, $i;
    }
  }

  return unless @splices;

  # Apply each splice in turn, since there may be multiple, even
  # consecutive, splices in one sentence.
  my $progress = 0;
  for my $s (@splices) {
    $s -= $progress;

    # transform @words
    my $x = $$a{words}[$s];
    my $y = $$a{words}[$s+1];
    $$x{word} .= $$y{word};

    my @words;
    my $i;
    for(@{$$a{words}}) {
      next if $y == $_;
      $$_{no} = $i++;
      push @words, $_;
    }
    $$a{words} = [ @words ];

    # transform @edges
    my @edges = ();

    for my $e (@{$$a{edges}}) {
      next if $$e{child} == $s+1;
      if($$e{child} > $s) {
        $$e{child}--;
      }
      if($$e{parent} > $s) {
        $$e{parent}--;
      }

      push @edges, $e;
    }
    $$a{edges} = [ @edges ];

    $progress++;
  }

  write_anno $a;
}

die "usage: $0 <word pattern> <word pattern> <files>\n" unless @ARGV > 2;

my $x = shift; $p1 = qr /$x/;
$x = shift; $p2 = qr/$x/;

commandEval "load @ARGV";

# import annotations to perl for easier manipulation
our @annotations;
my $annotations = InputStruct_annotations_get($CDG::inputCurrentGrammar);
my $n = hashSize($annotations);
print "Importing $n annotations to perl...\n" if $n;
for (my $l = hashListOfKeys($annotations); $l; $l = listNext $l) {
  my $ae = hashGet($annotations, listElement($l));
  my $h = AnnoEntry2hash $ae;
  push @annotations, $h;
}

condense $_ for @annotations;


