# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaConfigItem - interface for configurable objects.
## This class has no implementation and only serves as a marker of derived
## configurable items in YadaConfigDocument.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaConfigItem.tcl,v 1.6 2004/03/25 13:58:24 ddreyer Exp $
## ----------------------------------------------------------------------------
class YadaConfigItem {
    protected variable _modified 0
  
    public variable name

    public method load {}
    public method save {}
    public method toDOM {}
    public method fromDOM {node}; # \type domNode
    public method hasSavedState {}
    public method isModified {}
    public method setModified {}
    public method resetModified {}
    public method displayTitle {}
    public method getPersistanceFileName {}; ## purevirtual
    
    
};



## ----------------------------------------------------------------------------
## Serializes object state to unique file (depends on object name)
## ----------------------------------------------------------------------------
body YadaConfigItem::save {} {
    set domNode [toDOM]	
    set fileNameToSave [getPersistanceFileName]

    if { [::file exists ${fileNameToSave}.gz] } {
	::file delete ${fileNameToSave}.gz
    }

    set fileID [open $fileNameToSave {CREAT WRONLY}]
    puts $fileID [$domNode asXML]
    close $fileID
 
    if {[catch {exec gzip -f $fileNameToSave} errMsg]} {
	error "Saving failed: $errMsg"
    }
    resetModified
}

## ----------------------------------------------------------------------------
## Loads object state from unique file (depends on object name)
## ----------------------------------------------------------------------------
body YadaConfigItem::load {} {
    set zippedFileName [getPersistanceFileName].gz

    if { ![::file exists $zippedFileName]} {   
	error "File $zippedFileName not found."
    }
 
    set fileID [open "|zcat $zippedFileName" {RDONLY}]
    set xmlData [read  $fileID]
    close $fileID

    set document [dom parse $xmlData]
    fromDOM [$document documentElement] 
    resetModified
 
}


## ----------------------------------------------------------------------------
## Predicate returning 1, if configuration data has changed, 0 otherwise
## ----------------------------------------------------------------------------
body YadaConfigItem::isModified {} {
    return $_modified
}

## ----------------------------------------------------------------------------
## Marks item as modified
## ----------------------------------------------------------------------------
body YadaConfigItem::setModified {} {
    set _modified 1
    if { [string compare $name "<none>"] == 0 } {
	return
    }
    displayTitle
}

## ----------------------------------------------------------------------------
## Removes modification mark
## ----------------------------------------------------------------------------
body YadaConfigItem::resetModified {} {
    set _modified 0
    displayTitle
}


## ----------------------------------------------------------------------------
## Predicate returning 1, if file containing persistant state of YadaGrammar
## exists, otherwise 0
## ----------------------------------------------------------------------------
body YadaConfigItem::hasSavedState {} {
    set zippedFileName [getPersistanceFileName].gz 
    if { [::file exists $zippedFileName] } {
	return 1
    } else {
	return 0
    }
}


## ----------------------------------------------------------------------------
## Displays title in main window (reflecting modification state of document)
## ----------------------------------------------------------------------------
body YadaConfigItem::displayTitle {} {
    if { [isModified] && $name != "<none>" } {
	wm title .  "$name *** changed *** " 
    } else {
	wm title . "$name"
    }
}




