# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaDifference - difference between two experiments.
## \ingroup YadaRunnableDocuments
##
## \author Michael Daum
##
## $Id: YadaDifference.tcl,v 1.4 2004/02/25 14:42:08 micha Exp $
## ----------------------------------------------------------------------------
class YadaDifference {
  inherit YadaRunnableDocument

  # variables ----------------------------------------------------------------
  public variable betterAndFasterBackground "green1"
  public variable betterAndFasterForeground "black"
  public variable betterBackground "green2"
  public variable betterButSlowerBackground "green3"
  public variable betterButSlowerForeground "black"
  public variable betterForeground "black"
  public variable fasterBackground "lightblue"
  public variable fasterForeground "black"
  public variable harderBackground "deepskyblue3"
  public variable harderForeground "black"
  public variable sameBackground "gray"
  public variable sameForeground "black"
  public variable slowerBackground "yellow"
  public variable slowerForeground "black"
  public variable softerBackground "deepskyblue1"
  public variable softerForeground "black"
  public variable weirdBackground "pink"
  public variable weirdForeground "black"
  public variable worseAndSlowerBackground "red3"
  public variable worseAndSlowerForeground "white"
  public variable worseBackground "red1"
  public variable worseButFasterBackground "orange"
  public variable worseButFasterForeground "black"
  public variable worseForeground "white"

  private variable _totalWordgraphs 0
  private variable _fasterWordgraphs 0
  private variable _betterWordgraphs 0
  private variable _betterAndFasterWordgraphs 0
  private variable _betterButSlowerWordgraphs 0
  private variable _softerWordgraphs 0
  private variable _harderWordgraphs 0
  private variable _sameWordgraphs 0
  private variable _worseButFasterWordgraphs 0
  private variable _worseAndSlowerWordgraphs 0
  private variable _worseWordgraphs 0
  private variable _slowerWordgraphs 0
  private variable _weirdWordgraphs 0

  private variable _totalTime1 0
  private variable _score1 0
  private variable _totalTime2 0
  private variable _score2 0
  private variable _noSolutions1 0
  private variable _totalSolTime1 0
  private variable _totalSoftTime1 0
  private variable _noSolutions2 0
  private variable _totalSolTime2 0
  private variable _totalSoftTime2 0
  private variable _noErrors1 0
  private variable _noErrors2 0
  private variable _totalStructRecall1 0
  private variable _totalLexicalRecall1 0
  private variable _totalLabelledRecall1 0
  private variable _totalStructRecall2 0
  private variable _totalLexicalRecall2 0
  private variable _totalLabelledRecall2 0

  private variable _typeOfColumn ; ## \type TclArray
  private variable _scoreOf ; ## \type TclArray
  private variable _structHitsOf ; ## \type TclArray
  private variable _lexicalHitsOf ; ## \type TclArray
  private variable _labelledHitsOf ; ## \type TclArray
  private variable _triesOf ; ## \type TclArray

  private variable _prevExperimentName1 ""
  private variable _prevExperimentName2 ""

  # methods ------------------------------------------------------------------
  public method activationHandle {}
  public method load {args}; ## \type TclList
  public method save {{fileName ""}}; ## \type TclString
  public method step {{wordgraph ""}}; ## \type TclString
  public method clear {}
  public method runall {}
  public method run {}
  public method selectExperiment {experimentNo {experimentName ""}}; ## \type TclNumber, TclString

  constructor {args} {}; ## \type TclList

  private method _analyseDocument {document experiment}; ## \type DomDocument, YadaExperiment
  private method _judgeRow {row}; ## \type TclNumber
  private method _computeStatistics {}
  private method _judgeData {measure1a measure1b time1 solutionTime1 softTime1 measure2a measure2b time2 solutionTime2 softTime2}; ## \type TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber
  private method _getJudgeOfRow {row}; ## \type TclNumber
  private method _getWordgraphs {}
  private method _getSelectedExperiments {}
  private method _getTypeOfColumn {colNo} ; ## \type TclNumber
  private method _resetDataOf {wordgraphs}; ## \type TclList
  private method _jobConsumer {job}; ## \type YadaJob
  private method _getDoneWork {}
  private method _getUndoneWork {}

  protected method _viewLogfile {}
  protected method _viewXmlfile {}
  protected method _traceProcess {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaDifference::constructor {args} {

  # reconfigure table
  $itk_component(table) configure -rows 300 -cols 18 \
    -rowtagcommand [code $this _judgeRow]

  # set types of columns
  array set _typeOfColumn {
    0 "dictionary"
    1 "integer"
    2 "real"
    3 "integer"
    4 "integer"
    5 "integer"
    6 "integer"
    7 "real"
    8 "real"
    9 "integer"
    10 "integer"
    11 "integer"
    12 "integer"
    13 "real"
    14 "real"
    15 "real"
    16 "real"
    17 "ascii"
  }

  # experiments
  foreach i {1 2} {

    itk_component add experiment${i}BoxLabel {
      label $itk_component(heading).experiment${i}BoxLabel \
	  -text "Experiment ${i}:" \
	  -anchor w
    }  {}

    itk_component add experiment${i}Box {
      iwidgets::combobox $itk_component(heading).experiment${i}Box \
	  -completion false \
	  -borderwidth 2 \
	  -margin 4 \
	  -selectioncommand [code $this selectExperiment $i]
    }  {}

    itk_component add score${i}Label {
      label $itk_component(heading).score${i}Label \
	-text "Score:" \
	-anchor w \
    } {}

    itk_component add score${i} {
      entry $itk_component(heading).score${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    itk_component add totalTime${i}Label {
      label $itk_component(heading).totalTime${i}Label \
	-text "Time:" \
	-anchor w 
    } {}

    itk_component add totalTime${i} {
      entry $itk_component(heading).totalTime${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    itk_component add noSolutions${i}Label {
      label $itk_component(heading).noSolutions${i}Label \
	-text "No.Solutions:" \
	-anchor w 
    } {}

    itk_component add noSolutions${i} {
      entry $itk_component(heading).noSolutions${i} \
	-textvariable [scope _noSolutions${i}] \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    itk_component add noErrors${i}Label {
      label $itk_component(heading).noErrors${i}Label \
	-text "No.Errors:" \
	-anchor w 
    } {}

    itk_component add noErrors${i} {
      entry $itk_component(heading).noErrors${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    itk_component add totalSolTime${i}Label {
      label $itk_component(heading).totalSolTime${i}Label \
	-text "Sol.Time:" \
	-anchor w
    } {}

    itk_component add totalSolTime${i} {
      entry $itk_component(heading).totalSolTime${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    itk_component add totalSoftTime${i}Label {
      label $itk_component(heading).totalSoftTime${i}Label \
	-text "Soft Time:" \
	-anchor w
    } {}

    itk_component add totalSoftTime${i} {
      entry $itk_component(heading).totalSoftTime${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    itk_component add totalStructRecall${i}Label {
      label $itk_component(heading).totalStructRecall${i}Label \
	-text "Struct.Recall:" \
	-anchor w
    } {}
    
    itk_component add totalStructRecall${i} {
      entry $itk_component(heading).totalStructRecall${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    itk_component add totalLexicalRecall${i}Label {
      label $itk_component(heading).totalLexicalRecall${i}Label \
	-text "Lex.Recall:" \
	-anchor w
    } {}

    itk_component add totalLabelledRecall${i}Label {
      label $itk_component(heading).totalLabelledRecall${i}Label \
	-text "Label.Recall:" \
	-anchor w
    } {}
    
    itk_component add totalLabelledRecall${i} {
      entry $itk_component(heading).totalLabelledRecall${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

    
    itk_component add totalLexicalRecall${i} {
      entry $itk_component(heading).totalLexicalRecall${i} \
	-borderwidth 2 \
	-state disabled \
	-width 20
    } {}

   # bindings
    set entry [$itk_component(experiment${i}Box) component entry]
    bind $entry <Return> [code $this selectExperiment $i]
    bind $entry <Up> +[code $this selectExperiment $i]
    bind $entry <Down> +[code $this selectExperiment $i]
  }

  # separator
  itk_component add separator {
    frame $itk_component(heading).separator \
	-height 2 \
	-borderwidth 2 \
	-relief sunk
  }

  #
  # judge bar
  #
  itk_component add judgeFrame {
    frame $itk_component(heading).judgeFrame
  }

  itk_component add totalWordgraphs {
    iwidgets::entryfield $itk_component(judgeFrame).totalWordgraphs \
      -labeltext "Total:" \
      -textvariable [scope _totalWordgraphs] \
      -labelpos w \
      -borderwidth 2 \
      -state disabled \
      -disabledforeground black \
      -width 5
  } {}

  itk_component add judgeBar {
    YadaJudgeBar $itk_component(judgeFrame).judgeBar \
      -height 15 \
      -borderwidth 2 \
      -relief sunk \
      -betterAndFasterBackground $betterAndFasterBackground \
      -betterBackground $betterBackground \
      -betterButSlowerBackground $betterButSlowerBackground \
      -fasterBackground $fasterBackground \
      -softerBackground $softerBackground \
      -sameBackground $sameBackground \
      -harderBackground $harderBackground \
      -slowerBackground $slowerBackground \
      -worseButFasterBackground $worseButFasterBackground \
      -worseBackground $worseBackground \
      -worseAndSlowerBackground $worseAndSlowerBackground \
      -betterForeground $betterForeground \
      -betterAndFasterForeground $betterAndFasterForeground \
      -betterButSlowerForeground $betterButSlowerForeground \
      -fasterForeground $fasterForeground \
      -softerForeground $softerForeground \
      -sameForeground $sameForeground \
      -harderForeground $harderForeground \
      -slowerForeground $slowerForeground \
      -worseButFasterForeground $worseButFasterForeground \
      -worseAndSlowerForeground $worseAndSlowerForeground \
      -worseForeground $worseForeground \
      -weirdForeground $weirdForeground \
      -weirdBackground $weirdBackground
  } {}

  # packing
  grid columnconfigure $itk_component(heading) 1 -weight 1
  grid columnconfigure $itk_component(heading) 2 -pad 5
  grid columnconfigure $itk_component(heading) 3 -weight 1
  grid columnconfigure $itk_component(heading) 4 -pad 5
  grid columnconfigure $itk_component(heading) 5 -weight 1 
  grid columnconfigure $itk_component(heading) 6 -pad 5
  grid columnconfigure $itk_component(heading) 7 -weight 1
  grid columnconfigure $itk_component(heading) 8 -pad 5
  grid columnconfigure $itk_component(heading) 9 -weight 1
  grid columnconfigure $itk_component(heading) 10 -pad 5
  grid columnconfigure $itk_component(heading) 11 -weight 1
  grid columnconfigure $itk_component(heading) 12 -pad 5
  grid columnconfigure $itk_component(heading) 13 -weight 1
  grid columnconfigure $itk_component(heading) 14 -pad 5

  foreach i {1 2} {
    grid $itk_component(experiment${i}BoxLabel) \
      -row 0 -column [expr ($i - 1) * 8]     -sticky ew -pady 2
    grid $itk_component(experiment${i}Box) \
      -row 0 -column [expr ($i - 1) * 8 + 1]     -sticky ew -pady 2 -columnspan 5 

    grid $itk_component(totalTime${i}Label) \
      -row 1 -column [expr ($i - 1) * 8]     -sticky ew -pady 2
    grid $itk_component(totalTime${i}) \
      -row 1 -column [expr ($i - 1) * 8 + 1]     -sticky ew -pady 2

    grid $itk_component(score${i}Label)\
      -row 1 -column [expr ($i - 1) * 8 + 2] -sticky ew -pady 2
    grid $itk_component(score${i})\
      -row 1 -column [expr ($i - 1) * 8 + 3] -sticky ew -pady 2

    grid $itk_component(totalStructRecall${i}Label) \
      -row 1 -column [expr ($i - 1) * 8 + 4]     -sticky ew -pady 2
    grid $itk_component(totalStructRecall${i}) \
      -row 1 -column [expr ($i - 1) * 8 + 5] -sticky ew -pady 2

    grid $itk_component(totalSolTime${i}Label) \
      -row 2 -column [expr ($i - 1) * 8 ] -sticky ew -pady 2
    grid $itk_component(totalSolTime${i}) \
      -row 2 -column [expr ($i - 1) * 8 + 1] -sticky ew -pady 2

    grid $itk_component(noErrors${i}Label) \
      -row 2 -column [expr ($i - 1) * 8 + 2]     -sticky ew -pady 2
    grid $itk_component(noErrors${i}) \
      -row 2 -column [expr ($i - 1) * 8 + 3] -sticky ew -pady 2

    grid $itk_component(totalLabelledRecall${i}Label) \
      -row 2 -column [expr ($i - 1) * 8 + 4]     -sticky ew -pady 2
    grid $itk_component(totalLabelledRecall${i}) \
      -row 2 -column [expr ($i - 1) * 8 + 5] -sticky ew -pady 2

    grid $itk_component(totalSoftTime${i}Label) \
      -row 3 -column [expr ($i - 1) * 8 ] -sticky ew -pady 2
    grid $itk_component(totalSoftTime${i}) \
      -row 3 -column [expr ($i - 1) * 8 + 1] -sticky ew -pady 2

    grid $itk_component(noSolutions${i}Label) \
      -row 3 -column [expr ($i - 1) * 8 + 2]     -sticky ew -pady 2
    grid $itk_component(noSolutions${i}) \
      -row 3 -column [expr ($i - 1) * 8 + 3] -sticky ew -pady 2

    grid $itk_component(totalLexicalRecall${i}Label) \
      -row 3 -column [expr ($i - 1) * 8 + 4]     -sticky ew -pady 2
    grid $itk_component(totalLexicalRecall${i}) \
      -row 3 -column [expr ($i - 1) * 8 + 5] -sticky ew -pady 2

  }

  grid $itk_component(separator)  -sticky ew -row 4 -column 0 -pady 10 -columnspan 14
  grid $itk_component(judgeFrame) -sticky ew -row 7 -column 0 -pady 2 -columnspan 14

  grid columnconfigure $itk_component(judgeFrame) 1 -weight 1

  grid $itk_component(totalWordgraphs) -sticky w -row 0 -column 0 -pady 2
  grid $itk_component(judgeBar)     -sticky ew -row 0 -column 1 -pady 2 

  eval itk_initialize $args
  .main registerDocument $this
}

## ----------------------------------------------------------------------------
## _judgeRow
## ----------------------------------------------------------------------------
body YadaDifference::_judgeRow {row} {
  set judge [_getJudgeOfRow $row]
  set wordgraph [$itk_component(table) getCell $row,0]

  if {$judge == "" || $_doneWordgraph($wordgraph) != 3} {
    if {[expr $row % 2]} {
      return "default"
    } else {
      return "colored"
    }
  }
  
  return $judge
}


## ----------------------------------------------------------------------------
## clear
## ----------------------------------------------------------------------------
body YadaDifference::clear {} {
  chain ;# call YadaRunnableDocument::clear

  # reset the table 
  set table $itk_component(table)
  $table setCell \
       	0,0 "wordgraph" \
       	0,1 "values" \
       	0,2 "score" \
	0,3 "errors" \
       	0,4 "time" \
       	0,5 "sol.time" \
       	0,6 "soft time" \
	0,7 "lex.recall" \
       	0,8 "score" \
	0,9 "errors" \
       	0,10 "time" \
       	0,11 "sol.time" \
       	0,12 "soft time" \
	0,13 "lex.recall" \
	0,14 "speed up" \
       	0,15 "sol.speed up" \
       	0,16 "soft speed up" \
	0,17 "judge"


  $table tag config default -anchor e
  $table tag config colored -bg gray80 -anchor e
  $table tag config better -bg $betterBackground -fg $betterForeground -anchor e -relief flat
  $table tag config betterAndFaster -bg $betterAndFasterBackground -fg $betterAndFasterForeground -anchor e -relief flat
  $table tag config betterButSlower -bg $betterButSlowerBackground -fg $betterButSlowerForeground -anchor e -relief flat
  $table tag config faster -bg $fasterBackground -fg $fasterForeground -anchor e -relief flat
  $table tag config softer -bg $softerBackground -fg $softerForeground -anchor e -relief flat
  $table tag config harder -bg $harderBackground -fg $harderForeground -anchor e -relief flat
  $table tag config worseButFaster  -bg $worseButFasterBackground  -fg $worseButFasterForeground  -anchor e -relief flat
  $table tag config worseAndSlower  -bg $worseAndSlowerBackground  -fg $worseAndSlowerForeground  -anchor e -relief flat
  $table tag config worse  -bg $worseBackground  -fg $worseForeground  -anchor e -relief flat
  $table tag config slower -bg $slowerBackground -fg $slowerForeground -anchor e -relief flat
  $table tag config same   -bg $sameBackground   -fg $sameForeground   -anchor e -relief flat
  $table tag config weird  -bg $weirdBackground -fg $weirdForeground -anchor e -relief flat
  
  $table tag config sel -fg "" -bg "" -relief raised -bd 1
  $table tag config active -anchor w -fg black -bg gray -relief sunk -bd 1
  $table tag config title -font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} -bg #a3a3a3 -fg white
  $table tag config runner -fg red3 -bg #a3a3a3
  $table tag config judge -anchor w -bg "" -fg ""
  $table tag raise runner title

  $table width 0 15 1 5

  # reset scores of selected wordgraphs
  _resetDataOf "[_getActiveWordgraphs]"

  _computeStatistics

}

## ----------------------------------------------------------------------------
## load
## ----------------------------------------------------------------------------
body YadaDifference::load {args} {

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set wordgraphsInTable [$table getCell 0,0 $rmax,0]
  #$table configure -rowtagcommand ""

  # build the list of all available data
  if {[llength $args] == 0} {
    set activeWordgraphs [_getActiveWordgraphs]
  } else {
    set activeWordgraphs $args
  }
  set allWordgraphs ""
  set experiments [_getSelectedExperiments]
  if {[llength $experiments] != 2} {
    return
  }
  set experiment1 [lindex $experiments 0]
  set experiment2 [lindex $experiments 1]

  foreach wordgraph $activeWordgraphs {
    if {$_doneWordgraph($wordgraph) == 3} {
      continue
    }
    
    set file1 [$experiment1 getXmlFileName $wordgraph]
    set file2 [$experiment2 getXmlFileName $wordgraph]
    if {![file exists $file1]} {
      set file1 ""
    }
    if {![file exists $file2]} {
      set file2 ""
    }
    set allFiles($wordgraph,1) $file1
    set allFiles($wordgraph,2) $file2
    if {$file1 != "" || $file2 != ""} {
      lappend allWordgraphs $wordgraph
    }
  }

  # load all xml files 
  set noLoadedFiles 0
  foreach wordgraph $allWordgraphs {

    # did we press the break button
    if {$_gotBreak} {
      unbreakProcess
      print "WARNING: interrupted loading\n"
      break
    }

    # load data for both experiments
    foreach i {1 2} {
      if {($_doneWordgraph($wordgraph) & $i) == $i} {
	continue
      }
      set file $allFiles($wordgraph,$i)
      if {$file == ""} {
	set _doneWordgraph($wordgraph) [expr $_doneWordgraph($wordgraph) & ~$i]
	continue
      }
      set _doneWordgraph($wordgraph) [expr $_doneWordgraph($wordgraph) | $i]

      # say what we do
      print "INFO: loading file `$file'\n"

      # analyse the file
      set fd [open "|zcat $file" r]
      set data [read $fd]
      close $fd
      if {[catch {
	set document [dom parse $data]
	_analyseDocument $document [subst \$experiment$i]
      } errMsg ]} {
	print "ERROR: invalid xml file\n"
	print "$errMsg\n$::errorInfo"
      }
      catch {
	$document delete
	set document ""
      }

      incr noLoadedFiles
      update
    }
  }

  # draw statistics
  $itk_component(judgeBar) draw \
      $_fasterWordgraphs $_betterWordgraphs \
      $_betterAndFasterWordgraphs $_betterButSlowerWordgraphs $_softerWordgraphs \
      $_harderWordgraphs $_sameWordgraphs $_worseButFasterWordgraphs \
      $_worseAndSlowerWordgraphs $_worseWordgraphs $_slowerWordgraphs \
      $_weirdWordgraphs

  print "INFO: loaded $noLoadedFiles files\n"
  #$table configure -rowtagcommand [code $this _judgeRow]
  resortRows
  _computeStatistics
  .main printStatus ""
  $table selection clear all
}

## ----------------------------------------------------------------------------
## save
## ----------------------------------------------------------------------------
body YadaDifference::save {{fileName ""}} {
  return

  if {$fileName == "" } {
    set types {
      {"Xml logfile"	{.xml.gz}}
      {"CDG logfile"	{.log.gz}}
    }
    set fileName [tk_getSaveFile -filetypes $types -parent .]
    if {$fileName == ""} { 
      return 
    }
  }

}

## ----------------------------------------------------------------------------
## _computeStatistics
## ----------------------------------------------------------------------------
body YadaDifference::_computeStatistics {} {
  chain

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set cmax [$table cget -cols]

  set _totalWordgraphs 0
  set _fasterWordgraphs 0
  set _betterWordgraphs 0
  set _betterAndFasterWordgraphs 0
  set _betterButSlowerWordgraphs 0
  set _softerWordgraphs 0
  set _harderWordgraphs 0
  set _sameWordgraphs 0
  set _worseWordgraphs 0
  set _worseButFasterWordgraphs 0
  set _worseAndSlowerWordgraphs 0
  set _slowerWordgraphs 0
  set _weirdWordgraphs 0
  set _noErrors1 0
  set _noErrors2 0
  set _noSolutions1 0
  set _noSolutions2 0
  set _totalSolTime1 0
  set _totalSolTime2 0
  set _totalSoftTime1 0
  set _totalSoftTime2 0
  set _score1 0.0
  set _score2 0.0
  set _totalTime1 0
  set _totalTime2 0
  set _totalStructRecall1 0
  set _totalLexicalRecall1 0
  set _totalLabelledRecall1 0
  set _totalStructRecall2 0
  set _totalLexicalRecall2 0
  set _totalLabelledRecall2 0
  set totalTries1 0
  set totalStructHits1 0
  set totalLexicalHits1 0
  set totalLabelledHits1 0
  set totalTries2 0
  set totalStructHits2 0
  set totalLexicalHits2 0
  set totalLabelledHits2 0

  for {set row 1} {$row < $rmax} {incr row} {

    set dataRow [$table getCell $row,0 $row,$cmax]
    set noValues	[lindex $dataRow 1]
    if {$noValues == ""} {
      continue
    }

    set wordgraph	[lindex $dataRow 0]
    set noErrors1	[lindex $dataRow 3]
    set time1		[lindex $dataRow 4]
    set solutionTime1	[lindex $dataRow 5]
    set softTime1	[lindex $dataRow 6]
    set noErrors2	[lindex $dataRow 9]
    set time2		[lindex $dataRow 10]
    set solutionTime2	[lindex $dataRow 11]
    set softTime2	[lindex $dataRow 12]
    set speedUp		[lindex $dataRow 14]
    set solSpeedUp	[lindex $dataRow 15]
    set softSpeedUp	[lindex $dataRow 16]
    set judge		[lindex $dataRow 17]

    if {$_doneWordgraph($wordgraph) == 3} {
      incr _totalWordgraphs
    }
    incr _noErrors1 $noErrors1
    incr _noErrors2 $noErrors2
    incr _totalTime1 $time1
    incr _totalTime2 $time2
    incr _totalSolTime1 $solutionTime1
    incr _totalSoftTime1 $softTime1
    incr _totalSolTime2 $solutionTime2
    incr _totalSoftTime2 $softTime2

    set _score1 [expr $_score1 + $_scoreOf($wordgraph,1)]
    if {$_scoreOf($wordgraph,1) > 0.0} {
      incr _noSolutions1
    }
    set _score2 [expr $_score2 + $_scoreOf($wordgraph,2)]
    if {$_scoreOf($wordgraph,2) > 0.0} {
      incr _noSolutions2
    }

    incr totalTries1 $_triesOf($wordgraph,1)
    incr totalStructHits1 $_structHitsOf($wordgraph,1)
    incr totalLexicalHits1 $_lexicalHitsOf($wordgraph,1)
    incr totalLabelledHits1 $_labelledHitsOf($wordgraph,1)
    incr totalTries2 $_triesOf($wordgraph,2)
    incr totalStructHits2 $_structHitsOf($wordgraph,2)
    incr totalLexicalHits2 $_lexicalHitsOf($wordgraph,2)
    incr totalLabelledHits2 $_labelledHitsOf($wordgraph,2)

    if {$judge != ""} {
      incr _${judge}Wordgraphs
    }
  }

  if {$totalTries1 > 0} {
    set _totalStructRecall1 [expr $totalStructHits1 * 100.0 / $totalTries1  ]
    set _totalLexicalRecall1 [expr $totalLexicalHits1 * 100.0 / $totalTries1  ]
    set _totalLabelledRecall1 [expr $totalLabelledHits1 * 100.0 / $totalTries1  ]
  }
  if {$totalTries2 > 0} {
    set _totalStructRecall2 [expr $totalStructHits2 * 100.0 / $totalTries2  ]
    set _totalLexicalRecall2 [expr $totalLexicalHits2 * 100.0 / $totalTries2  ]
    set _totalLabelledRecall2 [expr $totalLabelledHits2 * 100.0 / $totalTries2  ]
  }

  # show statistics
  foreach i {1 2} {
    set noSolutions [subst \$_noSolutions${i}]
    set totalTime [subst \$_totalTime${i}]
    set totalSolTime [subst \$_totalSolTime${i}]
    set totalSoftTime [subst \$_totalSoftTime${i}]
    set score [subst \$_score${i}]
    set noErrors [subst \$_noErrors${i}]
    set totalStructRecall [subst \$_totalStructRecall${i}]
    set totalLexicalRecall [subst \$_totalLexicalRecall${i}]
    set totalLabelledRecall [subst \$_totalLabelledRecall${i}]
    
    $itk_component(totalTime${i}) configure -state normal
    $itk_component(totalSolTime${i}) configure -state normal
    $itk_component(totalSoftTime${i}) configure -state normal
    $itk_component(score${i}) configure -state normal
    $itk_component(noErrors${i}) configure -state normal
    $itk_component(totalStructRecall${i}) configure -state normal
    $itk_component(totalLexicalRecall${i}) configure -state normal
    $itk_component(totalLabelledRecall${i}) configure -state normal

    $itk_component(totalTime${i}) delete 0 end
    $itk_component(totalSolTime${i}) delete 0 end
    $itk_component(totalSoftTime${i}) delete 0 end
    $itk_component(score${i}) delete 0 end
    $itk_component(noErrors${i}) delete 0 end
    $itk_component(totalStructRecall${i}) delete 0 end
    $itk_component(totalLexicalRecall${i}) delete 0 end
    $itk_component(totalLabelledRecall${i}) delete 0 end

    $itk_component(score${i}) insert 0 [format "%.4f" $score]
    $itk_component(noErrors${i}) insert 0 $noErrors
    $itk_component(totalTime${i}) insert 0 "[timeToString $totalTime]"
    $itk_component(totalSolTime${i}) insert 0 "[timeToString $totalSolTime]"
    $itk_component(totalSoftTime${i}) insert 0 "[timeToString $totalSoftTime]"
    $itk_component(totalStructRecall${i}) insert 0 [format "%.4f%%" $totalStructRecall]
    $itk_component(totalLexicalRecall${i}) insert 0 [format "%.4f%%" $totalLexicalRecall]
    $itk_component(totalLabelledRecall${i}) insert 0 [format "%.4f%%" $totalLabelledRecall]

    $itk_component(totalTime${i}) configure -state disabled
    $itk_component(totalSolTime${i}) configure -state disabled
    $itk_component(totalSoftTime${i}) configure -state disabled
    $itk_component(score${i}) configure -state disabled
    $itk_component(noErrors${i}) configure -state disabled
    $itk_component(totalStructRecall${i}) configure -state disabled
    $itk_component(totalLexicalRecall${i}) configure -state disabled
    $itk_component(totalLabelledRecall${i}) configure -state disabled
  }

  # redraw the judge canvas
  $itk_component(judgeBar) draw \
      $_fasterWordgraphs $_betterWordgraphs \
      $_betterAndFasterWordgraphs $_betterButSlowerWordgraphs $_softerWordgraphs \
      $_harderWordgraphs $_sameWordgraphs $_worseButFasterWordgraphs \
      $_worseAndSlowerWordgraphs $_worseWordgraphs $_slowerWordgraphs \
      $_weirdWordgraphs

}

## ----------------------------------------------------------------------------
## selectExperiment
## ----------------------------------------------------------------------------
body YadaDifference::selectExperiment {experimentNo {experimentName ""}} {
  set box $itk_component(experiment${experimentNo}Box)
  if {$experimentName == ""} {
    set experimentName [$box get]
    if {$experimentName == ""} {
      set experimentName "<none>"
    }
  }

  set prevExperimentName [subst \$_prevExperimentName${experimentNo}]
  if {$prevExperimentName != $experimentName} {
    set _prevExperimentName${experimentNo} $experimentName
    foreach wordgraph [array names _doneWordgraph] {
      unset _doneWordgraph($wordgraph)
    }
    clear
    $box clear entry
    $box insert entry 0 "$experimentName"
    
  }
  
  set listOfEntries [[$box component list] get 0 end]
  set index [lsearch $listOfEntries $experimentName]

  if {$index < 0} {
    selectExperiment $experimentNo "<none>"
  } else {
    $box selection set $index $index
  }
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaDifference::activationHandle {} {
  chain ;# call YadaRunnableDocument::activationHandle

  set experimentNames [.main getActiveExperiments]

  foreach i {1 2} {

  # get the previous selection
    set box $itk_component(experiment${i}Box)
    set experimentName [$box get]

    # default experiments to choose
    if {$experimentName == ""} {
      set experimentName "<none>"
    }

    set listOfEntries [[$box component list] get 0 end]
    if {$listOfEntries != $experimentNames} {
      $box clear list 
      eval $box insert list end $experimentNames
      $box insert entry 0 "$experimentName"
    }
  }
}

## ----------------------------------------------------------------------------
## _judgeData
## ----------------------------------------------------------------------------
body YadaDifference::_judgeData {measure1a measure1b time1 solutionTime1 softTime1 measure2a measure2b time2 solutionTime2 softTime2} {

  set significance [.main getSignificance]

  # is there an annotation? if not then
  if {$measure1a == 0 && $measure2a == 0} {
    # ... use scores
    set measure1 $measure1b
    set measure2 $measure2b
  } else {
    # ... otherwise use recall
    set measure1 $measure1a
    set measure2 $measure2a
  }

  # compare properties
  set measureDiff [expr $measure1 - $measure2 ]
  set speedUp [_computeSpeedup $time1 $time2]
  set solSpeedUp [_computeSpeedup $solutionTime1 $solutionTime2]
  set softSpeedUp [_computeSpeedup $softTime1 $softTime2]

  # judge
  if {$measureDiff > 1e-5} {
    if {$solutionTime2 != 0.0} { ;# does experiment 2 give us anything
      if {$speedUp < [expr 1.0 / $significance]} {
	set judge "betterButSlower"
      } elseif {$speedUp > $significance} {
	set judge "betterAndFaster"
      } else {
	set judge "better"
      }
    } else {
      set judge "better"
    }
  } elseif {$measureDiff < -1e-5} {
    if {$solutionTime1 != 0.0} { ;# does experiment 1 give us anything
      if {$speedUp > $significance} {
	set judge "worseButFaster"
      } elseif {$speedUp < [expr 1.0 / $significance]} {
	set judge "worseAndSlower"
      } else {
	set judge "worse"
      }
    } else {
      set judge "worse"
    }
  } elseif {$speedUp < [expr 1.0 / $significance]} {
    set judge "slower"
  } elseif {$speedUp > $significance} {
    set judge "faster"
  } elseif {$solSpeedUp > $significance} {
    set judge "softer"
  } elseif {$solSpeedUp < [expr 1.0 / $significance]} {
    set judge "harder"
  } elseif {$softSpeedUp > $significance} {
    set judge "softer"
  } elseif {$softSpeedUp < [expr 1.0 / $significance]} {
    set judge "harder"
  } else {
    set judge "same"
  }

  # weirdness check, this relys on measure1b and measure2b to be scores
  # measure a: recall
  # measure b: score
  if {[.main getWeirdness]} {
    if {(
      	  $measure1a == 100.0 &&			
  	  $measure2a < [expr $measure1a - 1e-10] &&
  	  $measure1b <= $measure2b                
    	) || (
  	  $measure2a == 100.0 &&			
  	  $measure1a < [expr $measure2a - 1e-10] &&
  	  $measure2b <= $measure1b                
    	)} {
      set judge "weird"
    } 
  } else {
    if {(
	  $measure2b != 0.0 &&
	  $measure2a < [expr $measure1a - 1e-10] &&
	  $measure1b <= $measure2b                
	) || (
	  $measure1b != 0.0 &&
	  $measure1a < [expr $measure2a - 1e-10] &&
	  $measure2b <= $measure1b                
	)} {
      set judge "weird"
    }
  }

  return $judge
}

## ----------------------------------------------------------------------------
## _getSelectedExperiments
## ----------------------------------------------------------------------------
body YadaDifference::_getSelectedExperiments {} {

  set experiments ""
  set experimentName [$itk_component(experiment1Box) get]
  if {$experimentName != "" && $experimentName != "<none>"} {
    lappend experiments [.main getExperiment $experimentName]
  }

  set experimentName [$itk_component(experiment2Box) get]
  if {$experimentName != "" && $experimentName != "<none>"} {
    lappend experiments [.main getExperiment $experimentName]
  }

  return $experiments
}

## ----------------------------------------------------------------------------
## _getWordgraphs
## ----------------------------------------------------------------------------
body YadaDifference::_getWordgraphs {} {

  set experiments [_getSelectedExperiments]
  if {[llength $experiments] != 2} {
    return ""
  }
  
  set experiment1 [lindex $experiments 0]
  set experiment2 [lindex $experiments 1]

  # shortcut for two experiments using the same grammar
  if {[$experiment1 cget -grammarName] == [$experiment2 cget -grammarName]} {
    return [.main getWordgraphsOfExperiment [$experiment1 cget -name]]
  }
  
  # compute intersection of the used wordgraphs
  set wordgraphs1 [.main getWordgraphsOfExperiment [$experiment1 cget -name]]
  set wordgraphs2 [.main getWordgraphsOfExperiment [$experiment2 cget -name]]

  set commonWordgraphs ""
  foreach wordgraph $wordgraphs1 {
    if {[lsearch $wordgraphs2 $wordgraph] >=0} {
      lappend commonWordgraphs $wordgraph
    }
  }

  return $commonWordgraphs
}

## ----------------------------------------------------------------------------
## _getJudge
## ----------------------------------------------------------------------------
body YadaDifference::_getJudgeOfRow {row} {
  set cmax [$itk_component(table) cget -cols]
  return [$itk_component(table) getCell $row,$cmax]
}

## ----------------------------------------------------------------------------
## _getTypeOfColumn
## ----------------------------------------------------------------------------
body YadaDifference::_getTypeOfColumn {colNo} {
  return $_typeOfColumn($colNo)
}

## ----------------------------------------------------------------------------
## _resetDataOf
## ----------------------------------------------------------------------------
body YadaDifference::_resetDataOf {wordgraphs} {
  chain $wordgraphs

  foreach wordgraph $wordgraphs {
    foreach experimentNo {1 2} {
      set _scoreOf($wordgraph,$experimentNo) 0.0
      set _structHitsOf($wordgraph,$experimentNo) 0
      set _lexicalHitsOf($wordgraph,$experimentNo) 0
      set _labelledHitsOf($wordgraph,$experimentNo) 0
      set _triesOf($wordgraph,$experimentNo) 0
    }
  }
}

## ----------------------------------------------------------------------------
## step
## ----------------------------------------------------------------------------
body YadaDifference::step {{wordgraph ""}} {

  if {$wordgraph == ""} {
    set wordgraph [lindex [_getActiveWordgraphs] 0]
    if {$wordgraph == ""} {
      return 0
    }
  }

  # check and load
  set experiments [_getSelectedExperiments]
  if {[llength $experiments] != 2} {
    return 0
  }
  set experiment1 [lindex $experiments 0]
  set experiment2 [lindex $experiments 1]

  set file1 [$experiment1 getXmlFileName $wordgraph]
  set file2 [$experiment2 getXmlFileName $wordgraph]

  set machine1 [.main getMachine [$experiment1 cget -machineName]]
  set machine2 [.main getMachine [$experiment2 cget -machineName]]

  if {[file exists $file1] || [file exists $file2]} {
    load $wordgraph
  }

    # construct the job
  if {![file exists $file1]} {
    set job1 [YadaJob ::#auto \
      -wordgraphName $wordgraph \
      -experiment $experiment1 \
      -runner $this \
      -machine $machine1 \
      -consumerCommand [code $this _jobConsumer]]
    _registerJob $job1
    $machine1 queueJob $job1
  }
  if {![file exists $file2]} {
    set job2 [YadaJob ::#auto \
      -wordgraphName $wordgraph \
      -experiment $experiment2 \
      -runner $this \
      -machine $machine2 \
      -consumerCommand [code $this _jobConsumer]]
    _registerJob $job2
    $machine2 queueJob $job2
  }

  # start the job
  if {$machine1 != $machine2} {
    $machine1 startJob
  }
  $machine2 startJob

  # wait
  set _isRunning 1
  while {$_isRunning} {
    vwait [scope _isRunning]
  }

  # finish message
  if {$_gotBreak} {
    printMessage "interrupted job(s)"
    unbreakJob
  } else {
    printMessage "finished job(s)"
  }

  return 1
}

## ----------------------------------------------------------------------------
## run
## ----------------------------------------------------------------------------
body YadaDifference::run {} {
  set activeWordgraphs [_getActiveWordgraphs]
  
  # get the current experiments
  foreach experiment [_getSelectedExperiments] {

    # check and load
    .main printStatus "... checking experiment '[$experiment cget -name]'"
    $experiment check
    load
    .main printStatus ""
    if {[$experiment cget -isComplete]} {
      continue
    }

    # construct all jobs 
    set machine [.main getMachine [$experiment cget -machineName]]
    foreach wordgraph $activeWordgraphs {
      set job [YadaJob ::#auto \
	-wordgraphName $wordgraph \
	-experiment $experiment \
	-runner $this \
	-machine $machine \
	-consumerCommand [code $this _jobConsumer]]
      _registerJob $job

      # queue jobs of the same wordgraph
      lappend jobsOfWordgraph($wordgraph) $job
    }
  }

  foreach wordgraph [lsort -dictionary [array names jobsOfWordgraph]] {
    set jobs $jobsOfWordgraph($wordgraph)
    foreach job $jobs {
      set machine [$job cget -machine]
      $machine queueJob $job
      set machines($machine) "horst"
    }
  }

  foreach $machine [array names machines] {
    $machine startJob
  }
  
  # wait
  set _isRunning 1
  while {$_isRunning} {
    vwait [scope _isRunning]
  }

  # finish message
  if {$_gotBreak} {
    printMessage "interrupted job(s)"
    unbreakJob
  } else {
    printMessage "finished job(s)"
  }
}

## ----------------------------------------------------------------------------
## runall
## ----------------------------------------------------------------------------
body YadaDifference::runall {} {
  set refExperimentName [$itk_component(experiment1Box) get]

  # need a ref experiment to run all
  if {$refExperimentName == "" || $refExperimentName == "<none>"} {
    printMessage "ERROR: need a reference experiment to run all other experiments" error
    return
  }
  
  set experimentNames [[$itk_component(experiment2Box) component list] get 0 end]
  set refExperiment [.main getExperiment $refExperimentName]
  
  # construct possible jobs
  set experiments ""
  set refExperimentGrammarName [$refExperiment cget -grammarName]
  foreach experimentName $experimentNames {
    set experiment [.main getExperiment $experimentName]
    if {[$experiment cget -grammarName] != $refExperimentGrammarName ||
	$experimentName == "<none>" ||
	$experimentName == "" ||
	$experimentName == $refExperimentName ||
	![$experiment cget -isActive]} {
	continue
    }

    # force experiment checking
    .main printStatus "... checking experiment `$experimentName'"
    if {[$experiment check]} { 
      continue ;# is complete
    }

    # construct jobs
    lappend experiments $experiment
    set noJobs 0
    set grammar [.main getGrammar [$experiment cget -grammarName]]
    set machine [.main getMachine [$experiment cget -machineName]]
    foreach wordgraph [$grammar getSelection] {
      set xmlFile [$experiment getXmlFileName $wordgraph]
      # dont recompute existing data
      if {[file exists $xmlFile]} {
	continue 
      }

      set job [YadaJob ::#auto \
	-wordgraphName $wordgraph \
	-experiment $experiment \
	-runner $this \
	-machine $machine \
	-consumerCommand [code $this _jobConsumer]]
      _registerJob $job
      incr noJobs
    }

    print "INFO: $experimentName needs $noJobs jobs\n"
  }

  # confirm batch of jobs
  if {![askRunall]} {
    foreach experiment [array names _currentJobs] {
      foreach  job $_currentJobs($experiment) {
	_unregisterJob $job
	itcl::delete object $job
	update ;# this loop could be time consuming
      }
    }
    return
  }
  
  # do it
  set _timer [clock seconds] 
  foreach experiment $experiments {

    # select experiment
    selectExperiment 2 [$experiment cget -name]
    load
    if {[$experiment cget -isComplete]} {
      foreach job $_currentJobs($experiment) {
	_unregisterJob $job
	itcl::delete object $job
      }
      continue
    }

    # queue jobs of the current experiment
    foreach job $_currentJobs($experiment) {
      set machine [$job cget -machine]
      $machine queueJob $job
    }

    # start
    $machine startJob
    
    # wait
    set _isRunning 1
    while {$_isRunning} {
      vwait [scope _isRunning]
    }

    # escape
    if {$_gotBreak} {
      break
    }
  }

  # finish message
  if {$_gotBreak} {
    printMessage "interrupted job(s)"
    unbreakJob
  } else {
    printMessage "finished job(s)"
  }

}


## ----------------------------------------------------------------------------
## _jobConsumer
## ----------------------------------------------------------------------------
body YadaDifference::_jobConsumer {job} {
  $job setState "consuming"
  set wordgraph [$job cget -wordgraphName]
  set experiment [$job cget -experiment]
  set machine [$job cget -machine]
  set experiments [_getSelectedExperiments]
  set experiment1 [lindex $experiments 0]
  set experiment2 [lindex $experiments 1]


  # remove the job from the list of current once
  unmarkWordgraph $wordgraph
  _unregisterJob $job
  $machine stopJob $job

  # refresh the statistics
  if {$experiment == $experiment1} {
    set _doneWordgraph($wordgraph) [expr $_doneWordgraph($wordgraph) | 1]
  } elseif {$experiment == $experiment2} {
    set _doneWordgraph($wordgraph) [expr $_doneWordgraph($wordgraph) | 2]
  } else {
    # never reach
    error "error consuming foreign experiment '$experiment',\nknown experiments are '$experiment1', '$experiment2'"
  }

  # analyse the new data
  set xmlFile [$experiment getXmlFileName $wordgraph]
  print "INFO: loading `$xmlFile'\n"
  set fd [open "|zcat $xmlFile" r]
  set data [read $fd]
  close $fd

  if {[catch {
    set document [dom parse $data]
    _analyseDocument $document $experiment
  } errMsg ]} {
    print "ERROR: invalid xml file\n"
    print "$errMsg\n$::errorInfo"
  }
  catch {
    $document delete
    set document ""
  }

  # start a new job
  if {!$_gotBreak} {
    if {[$machine startJob]} {
      return
    }
  }

  # detect finish
  if {$_isRunning} {
    if {(![info exists _currentJobs($experiment1)] ||
	  [llength $_currentJobs($experiment1)] == 0)
	&&
	(![info exists _currentJobs($experiment2)] ||
	  [llength $_currentJobs($experiment2)] == 0)} {
      set _isRunning 0 ;# signal vwait
    }
  }
}

## ----------------------------------------------------------------------------
## _analyseDocument
## ----------------------------------------------------------------------------
body YadaDifference::_analyseDocument {document experiment} {

  # default analysis
  YadaRunnableDocument::_analyseDocument $document $experiment

  # the wordgraph is needed at least
  set wordgraph $_analyseResult(wordgraph)
  if {$wordgraph == ""} {
    print "ERROR: no information available in this run\n"
    return
  }

  # compute derived data
  set row [_getRowOfWordgraph $wordgraph]
  set dataRow [$itk_component(table) getCell $row,3 $row,13]
  set noErrors1		[lindex $dataRow 0]
  set time1		[lindex $dataRow 1]
  set solutionTime1	[lindex $dataRow 2]
  set softTime1		[lindex $dataRow 3]
  set noErrors2		[lindex $dataRow 6]
  set time2		[lindex $dataRow 7]
  set solutionTime2	[lindex $dataRow 8]
  set softTime2		[lindex $dataRow 9]

  if {$_triesOf($wordgraph,1) == 0} {
    set lexicalRecall1 0
  } else {
    set lexicalRecall1 [expr $_lexicalHitsOf($wordgraph,1) * 100.0 / $_triesOf($wordgraph,1)]
  }
  if {$_triesOf($wordgraph,2) == 0} {
    set lexicalRecall2 0
  } else {
    set lexicalRecall2 [expr $_lexicalHitsOf($wordgraph,2) * 100.0 / $_triesOf($wordgraph,2)]
  }

  # save score and edge data
  if {[$itk_component(experiment1Box) get] == [$experiment cget -name]} {
    set experimentNo 1
  } else {
    set experimentNo 2
  }
  set _scoreOf($wordgraph,$experimentNo) $_analyseResult(score)
  set _structHitsOf($wordgraph,$experimentNo) $_analyseResult(structHits)
  set _lexicalHitsOf($wordgraph,$experimentNo) $_analyseResult(lexicalHits)
  set _labelledHitsOf($wordgraph,$experimentNo) $_analyseResult(labelledHits)
  set _triesOf($wordgraph,$experimentNo) $_analyseResult(tries)

  # overwrite the given experiment data
  set score1 $_scoreOf($wordgraph,1)
  set score2 $_scoreOf($wordgraph,2)
  set time${experimentNo} $_analyseResult(totalTime)
  set solutionTime${experimentNo} $_analyseResult(solutionTime)
  set softTime${experimentNo} $_analyseResult(softTime)
  set noErrors${experimentNo} [expr $_analyseResult(tries) - $_analyseResult(lexicalHits) ]
  set lexicalRecall${experimentNo} [expr $_analyseResult(lexicalHits) * 100.0 / $_analyseResult(tries)]

  if {($_doneWordgraph($wordgraph) & 1) == 1} {
    if {($_doneWordgraph($wordgraph) & 2) == 2} {
      set speedUp [_computeSpeedup $time1 $time2]
      set solSpeedUp [_computeSpeedup $solutionTime1 $solutionTime2]
      set softSpeedUp [_computeSpeedup $softTime1 $softTime2]
      set judge [_judgeData \
	$lexicalRecall1 $score1 $time1 $solutionTime1 $softTime1 \
	$lexicalRecall2 $score2 $time2 $solutionTime2 $softTime2]
      if {$judge == "weird"} {
	print "ERROR: wordgraph '$wordgraph' is weird: score1=$score1 score2=$score2 lexicalRecall1=$lexicalRecall1 lexicalRecall2=$lexicalRecall2\n"
      }
    } else {
      set noErrors2 0
      set time2 0
      set solutionTime2 0
      set softTime2 0
      set lexicalRecall2 0
      set speedUp 1.0
      set solSpeedUp 1.0
      set softSpeedUp 1.0
      set judge ""
    }
  } else {
    set noErrors1 0
    set time1 0
    set solutionTime1 0
    set softTime1 0
    set lexicalRecall1 0
    set speedUp 1.0
    set solSpeedUp 1.0
    set softSpeedUp 1.0
    set judge ""
  }

  # add the collected data to the table
  _insertData $wordgraph \
	      $_analyseResult(noValues) \
	      [format "%4.3e" $score1] \
	      $noErrors1 \
	      $time1 \
	      $solutionTime1 \
	      $softTime1 \
	      [format "%3.2f" $lexicalRecall1] \
	      [format "%4.3e" $score2] \
	      $noErrors2 \
	      $time2 \
	      $solutionTime2 \
	      $softTime2 \
	      [format "%3.2f" $lexicalRecall2] \
	      [format "%.2g" $speedUp] \
	      [format "%.2g" $solSpeedUp] \
	      [format "%.2g" $softSpeedUp] \
	      $judge
}

## ----------------------------------------------------------------------------
## _getDoneWork
## ----------------------------------------------------------------------------
body YadaDifference::_getDoneWork {} {
  set result ""
  foreach {wordgraph experimentNo} [array names _doneWordgraph] {
    if {$_doneWordgraph($wordgraph) == 3} {
      lappend result $wordgraph
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## _getUndoneWork
## ----------------------------------------------------------------------------
body YadaDifference::_getUndoneWork {} {
  set result ""

  foreach wordgraph [array names _doneWordgraph] {
    if {$_doneWordgraph($wordgraph) != 3} {
      lappend result $wordgraph
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## _viewLogfile: wrapper to YadaRunnableDocument::_viewLogfile
## ----------------------------------------------------------------------------
body YadaDifference::_viewLogfile {} {
  chain [_getSelectedExperiments]
}

## ----------------------------------------------------------------------------
## _viewXmlfile: wrapper to YadaRunnableDocument::_viewXmlfile
## ----------------------------------------------------------------------------
body YadaDifference::_viewXmlfile {} {
  chain [_getSelectedExperiments]
}

## ----------------------------------------------------------------------------
## _traceProcess
## ----------------------------------------------------------------------------
body YadaDifference::_traceProcess {} {
  # TODO
}
