# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaExperiments - configure document for all experiments.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaExperiments.tcl,v 1.7 2004/03/25 13:58:24 ddreyer Exp $
## ----------------------------------------------------------------------------
class YadaExperiments {
  inherit YadaConfigDocument


  # variables ----------------------------------------------------------------
  private variable _experimentTypes	""
  private variable _isActive 1
  private variable _isComplete 0

  # methods ------------------------------------------------------------------
  public method init {}
  public method getActiveExperiments {args}; ## \type TclList
  public method activationHandle {}
  public method load {}
  public method save {}

  constructor {args} {}; ## \type TclList

  private method _select {{experimentName ""}}; ## \type TclString
  private method _commit {}
  private method _mark {aspect}; ## \type TclString
  private method _new {}
  private method _dataDirChooser {}
  private method _delete {}
  private method _defaults {{fileName ""}}; ## \type TclString
  private method _initGrammarNames {}
  private method _checkExperiment {}

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaExperiments::constructor {args} {

  set _allAspects {name type command dataDir grammarName machineName \
      script status} 

  # edit menu
  set index [$itk_component(editMenu) index "Delete"]
  incr index
  $itk_component(editMenu) insert [incr index] command \
       	-label "Check Experiment" \
       	-underline 0 \
       	-command [code $this _checkExperiment]

  # type
  itk_component add typeLabel {
    label $itk_component(childsite).typeLabel \
      -text "Type:" \
      -anchor nw
  } {}

  itk_component add typeMark {
    frame $itk_component(childsite).typeMark \
      -borderwidth 2
  } {}

  itk_component add type {
    iwidgets::combobox $itk_component(typeMark).box \
      -completion false \
      -unique true \
      -borderwidth 2 \
      -editable 0 \
  }  {}
  set entry [$itk_component(type) component entry]
  $entry configure \
    -validate all \
    -validatecommand [code $this _validateCommand "type" %P]

  # grammar 
  itk_component add grammarLabel {
    label $itk_component(childsite).grammarLabel \
      -text "Grammar:" \
      -anchor nw 
  } {}

  itk_component add grammarNameMark {
    frame $itk_component(childsite).grammarNameMark \
      -borderwidth 2
  } {}

  itk_component add grammarName {
    iwidgets::combobox $itk_component(grammarNameMark).box \
      -completion false \
      -unique true \
      -borderwidth 2 \
      -editable 0 \
  }  {}
  set entry [$itk_component(grammarName) component entry]
  $entry configure \
    -validate all \
    -validatecommand [code $this _validateCommand "grammarName" %P]

  # machine 
  itk_component add machineLabel {
    label $itk_component(childsite).machineLabel \
      -text "Machine:" \
      -anchor nw 
  } {}

  itk_component add machineNameMark {
    frame $itk_component(childsite).machineNameMark \
      -borderwidth 2
  } {}

  itk_component add machineName {
    iwidgets::combobox $itk_component(machineNameMark).box \
      -completion false \
      -unique true \
      -borderwidth 2 \
      -editable 0 \
  }  {}
  set entry [$itk_component(machineName) component entry]
  $entry configure \
    -validate all \
    -validatecommand [code $this _validateCommand "machineName" %P]

  # experiment script
  itk_component add scriptLabel {
    label $itk_component(childsite).scriptLabel \
      -text "Script:" \
      -anchor nw
  } {}

  itk_component add scriptMark {
    frame $itk_component(childsite).scriptMark \
      -borderwidth 2
  } {}

  itk_component add script {
    iwidgets::scrolledtext $itk_component(scriptMark).text \
      -hscrollmode none \
      -vscrollmode static \
      -wrap none \
      -borderwidth 2 \
      -sbwidth 12 \
      -troughcolor gray \
      -textbackground gray80 \
      -height 100 \
      -highlightthickness 2 \
      -highlightcolor gray90 \
      -textfont fixed
  } {}


  # command 
  itk_component add commandLabel {
    label $itk_component(childsite).commandLabel \
      -text "Command:" \
      -anchor nw
  } {}

  itk_component add commandMark {
    frame $itk_component(childsite).commandMark \
      -borderwidth 2
  } {}

  itk_component add command {
    iwidgets::combobox $itk_component(commandMark).box \
      -completion false \
      -unique true \
      -borderwidth 2 \
  }  {}
  set entry [$itk_component(command) component entry]
  $entry configure \
    -validate all \
    -validatecommand [code $this _validateCommand "command" %P]

  # data dir
  itk_component add dataDirLabel {
    label $itk_component(childsite).dataDirLabel\
      -text "Data Dir:" \
      -anchor nw
  } {}

  itk_component add dataDirMark {
    frame $itk_component(childsite).dataDirMark \
      -borderwidth 2
  } {}

  itk_component add dataDir {
    entry $itk_component(dataDirMark).entry \
      -borderwidth 2 \
      -validate all \
      -validatecommand [code $this _validateCommand "dataDir" %P]
  } {}

  itk_component add dataDirButton {
    button $itk_component(dataDirMark).button \
      -image [YadaImages::get defaultFolder] \
      -padx 5 \
      -pady 0 \
      -command [code $this _dataDirChooser]
  } {}

  # status buttons
  itk_component add statusLabel {
    label $itk_component(childsite).statusLabel \
	-text "Status:" \
	-anchor nw
  }

  itk_component add statusMark {
    frame $itk_component(childsite).statusMark \
      -borderwidth 2
  } {}

  itk_component add isActive {
    checkbutton $itk_component(statusMark).isActive \
      -text "Active" \
      -variable [scope _isActive] \
      -command [code $this _mark "isActive"]
  } {}

  itk_component add isComplete {
    checkbutton $itk_component(statusMark).isComplete \
      -text "Complete" \
      -variable [scope _isComplete] \
      -command [code $this _mark "isComplete"]
  } {}

  # more buttons
  itk_component add checkButton {
    button $itk_component(topButtonFrame).checkButton \
	-text "Check" \
	-command [code $this _checkExperiment]
  }


  # 
  # bindings
  #

  # script texts
  set text [$itk_component(script) childsite]
  bind $text <FocusOut> [code $this _mark "script"]
  bind $text <Leave> [code $this _mark "script"]
  bind $text <KeyPress> [code $this _mark "script"]

  #
  # packing + gridding
  #
  grid rowconfigure $itk_component(childsite) 7 -weight 1

  grid $itk_component(typeLabel) -sticky w -row 2 -column 0 
  grid $itk_component(typeMark) -sticky ew -row 2 -column 1 -pady 10
  pack $itk_component(type) -fill both -expand 1

  grid $itk_component(grammarLabel) -sticky ew -row 3 -column 0
  grid $itk_component(grammarNameMark) -sticky ew -row 3 -column 1 -pady 10 -columnspan 2
  pack $itk_component(grammarName) -fill both -expand 1

  grid $itk_component(commandLabel) -sticky ew -row 4 -column 0  -pady 10
  grid $itk_component(commandMark) -sticky ew -row 4 -column 1  -pady 10 -columnspan 2
  pack $itk_component(command) -fill both -expand 1 
  
  grid $itk_component(machineLabel) -sticky ew -row 5 -column 0  -pady 10
  grid $itk_component(machineNameMark) -sticky ew -row 5 -column 1  -pady 10 -columnspan 2
  pack $itk_component(machineName) -fill both -expand 1 
  
  grid $itk_component(dataDirLabel) -sticky ew -row 6 -column 0 -pady 10
  grid $itk_component(dataDirMark) -sticky ew -row 6 -column 1 -pady 10 -columnspan 2
  pack $itk_component(dataDir) -fill both -expand 1 -side left
  pack $itk_component(dataDirButton) -side left
  
  grid $itk_component(scriptLabel) -sticky new -row 7 -column 0 -pady 10
  grid $itk_component(scriptMark) -sticky news -row 7 -column 1 -rowspan 2 -pady 10 -columnspan 2
  pack $itk_component(script) -fill both -expand 1

  grid $itk_component(statusLabel) -sticky w -row 9 -column 0
  grid $itk_component(statusMark) -sticky w -row 9 -column 1 
  pack $itk_component(isActive) \
       $itk_component(isComplete) -fill both -side left -expand 1

  pack $itk_component(checkButton) -side top -fill x -pady 5

  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaExperiments::init {} {
  if {$_isInitialized} {
    return
  }
  chain

  # initialize the comboboxes
  set commands ""
  foreach item [getAllItems] {

      #if { [$item hasSavedState] } {
	#  $item setModified
      #}

    set command [$item cget -command]
    if {[lsearch $commands $command] < 0} {
      lappend commands $command
    }
  }
  eval $itk_component(type) insert list 0 [lsort -dictionary $_experimentTypes]
  eval $itk_component(command) insert list 0 [lsort -dictionary $commands]

  # read the grammars and machines
  eval $itk_component(grammarName) insert list end [.main getGrammarNames]
  eval $itk_component(machineName) insert list end "[.main getMachineNames] {}"

  _select "<none>"
}

## ----------------------------------------------------------------------------
## _select
## ----------------------------------------------------------------------------
body YadaExperiments::_select {{experimentName ""}} {

  # get a experiment name from the combobox if we didn't provide one
  if {$experimentName == ""} {
    set experimentName [$itk_component(name) get]
  } 
  set currentItem [getItem $experimentName]
  if {$currentItem == ""} {
    printMessage "Experiment `$experimentName' is unknown." error
    return
  }
  
  # check if we didn't commit changes
  if {![_askChange]} {
    return
  }


  # set the new active experiment
  setCurrentItem $currentItem

  ## update title bar
  displayTitle

  # update the name box
  set box $itk_component(name)
  set entries [[$box component list] get 0 end]
  set index [lsearch $entries $experimentName]
  if {$index < 0} {
    error "ERROR: experiment `$experimentName' not in combobox"
  }
  $box selection set $index $index
  $box delete entry 0 end
  $box insert entry 0 $experimentName

  # update the command box
  set box $itk_component(command)
  set command [$currentItem cget -command]
  set entries [[$box component list] get 0 end]
  set index [lsearch $entries $command]
  if {$index < 0} {
    error "ERROR: command `$command' not in combobox"
  }
  $box selection set $index $index
  $box delete entry 0 end
  $box insert entry 0 $command


  # update the boxes
  foreach aspect {type grammarName machineName} {
    set box $itk_component($aspect)
    set value [$currentItem cget -$aspect]
    set index [getIndexOfName [$box component list] $value]
    if {$index < 0} {
      error "ERROR: $aspect `$value' is unknown"
    }
    $box selection set $index $index
    set isEditable [$box cget -editable]
    if {!$isEditable} {
      $box configure -editable 1
    }
    $box delete entry 0 end
    $box insert entry 0 $value
    if {!$isEditable} {
      $box configure -editable 0
    }
  }

  # update the script
  set text $itk_component(script)
  $text delete 1.0 end
  $text insert 1.0 [$currentItem cget -script]

  # update the entries
  foreach entryName {dataDir} {
    set entry $itk_component($entryName) 
    $entry delete 0 end
    $entry insert 0 [$currentItem cget -$entryName]
  }

  # update the checkbuttons
  set _isActive [$currentItem cget -isActive]
  set _isComplete [$currentItem cget -isComplete]

  # remove marking
  _unmark

  # remember this selected name
  set _oldItem $experimentName
}

## ----------------------------------------------------------------------------
## _commit
## ----------------------------------------------------------------------------
body YadaExperiments::_commit {} {
  
  if {![chain]} {
    return
  }

  set currentItem [getCurrentItem]

  # set all aspects of the object
  foreach aspect {type grammarName command dataDir machineName} {
    $currentItem configure -$aspect [$itk_component($aspect) get]
  }

  # set the script property of the object
  set text $itk_component(script)
  $currentItem configure -script [$text get 1.0 end]

  # save the checkbuttons
  $currentItem configure -isActive $_isActive
  $currentItem configure -isComplete $_isComplete

  # save data unique in comboboxes
  foreach aspect {type grammarName command machineName} {
    set box $itk_component($aspect)
    set value [$currentItem cget -$aspect]
    set listBox [$box component list]
    if {[getIndexOfName $listBox $value] < 0} {
      set entries [$listBox get 0 end] 
      lappend entries $value
      $listBox delete 0 end
      eval $listBox insert 0 [lsort -dictionary $entries]
    }
  }

  [getCurrentItem] setModified

  # remember this selected name
  set _oldItem [$currentItem cget -name]

  # remove marking
  _unmark
  _select
}

## ----------------------------------------------------------------------------
## _new
## ----------------------------------------------------------------------------
body YadaExperiments::_new {} {

  # ask for a new name
  set prd [iwidgets::promptdialog .prd \
    -title "Yada - New experiment" \
    -modality application \
    -labeltext "Name:" \
    -labelpos w]

  set experimentName [$itk_component(name) get]
  set entry [$prd component prompt]
  $entry configure -width 40
  $entry insert 0 $experimentName

  $prd center .
  $prd hide Apply
  $prd hide Help
  after idle "focus [$prd component prompt]"

  set result [$prd activate]
  set experimentName [$prd get]
  destroy $prd
  
  if {!$result} {
    return
  }

  set item [getItem $experimentName]
  if {$item != ""} {
    set message "ERROR: experiment `$experimentName' exists already."
    .main printStatus $message
    printMessage $message error
    return [_new]
  }
  .main printStatus ""

  # build a new experiment
  set currentItem [getCurrentItem]
  setCurrentItem [$currentItem clone -name $experimentName]
  .main printStatus "New experiment `$experimentName'"

  # store all aspects
  set box $itk_component(name)
  $box delete entry 0 end
  $box insert entry 0 $experimentName
  _commit

  # display it 
  _select $experimentName
}

## ----------------------------------------------------------------------------
## _dataDirChooser
## ----------------------------------------------------------------------------
body YadaExperiments::_dataDirChooser {} {
  focus $itk_component(dataDir)

  set experimentName [$itk_component(name) get]

  set currentItem [getCurrentItem]
  set dataDir [$currentItem cget -dataDir]
  set directory [file dirname $dataDir]
  set selection [.fileSelector activate \
    -directory $directory  \
    -title "Choose a data directory" \
    -filter [list "*.log.gz *.xml.gz" "*log.gz" "*.xml.gz" "*"] \
    -selection $dataDir]
    #-mode "singledir" \

  if {$selection != ""} { 
    if {[llength $selection] > 1 || ![file isdirectory $selection]} {
      set message "ERROR: please select one directory."
      .main printStatus $message
      printMessage $message error
      return [_dataDirChooser]
    } else {
      .main printStatus ""
      $itk_component(dataDir) delete 0 end
      $itk_component(dataDir) insert 0 $selection
      _mark "dataDir"
    }
  }
}

## ----------------------------------------------------------------------------
## _delete
## ----------------------------------------------------------------------------
body YadaExperiments::_delete {} {

  # get the experiment name
  set currentItem [getCurrentItem]
  set experimentName [$currentItem cget -name]

  # experiment "<none>" is undeleteable
  if {$experimentName == "<none>"} {
    return
  }

  # are you sure?
  set msg [iwidgets::messagedialog .msg \
    -text "Are you sure that you want to delete experiment\n`$experimentName'?" \
    -title "Yada - Caution"  \
    -image [YadaImages::get question] \
    -modality application \
    -master $itk_interior]
  $msg buttonconfigure OK -text "Yes"
  $msg buttonconfigure Cancel -text "No"
  $msg center .
  set result [$msg activate]
  destroy $msg
  if {!$result} {
    return
  }

  # delete the data
  unsetItem $currentItem
  itcl::delete object $currentItem

  # update the names
  set box $itk_component(name)
  set index [getIndexOfName [$box component list] $experimentName]
  if {$index >= 0} {
    $box delete list $index $index
  }

  # show something
  _select "<none>"
}

## ----------------------------------------------------------------------------
## _defaults
## ----------------------------------------------------------------------------
body YadaExperiments::_defaults {{fileName ""}} {

  # possible variables
  set experimentTypes "netsearch arcconsistency pruning isearch frobbing {incremental completition} gls"

  if {$fileName == ""} {
    # built in default experiments or ...
    getYadaDefaultExperiments
  } else {
    # ... configured experiments
    if {[catch {source $fileName} errMsg]} {
      bgerror "There's an error in your experiments init file $fileName!"
    }
  }

  # required <none> experiment
  lappend experimentTypes "<none>"
  YadaExperiment ::#auto \
    -name "<none>" \
    -type "<none>" \
    -grammarName "<none>" \
    -machineName "<none>" \
    -command "<none>" \
    -dataDir "<none>"

  # assert experiments and types into the class
  set experiments [itcl_info objects -class YadaExperiment]
  set _experimentTypes [lsort -dictionary $experimentTypes]
  foreach experiment $experiments {
    set experimentName [$experiment cget -name]
    set isValid 1

    # validation checks:
    # (a) check if the experiment type is known
    # (b) check if experiment already exists
    # (c) check if the used grammar exists
    # (d) check if the experiments data store is unique

    # (a)
    set experimentType [$experiment cget -type]
    if {[lsearch $_experimentTypes $experimentType] < 0} {
      printMessage \
"Unknown experiment type $experimentType!
Known experiment types are `$_experimentTypes'
... ignoring experiment `$experimentName'." error
      set isValid 0
      itcl::delete object $experiment
    } 

    # (b)
    if {$isValid} {
      set item [getItem $experimentName]
      if {$item != ""} {
	printMessage "Experiment `$experimentName' already exists!" error
	set isValid 0
      } 
    }
     
    # (c)
    if {$isValid} {
      set grammarName [$experiment cget -grammarName]
      set grammar [.main getGrammar $grammarName]
      if {$grammar == ""} {
	printMessage \
"Grammar `$grammarName' is unknown!
... ignoring experiment `$experimentName'." error
	set isValid 0
      }
    }

    # (d)
    if {$isValid} {
      set experimentDataDir [$experiment cget -dataDir]
      if {[info exists dataDirs($experimentDataDir)]} {
	printMessage \
"`$experimentName' uses the same data directory as $dataDirs($experimentDataDir)!
... ignoring experiment `$experimentName'." error
	set isValid 0
      } else {
	set dataDirs($experimentDataDir) $experimentName
      }
    }

    if {$isValid} {
      setItem $experiment
    } else {
      itcl::delete object $experiment
    }
  }

  # just to be tidy
  array unset dataDirs
  .main printStatus "... built [llength $experiments] experiments"
}

## ----------------------------------------------------------------------------
## getActiveExperiments
## ----------------------------------------------------------------------------
body YadaExperiments::getActiveExperiments {args} {
  set result ""
  foreach experiment [eval getAllItems $args] {
    if {[$experiment cget -isActive]} {
      lappend result [$experiment cget -name]
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## _mark
## ----------------------------------------------------------------------------
body YadaExperiments::_mark {aspect} {

  switch $aspect {
    "script" {
      set currentItem [getCurrentItem]
      set oldData [string trim [$currentItem cget -$aspect]]
      set newData [string trim [$itk_component(script) get 1.0 end]]
    }
    "dataDir" {
      set currentItem [getCurrentItem]
      set oldData [glob -nocomplain [$currentItem cget -$aspect]]
      set newData [glob -nocomplain [$itk_component($aspect) get]]
    }
    "isActive" {
      set currentItem [getCurrentItem]
      set oldData [$currentItem cget -isActive]
      set newData $_isActive
      set aspect "status"
    }
    "isComplete" {
      set currentItem [getCurrentItem]
      set oldData [$currentItem cget -isComplete]
      set newData $_isComplete
      set aspect "status"
    }
    default {
      set oldData ""
      set newData ""
    }
  }

  chain $aspect $newData $oldData
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaExperiments::activationHandle {} {
  chain

  # refresh the grammars
  $itk_component(grammarName) delete list 0 end
  eval $itk_component(grammarName) insert list end [.main getGrammarNames]

  # refresh the machines
  $itk_component(machineName) delete list 0 end
  eval $itk_component(machineName) insert list end [.main getMachineNames]

  # redisplay the current selection
  _select
}

## ----------------------------------------------------------------------------
## _checkExperiment
## ----------------------------------------------------------------------------
body YadaExperiments::_checkExperiment {} {
  set currentItem [getCurrentItem]
  if {$currentItem == ""} {
    return
  }
  $currentItem check
  _select
}


## ----------------------------------------------------------------------------
## Overriding default implementation. Delegates command to current grammar's
## load method
## ----------------------------------------------------------------------------
body YadaExperiments::load {} {
    set currentItem [getCurrentItem]

    if { [$currentItem cget -name] != "<none>"} {
	[getCurrentItem] load
	_select
	displayTitle
    }
}

## ----------------------------------------------------------------------------
## Overriding default implementation. Delegates command to current grammar's
## save method
## ----------------------------------------------------------------------------
body YadaExperiments::save {} {
    set currentItem [getCurrentItem]

    if { [$currentItem cget -name] != "<none>"} {
	[getCurrentItem] save
    }
    displayTitle
}



