# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaGenerals - stuff that didn't fit anywhere else til now.
## This class just offers a gui for some settings that we'd like to
## access but no special gui like the YadaExperiments or the YadaGrammars are
## done yet. That is there is no YadaConfigItems that we deal with here. Therefore
## The YadaGenerals don't inherit YadaConfigDocument.
## \ingroup YadaConfigDocument
##
## \author Michael Daum
##
## $Id: YadaGenerals.tcl,v 1.13 2004/02/26 14:23:37 ddreyer Exp $
## ----------------------------------------------------------------------------
class YadaGenerals {
  inherit itk::Widget YadaPlainDocument


  # variables ----------------------------------------------------------------

  ## percentage indicating a significant difference in a quotient of two numbers.
  public variable significance  1.1

  ## tolerated time difference which isn't significant for a profiling distinction
  public variable timeTolerance 500

  ## the name of an editor to launch when we want to edit something
  public variable editor ""

  ## flag indicating how to compute weirdness in the YadaRunnableDocument s.
  public variable weirdness 1

  # methods ------------------------------------------------------------------
  public method init {}
  public method toDOM {}
  public method fromDOM {root}; # \type domNode
  public method displayTitle {}
  public method getPersistanceFileName {}; ## overriding virtual method
    public method activationHandle {}
  

  constructor {args} {}; ## \type TclList

  private method _commit {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaGenerals::constructor {args} {
  global env

  #
  # create custom file menu
  #
  set fileButton [.main component fileButton]
  for {
    set i 1
    set fileMenu $fileButton.menu$i
  } {[winfo exists $fileMenu]} {incr i} {
    set fileMenu $fileButton.menu$i
  }
  itk_component add fileMenu {
      menu $fileMenu
  } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  ## create submenu entry "New"
  itk_component add newMenu {
     menu $itk_component(fileMenu).newMenu 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  ## add file menu entries
  $itk_component(fileMenu) add command \
       	-label "Load" \
       	-underline 0 \
       	-command [code $this load] 
  $itk_component(fileMenu) add command \
       	-label "Save" \
       	-underline 0 \
       	-command [code $this save] 
  $itk_component(fileMenu) add cascade \
	-label "New" \
	-underline 0 \
	-menu $itk_component(newMenu)
  $itk_component(fileMenu) add separator
  $itk_component(fileMenu) add command \
	-label "Quit" \
	-underline 0 \
	-command {.main quit}


  $itk_component(newMenu) add command \
	-label "Runner" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaRunner Runner}
  $itk_component(newMenu) add command \
	-label "Difference" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaDifference Difference}
  $itk_component(newMenu) add command \
	-label "Gls Statistics" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaGlsStats GlsStatistics}
  $itk_component(newMenu) add command \
	-label "Ranking" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaRanking Ranking}


  set name "Generals"

  itk_component add frame {
    frame $itk_interior.frame \
      -borderwidth 0
  }
  
  itk_component add significanceLabel {
    label $itk_component(frame).significanceLabel \
      -text "Significance:" \
      -anchor nw
  }

  itk_component add significance {
    entry $itk_component(frame).significance \
      -textvariable [scope significance] \
      -borderwidth 2 \
      -state normal \
      -width 20 \
  } {}

  itk_component add timeToleranceLabel {
    label $itk_component(frame).timeToleranceLabel \
      -text "Time Tolerance:" \
      -anchor nw
  }

  itk_component add timeTolerance {
    entry $itk_component(frame).timeTolerance \
      -textvariable [scope timeTolerance] \
      -borderwidth 2 \
      -state normal \
      -width 20 \
  } {}

  itk_component add editorLabel {
    label $itk_component(frame).editorLabel \
      -text "Editor:" \
      -anchor nw
  }

  itk_component add editor {
    entry $itk_component(frame).editor \
      -textvariable [scope editor] \
      -borderwidth 2 \
      -state normal \
      -width 20 \
  } {}

  itk_component add weirdnessLabel {
    label $itk_component(frame).weirdnessLabel \
      -text "Weirdness:" -anchor nw
  } {}

  itk_component add weirdness {
    checkbutton $itk_component(frame).strictWeirdness \
      -text "strict" \
      -variable [scope weirdness] \
      -command [code $this _commit]
  } {}


  # binding
  bind $itk_component(timeTolerance) <Return> [code $this _commit]
  bind $itk_component(significance) <Return> [code $this _commit]
  bind $itk_component(editor) <Return> [code $this _commit]

  # packing + gridding
  pack $itk_component(frame) -fill both -expand 1 -padx 10 -pady 10
  grid columnconfigure $itk_component(frame) 0 -pad 10
  grid columnconfigure $itk_component(frame) 1 -weight 1
  grid rowconfigure $itk_component(frame) 4 -weight 1

  grid $itk_component(editorLabel) -sticky nw -row 0 -column 0 -pady 10 
  grid $itk_component(editor) -sticky nw -row 0 -column 1 -pady 10 

  grid $itk_component(significanceLabel) -sticky ew -row 1 -column 0 -pady 10
  grid $itk_component(significance) -sticky nw -row 1 -column 1 -pady 10 

  grid $itk_component(timeToleranceLabel) -sticky ew -row 2 -column 0 -pady 10
  grid $itk_component(timeTolerance) -sticky nw -row 2 -column 1 -pady 10 

  grid $itk_component(weirdnessLabel) -sticky ew -row 3 -column 0 -pady 10
  grid $itk_component(weirdness) -sticky nw -row 3 -column 1 -pady 10 


  eval itk_initialize $args
  
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaGenerals::init {} {
    if {$_isInitialized} {
	return
    }
    global env
    chain
    setModified
    configure -editor $editor; #$env(YADA_EDITOR)

  
}

## ----------------------------------------------------------------------------
## _commit
## ----------------------------------------------------------------------------
body YadaGenerals::_commit {} {
  configure \
    -significance $significance \
    -timeTolerance $timeTolerance \
    -editor $editor \
    -weirdness $weirdness

    setModified
}


## ----------------------------------------------------------------------------
## Serializes Object into a DOM-XML-Node
## @return root DOM node representing the YadaGenerals-Object (root tag <yada>)
## ----------------------------------------------------------------------------
body YadaGenerals::toDOM {} {
    set document [dom createDocument "yada"]
    set root [$document documentElement]

    set nodeGenerals [$document createElement "yadaGenerals"]

    $nodeGenerals setAttribute "significance" $significance

    $nodeGenerals setAttribute "timeTolerance" $timeTolerance

    $nodeGenerals setAttribute "editor" $editor

    $nodeGenerals setAttribute "weirdness" $weirdness

    $root appendChild $nodeGenerals


    return $root
}



## ----------------------------------------------------------------------------
## Initialize Object with DOM node (that has saved the object state)
## @param rootNode root DOM-Node (Tag <yada>)
## ----------------------------------------------------------------------------
body YadaGenerals::fromDOM {rootNode} {
    set xpath "/yada/yadaGenerals"

    set nodeGenerals [$rootNode selectNodes $xpath]
    set significance [$nodeGenerals getAttribute significance]
    

    set timeTolerance [$nodeGenerals getAttribute timeTolerance]
    set editor [$nodeGenerals getAttribute editor]
    set weirdness [$nodeGenerals getAttribute weirdness]

 
}


## ----------------------------------------------------------------------------
## Update toolbar, menubar and titlebar
## ----------------------------------------------------------------------------
body YadaGenerals::displayTitle {} {
    chain

    set loadButton [[[.main getDocument Configure] component toolbar] component loadButton]
    set saveButton [[[.main getDocument Configure] component toolbar] component saveButton]
   

    if { [hasSavedState] } {
	$itk_component(fileMenu) entryconfigure 0 -state normal
	$loadButton configure -state normal
    } else {
	$itk_component(fileMenu) entryconfigure 0 -state disabled
	$loadButton configure -state disabled
    }

    if { [isModified] } {
	$itk_component(fileMenu) entryconfigure 1 -state normal   
	$saveButton configure -state normal
    } else {
	$itk_component(fileMenu) entryconfigure 1 -state disabled
	$saveButton configure -state disabled
    }
}

## ----------------------------------------------------------------------------
## Getting unique file name from registry
## ----------------------------------------------------------------------------
body YadaGenerals::getPersistanceFileName {} {
    return [.main getPersistanceFileNameForMisc]
}

body YadaGenerals::activationHandle {} {
    chain
    ## activate custom file menu
    set fileButton [.main component fileButton]
    $fileButton configure -menu $itk_component(fileMenu)
}


