# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaRanking - load sets of results.
## \ingroup YadaDocuments
##
## \author Michael Daum
##
## $Id: YadaRanking.tcl,v 1.16 2004/02/25 14:42:10 micha Exp $
## ----------------------------------------------------------------------------
class YadaRanking {
  inherit YadaDocument

  # variables ----------------------------------------------------------------
  private variable _isRunning 0
  private variable _gotBreak 0
  private variable _typeOfColumn; ## \type TclArray
  private variable _rankingFormula ""
  private variable _experimentFilters {}
  private variable _followDataFlag 1
  private variable _lastSorting 1
  private variable _rankingOf ""
  private variable _totalTime 0
  private variable _totalExperiments 0

  # methods ------------------------------------------------------------------
  public method init {}
  public method load {args}; ## \type TclList
  public method reload {}
  # public method save {}
  public method clear {}
  public method activationHandle {}
  public method breakProcess {}
  public method unbreakProcess {}
  public method interruptProcess {}
  public method sortRows {colNo args}; ## \type TclNumber, TclList
  public method resortRows {}
  public method toggleFollowData {}
  public method isFollowData {}

  constructor {args} {}; ## \type TclList

  private method _selectGrammar {{grammarName ""}}; ## \type TclString
  private method _getActiveWork {}
  private method _getSelectedWork {}
  private method _getUndoneWork {}
  private method _getDoneWork {}
  private method _getWork {}
  private method _analyseDocument {document experiment wordgraph}; ## \type DomDocument, YadaExperiment, TclString
  private method _breakButtonAction {}
  private method _leaveTableCommand {}
  private method _escapeCommand {}
  private method _returnCommand {}
  private method _getRowOfExperimentName {experimentName}; ## \type TclString
  private method _insertData {experiment}; ## \type YadaExperiment
  private method _selectRanking {{formula ""}}; ## \type TclString
  private method _selectRankingCommand {}
  private method _selectFilter {{filters ""}}; ## \type TclString
  private method _selectFilterCommand {}
  private method _normalizeRankings {}
  private method _getRankingOfExperiment {experimentName}; ## \type TclString
  private method _setRankingOfExperiment {experimentName ranking}; ## \type TclString, TclNumber
  private method _unsetRankingOfExperiment {experimentName}; ## \type TclString
  private method _rowTagCommand {row}; ## \type TclNumber
  private method _computeStatistics {}
  private method _showStatistics {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaRanking::constructor {args} {

  # edit menu
  $itk_component(editMenu) add command \
       	-label "Clear" \
       	-underline 0 \
       	-command [code $this clear]
  $itk_component(editMenu) add command \
       	-label "Break" \
       	-underline 0 \
       	-command [code $this breakProcess]
  $itk_component(editMenu) add command \
       	-label "Interrupt" \
       	-underline 0 \
       	-command [code $this interruptProcess]
  
  # view menu
  itk_component add sortIncMenu {
     menu $itk_component(viewMenu).sortIncMenu 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  itk_component add sortDecMenu {
     menu $itk_component(viewMenu).sortDecMenu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }
  $itk_component(viewMenu) add separator
  $itk_component(viewMenu) add cascade \
	-label "Sort inc. by" \
	-underline 0 \
	-menu $itk_component(sortIncMenu)
  $itk_component(viewMenu) add cascade \
	-label "Sort dec. by" \
	-underline 0 \
	-menu $itk_component(sortDecMenu)

  # set types of columns
  array set _typeOfColumn {
    0 "dictionary"
    1 "real"
    2 "real"
    3 "real"
    4 "integer"
    5 "integer"
    6 "integer"
  }

  # toolbar
  $itk_component(toolbar) add FlatButton reloadButton \
    -image [YadaImages::get reload] \
    -balloonstr "reload" \
    -command [code $this reload]

  $itk_component(toolbar) add FlatButton clearButton \
	-image [YadaImages::get erase] \
	-command [code $this clear] \
	-balloonstr "clear"

  $itk_component(toolbar) add FlatButton breakButton \
	-image [YadaImages::get giveup] \
	-command [code $this _breakButtonAction] \
	-balloonstr "break"

  $itk_component(toolbar) add FlatButton term \
	-image [YadaImages::get stop] \
	-command [code $this interruptProcess] \
	-balloonstr "interrupt"


  # heading space
  itk_component add heading {
    frame $itk_component(childsite).heading
  }

  # the grammar
  itk_component add grammarBoxLabel {
    label $itk_component(heading).grammarBoxLabel \
	-text "Grammar:" \
	-anchor w
  }  {}

  itk_component add grammarBox {
    iwidgets::combobox $itk_component(heading).grammarBox \
	-completion false \
	-borderwidth 2 \
	-margin 4 \
	-selectioncommand [code $this _selectGrammar]
  }  {}

  # the total time
  itk_component add totalTimeLabel {
    label $itk_component(heading).totalTimeLabel \
	-text "Time:" \
	-anchor w
  }  {}

  itk_component add totalTime {
    entry $itk_component(heading).totalTime \
      -borderwidth 2 \
      -state disabled \
      -width 0
  }  {}

  # the total number of experiments
  itk_component add totalLabel {
    label $itk_component(heading).totalLabel \
	-text "Total:" \
	-anchor w
  }  {}

  itk_component add totalExperiments {
    entry $itk_component(heading).totalExperiments \
      -borderwidth 2 \
      -state disabled \
      -textvariable [scope _totalExperiments] \
      -width 0
  }  {}


  # the sorting
  itk_component add sortingBoxLabel {
    label $itk_component(heading).sortingBoxLabel \
	-text "Sort by:" \
	-anchor w \
  }  {}

  itk_component add sortingBox {
    iwidgets::combobox $itk_component(heading).sortingBox \
	-completion false \
	-borderwidth 2 \
	-margin 4 \
	-selectioncommand [code $this _selectRankingCommand]
  }  {}
  $itk_component(sortingBox) insert entry 0 $_rankingFormula

  # the experiment filter
  itk_component add experimentFilterLabel {
    label $itk_component(heading).experimentFilterLabel \
	-text "Experiment Filter:" \
	-anchor w \
  }  {}

  itk_component add experimentFilterBox {
    iwidgets::combobox $itk_component(heading).experimentFilterBox \
	-completion false \
	-borderwidth 2 \
	-margin 4 \
	-selectioncommand [code $this _selectFilterCommand]
  }  {}
  $itk_component(experimentFilterBox) insert entry 0 $_experimentFilters


  # the table
  itk_component add pane {
      iwidgets::panedwindow $itk_component(childsite).pane \
          -orient horizontal \
	  -sashindent -8 \
	  -thickness 2 \
	  -sashheight 8 \
	  -sashwidth 18
  } {
      keep -background -cursor
  }
  $itk_component(pane) add top -margin 0
  set top [$itk_component(pane) childsite top]

  itk_component add table {
     MyTable $top.table \
       	-exportselection 1 \
	-selecttitles 1 \
	-rowseparator "\n" \
	-colseparator "\t" \
	-rows 300 \
       	-cols 8 \
       	-state normal \
       	-cellanchor w \
       	-colstretchmode last \
       	-rowstretchmode none \
       	-selectmode extended \
       	-selecttype row \
       	-titlerows 1 \
       	-titlecols 0 \
	-borderwidth 3 \
	-outerborderwidth 2 \
	-outerrelief sunk \
	-selforeground "" \
	-selbackground "" \
	-font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} \
	-hscrollmode dynamic \
	-rowtagcommand [code $this _rowTagCommand]
  } { 
     keep -background -cursor 
  }
  
  # bindings
  set hull $itk_component(hull)
  set tableInTable [$itk_component(table) component table]
  bind $hull <Leave> [code $this _leaveTableCommand]
  bind $hull <FocusOut> [code $this _leaveTableCommand]
  bind $tableInTable <KeyPress-Escape> [code "$this _escapeCommand; break"]
  bind $tableInTable <KeyPress-Return> [code "$this _returnCommand; break"]
  set entry [$itk_component(sortingBox) component entry]
  bind $entry <KeyPress-Return> [code $this _selectRankingCommand]
  set entry [$itk_component(experimentFilterBox) component entry]
  bind $entry <KeyPress-Return> [code $this _selectFilterCommand]
  
  # packing + gridding
  pack $itk_component(heading) -fill x -side top -anchor n
  pack $itk_component(pane) -fill both -expand 1 -padx 2 -pady 5 -side top
  pack $itk_component(table) -fill both -expand 1 -anchor n

  grid columnconfigure $itk_component(heading) 1 -weight 1
  grid columnconfigure $itk_component(heading) 3 -weight 1
  grid columnconfigure $itk_component(heading) 5 -weight 1
  grid columnconfigure $itk_component(heading) 6 -weight 1
  
  grid $itk_component(grammarBoxLabel) -row 0 -column 0 -sticky ew -pady 2
  grid $itk_component(grammarBox)      -row 0 -column 1 -sticky ew -pady 2 -columnspan 3
  grid $itk_component(sortingBoxLabel) -row 0 -column 4 -sticky ew -pady 2 
  grid $itk_component(sortingBox)      -row 0 -column 5 -sticky ew -pady 2 -columnspan 2

  grid $itk_component(totalLabel)      -row 1 -column 0 -sticky ew -pady 2
  grid $itk_component(totalExperiments)    -row 1 -column 1 -sticky ew -pady 2
  grid $itk_component(totalTimeLabel)  -row 1 -column 2 -sticky ew -pady 2
  grid $itk_component(totalTime)       -row 1 -column 3 -sticky ew -pady 2
  grid $itk_component(experimentFilterLabel) -row 1 -column 4 -sticky ew -pady 2
  grid $itk_component(experimentFilterBox)   -row 1 -column 5 -sticky ew -pady 2 -columnspan 2

  eval itk_initialize $args
  .main registerDocument $this
}

## ----------------------------------------------------------------------------
## clear
## ----------------------------------------------------------------------------
body YadaRanking::clear {} {
  set table $itk_component(table)
  set selection [$table curselection]
  set cmax [$table cget -cols]

  # only delete the seleted rows
  if {$selection != ""} {
    set minRow 9999
    set maxRow 0

    foreach index $selection {
      scan $index "%d,%d" row col
      $table erase $row,0 $row,$cmax
      $table clear tags $row,0 $row,$cmax
      
      # delete ranking information
      set experimentName [$table getCell $row,0]
      _unsetRankingOfExperiment $experimentName
    }
  } else {
  
    # reset the whole statistics
    $table clear tags 
    $table delete rows -holddimensions 1 [$table cget -rows]
    set _rankingOf ""
    _getActiveWork
  }

  # reset the table 
  $table setCell \
       	0,0 "experiment" \
	0,1 "ranking" \
       	0,2 "struct.recall" \
       	0,3 "label.recall" \
       	0,4 "lex.recall" \
       	0,5 "time" \
	0,6 "sol.time" \
       	0,7 "soft time" \

  $table tag config title -font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*}
  $table tag config number -anchor e
  $table tag config sel -fg "" -bg "" -relief raised -bd 1
  $table tag config active -anchor w -fg black -bg gray -relief sunk -bd 1
  $table tag config colored -bg gray80
  $table tag config runner -fg red3 -bg #a3a3a3
  $table tag raise runner title
  $table tag col number 1 2 3 4 5 6
  $table width 0 45 1 10 2 9 3 9 4 9 5 25 6 25 7 25

  .main printStatus ""
  _computeStatistics
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaRanking::init {} {
  if {$_isInitialized} {
    return
  }
  chain
  clear
  _selectRanking {$lexicalRecall / (0.7 * log($time) + 0.2 * log($soltime) + 0.01 * log($softtime))}
  _selectFilter {*}

  $itk_component(sortIncMenu) delete 0 end
  $itk_component(sortDecMenu) delete 0 end

  set cmax [$itk_component(table) cget -cols]
  set headings [$itk_component(table) getCell 0,0 0,$cmax]
  set colNo 0
  foreach heading $headings {
    if {$heading != ""} {
      $itk_component(sortIncMenu) add command \
	-label "$heading" \
	-command [code $this sortRows $colNo -increasing]
      $itk_component(sortDecMenu) add command \
	-label "$heading" \
	-command [code $this sortRows $colNo -decreasing]
    }
    incr colNo
  }
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaRanking::activationHandle {} {
  chain

  # get the previous selection
  set box $itk_component(grammarBox)
  set grammarName [$box get]

  # default grammar to choose
  if {$grammarName == ""} {
    set grammarName "<none>"
  }

  $box clear list 
  eval $box insert list end [.main getGrammarNames]
  _selectGrammar $grammarName
}

## ----------------------------------------------------------------------------
## _selectGrammar
## ----------------------------------------------------------------------------
body YadaRanking::_selectGrammar {{grammarName ""}} {
  set box $itk_component(grammarBox)
  if {$grammarName == ""} {
    set grammarName [$box get]
    if {$grammarName == ""} {
      set grammarName "<none>"
    }
  } 

  $box clear entry
  $box insert entry 0 $grammarName
  set listOfEntries [[$box component list] get 0 end]
  set index [lsearch $listOfEntries $grammarName]

  if {$index < 0} {
    _selectGrammar "<none>"
  } else {
    $box selection set $index $index
  }

  resortRows
}

## ----------------------------------------------------------------------------
## reload
## ----------------------------------------------------------------------------
body YadaRanking::reload {} {
  clear
  load
}


## ----------------------------------------------------------------------------
## _getActiveWork
## ----------------------------------------------------------------------------
body YadaRanking::_getActiveWork {} {

  # get the active set of experiments
  set experimentNames [_getSelectedWork]
  if {$experimentNames == ""} {
    set experimentNames [_getUndoneWork]
    if {$experimentNames == ""} {
      set experimentNames [_getWork]
      set doneWork [_getDoneWork]
      if {$experimentNames == $doneWork} {
	return ""
      }

      # FIXME: do we want to set the table here? dirty!
      set rmax [llength $experimentNames]
      set _totalExperiments $rmax
      set table $itk_component(table)
      if {$rmax < 20} {
	set rmax 20
      } else {
	incr rmax
      }
      $table configure -rows $rmax
      set row 1
      foreach experimentName $experimentNames {
	$table setCell $row,0 $experimentName
	incr row
      }
    }
  }


  return $experimentNames
}

## ----------------------------------------------------------------------------
## _getSelectedWork
## ----------------------------------------------------------------------------
body YadaRanking::_getSelectedWork {} {
  set table $itk_component(table)

  set cells [$table curselection]
  
  foreach index $cells {
    scan $index "%d,%d" row col
    set experimentName [$table getCell $row,0]
    if {$experimentName != ""} {
      set id($row,0) $experimentName
    }
  }

  set experimentNames ""
  foreach index [array names id] {
    lappend experimentNames $id($index)
  }


  return [lsort -dictionary $experimentNames]
}

## ----------------------------------------------------------------------------
## _getDoneWork
## ----------------------------------------------------------------------------
body YadaRanking::_getDoneWork {} {

  set table $itk_component(table)
  set rmax [$table cget -rows]

  set result ""
  foreach {experimentName ranking} [$table getCell 1,0 $rmax,1] {
    if {$experimentName != "" && $ranking != ""} {
      lappend result $experimentName
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## _getUndoneWork
## ----------------------------------------------------------------------------
body YadaRanking::_getUndoneWork {} {

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set experimentNames ""
  foreach {experimentName ranking} [$table getCell 1,0 $rmax,1] {
    if {$experimentName != "" && $ranking == ""} {
      lappend experimentNames $experimentName
    }
  }

  return $experimentNames
}

## ----------------------------------------------------------------------------
## _getWork
## ----------------------------------------------------------------------------
body YadaRanking::_getWork {} {

  set grammarName [$itk_component(grammarBox) get]

  # get all experiments using this grammar
  set experimentNames ""
  foreach experiment [.main getExperiments] {
    set experimentName [$experiment cget -name]
    if {$experimentName == "<none>" || 
	![regexp $grammarName [$experiment cget -grammarName]]} {
      continue
    }

    set found 0
    foreach filter $_experimentFilters {
      if {[string match $filter $experimentName]} {
	set found 1
	break
      }
    } 

    if {!$found} {
      continue
    }
    
    if {![$experiment cget -isComplete]} {
      .main printStatus "... checking experiment $experimentName"
      $experiment check
    }

    if {![$experiment cget -isComplete]} {
      continue
    }

    lappend experimentNames $experimentName
  }

  return $experimentNames
}

## ----------------------------------------------------------------------------
## load
## ----------------------------------------------------------------------------
body YadaRanking::load {args} {

  set _isRunning 1
  set noLoadedFiles 0
  set noExperiments 0
  set errorFiles ""
  set noErrors 0

  # load all experiments
  foreach experimentName [_getActiveWork] {

    set experiment [.main getExperiment $experimentName]
    set grammar [.main getGrammar [$experiment cget -grammarName]]
    .main printStatus "... loading experiment `$experimentName'"

    # did we already collect the sums
    if {![$experiment cget -isUpToDate]} {

      # reset experiment statistics
      $experiment resetStatistics

      # load all wordgraphs
      foreach wordgraph [$grammar getSelection] {
	
	# check if the file is present
	set file [$experiment getXmlFileName $wordgraph]
	if {![file exists $file]} {
	  continue
	}

	.main printStatus "... loading experiment `$experimentName', wordgraph `$wordgraph', $noErrors errors"
	update 

	# analyse the file
	set fd [open "|zcat $file" r]
	set data [read $fd]
	close $fd
	set document [dom parse $data]
	set isOk [_analyseDocument $document $experiment $wordgraph]
	if {!$isOk} {
	  incr noErrors
	  lappend errorFiles $file
	}

	$document delete
	set document ""

	incr noLoadedFiles
      }
      $experiment configure -isUpToDate 1
    }

    # insert the data into the table
    _insertData $experiment

    # did we press the break button
    if {$_gotBreak} {
      unbreakProcess
      .main printStatus "WARNING: interrupted loading after $noLoadedFiles files"
      break
    }

    incr noExperiments

  }

  # if we had errors in some files the notify the user
  if {[llength $errorFiles] > 0} {
    set listDialog [ListDialog .#auto \
      -title "Yada - Message" \
      -labeltext "The following files had errors:" \
      -modality application \
      -height 480 \
      -width 640 \
      -master .
    ]
    eval $listDialog insert list 0 $errorFiles
    $listDialog activate
    itcl::delete object $listDialog
  }

  # we tried to load nothing ...
  if {$noExperiments == 0} {
    printMessage "No experiments found!"
  }
  
  # update the display
  _normalizeRankings
  resortRows
  _computeStatistics
  .main printStatus "INFO: loaded $noLoadedFiles files"
  set _isRunning 0
}

## ----------------------------------------------------------------------------
## _breakButtonAction
## ----------------------------------------------------------------------------
body YadaRanking::_breakButtonAction {} {
  if {$_gotBreak} {
    unbreakProcess
  } else {
    breakProcess
  }
}

## ----------------------------------------------------------------------------
## breakProcess
## ----------------------------------------------------------------------------
body YadaRanking::breakProcess {} {
  .main printStatus "INFO: break received"
  if {$_isRunning} {
    set _gotBreak 1
    $itk_component(toolbar) itemconfigure breakButton -background red3
  }
}

## ----------------------------------------------------------------------------
## unbreakProcess
## ----------------------------------------------------------------------------
body YadaRanking::unbreakProcess {} {
  set _gotBreak 0
  $itk_component(toolbar) itemconfigure breakButton -background gray90
}

## ----------------------------------------------------------------------------
## _leaveTableCommand
## ----------------------------------------------------------------------------
body YadaRanking::_leaveTableCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  if {$lastActiveCell == ""} {
    return
  }

  scan $lastActiveCell "%d,%d" row col
  $table activate -1,-1
}

## ----------------------------------------------------------------------------
## _escapeCommand
## ----------------------------------------------------------------------------
body YadaRanking::_escapeCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  
  if {$col == 0} {
    $table selection clear all
  } else {
    $table activate -1,-1
  }
}

## ----------------------------------------------------------------------------
## interruptProcess
## ----------------------------------------------------------------------------
body YadaRanking::interruptProcess {} {
  breakProcess
}

## ----------------------------------------------------------------------------
## _analyseDocument
## ----------------------------------------------------------------------------
body YadaRanking::_analyseDocument {document experiment wordgraph} {

  # default values
  set lexicalHits 0
  set structHits 0
  set labelledHits 0
  set tries 0
  set score 0.0
  set totalTime 0
  set solutionTime 0
  set softTime 0
  set isOk 1

  # get the cdgp logfile root element
  set cdgpNode [$document getElement "cdgp"]
  if {$cdgpNode == ""} {
    printMessage "ERROR: no cdgp logfile" error
    set isOk 0
  }  else {
    # get the parse verification node
    set parseVerificationNode [$cdgpNode descendant 1 "parseverification"]
    if {$parseVerificationNode != ""} {
      foreach aspect {tries structHits labelledHits lexicalHits } {
	set $aspect [$parseVerificationNode selectNodes \
	  descendant::[string tolower $aspect]/number\[@name="SYN"\]/@value]
      }
      # parse the statistics 
      foreach aspect {score totalTime solutionTime softTime} {
	set $aspect [$cdgpNode selectNodes \
	  descendant::statistics/number\[@name="$aspect"\]/@value ]
	if {[subst \$$aspect] == ""} {
	  set $aspect 0.0
	}
      }
    } else {
      set isOk 0
    }
  }
  # end of analysis

  # verification fallback: compute tries from annotation
  if {!$isOk} {
    .main printStatus "ERROR: error in logfile of $wordgraph, consulting annotation"
    set grammar [.main getGrammarOfExperiment [$experiment cget -name]]
    $grammar scan; # CAUTION: might be costy
    set yadaWordgraph [$grammar getWordgraph $wordgraph]
    set tries [$yadaWordgraph cget -noAnnoEdges]
    if {$tries == 0} {
      .main printStatus "ERROR: no correct annotation for wordgraph '$wordgraph'"
    }
  }

  # fix solution and soft time in case of no score
  if {$solutionTime == 0} {
    set solutionTime $totalTime
  }
  if {$softTime == 0} {
    set softTime $solutionTime
  }

  # add the collected data to the experiment
  set sumScore [expr [$experiment cget -totalScore] + $score]
  set sumLabelledHits [expr [$experiment cget -labelledHits] + $labelledHits]
  set sumStructHits [expr [$experiment cget -structHits] + $structHits]
  set sumLexicalHits [expr [$experiment cget -lexicalHits] + $lexicalHits]
  set sumTime [expr [$experiment cget -totalTime] + $totalTime]
  set sumSolTime [expr [$experiment cget -totalSolTime] + $solutionTime]
  set sumSoftTime [expr [$experiment cget -totalSoftTime] + $softTime]
  set sumTries [expr [$experiment cget -totalTries] + $tries]

  $experiment configure \
    -totalScore $sumScore \
    -totalTries $sumTries \
    -totalTime $sumTime \
    -totalSolTime $sumSolTime \
    -totalSoftTime $sumSoftTime \
    -structHits $sumStructHits \
    -labelledHits $sumLabelledHits \
    -lexicalHits $sumLexicalHits \
    -totalTries $sumTries

  return $isOk
}

## ----------------------------------------------------------------------------
## _insertData
## ----------------------------------------------------------------------------
body YadaRanking::_insertData {experiment} {

  set score [expr double([$experiment cget -totalScore])]
  set lexicalHits [$experiment cget -lexicalHits]
  set labelledHits [$experiment cget -labelledHits]
  set structHits [$experiment cget -structHits]
  set tries [$experiment cget -totalTries]

  set lexicalRecall [expr $lexicalHits * 100.0 / $tries]
  set labelledRecall [expr $labelledHits * 100.0 / $tries]
  set structRecall [expr $structHits * 100.0 / $tries ]
  set time [expr double([$experiment cget -totalTime])]
  set soltime [expr double([$experiment cget -totalSolTime])]
  set softtime [expr double([$experiment cget -totalSoftTime])]
  set experimentName [$experiment cget -name]
  set ranking [expr $_rankingFormula]
  _setRankingOfExperiment $experimentName $ranking

  # clear it
  set table $itk_component(table)
  set cmax [$table cget -cols]
  set row [_getRowOfExperimentName $experimentName]
  $table erase $row,1 $row,$cmax

  # fill it
  $table setCell \
    $row,1 [format "%4.3g" $ranking] \
    $row,2 [format "%.2f" $structRecall] \
    $row,3 [format "%.2f" $labelledRecall] \
    $row,4 [format "%.2f" $lexicalRecall] \
    $row,5 [timeToString $time] \
    $row,6 [timeToString $soltime] \
    $row,7 [timeToString $softtime] 

#    $row,4 [expr int($time)] \
#    $row,5 [expr int($soltime)] \
#    $row,6 [expr int($softtime)] 

  # show it
  if {$_followDataFlag} {
    $table see $row,1
  }
}

## ----------------------------------------------------------------------------
## _getRowOfExperimentName
## ----------------------------------------------------------------------------
body YadaRanking::_getRowOfExperimentName {experimentName} {

  set table $itk_component(table)
  set rmax [$itk_component(table) cget -rows]
  set experimentNames [$itk_component(table) getCell 0,0 $rmax,0]
  set row [lsearch -exact $experimentNames $experimentName]

  return $row
}

## ----------------------------------------------------------------------------
## _returnCommand
## ----------------------------------------------------------------------------
body YadaRanking::_returnCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  set colMax [expr [$table cget -cols] -1]
  if {$col == $colMax} {
    incr row
    set col 1
  } else {
    incr col
  }

  $table activate $row,$col
  $table see $row,$col
}

## ----------------------------------------------------------------------------
## sortRows
## ----------------------------------------------------------------------------
body YadaRanking::sortRows {colNo args} {
  set type $_typeOfColumn($colNo)
  eval $itk_component(table) sortRows $colNo -$type $args
  $itk_component(table) see 1,1
  set _lastSorting $colNo
  eval lappend _lastSorting $args
}

## ----------------------------------------------------------------------------
## _selectRankingCommand
## ----------------------------------------------------------------------------
body YadaRanking::_selectRankingCommand {} {
  _selectRanking
  reload
  resortRows
}

## ----------------------------------------------------------------------------
## _selectRanking
## ----------------------------------------------------------------------------
body YadaRanking::_selectRanking {{formula ""}} {

  set box $itk_component(sortingBox)

  if {$formula == ""} {
    set _rankingFormula [$box get]
  } else {
    set _rankingFormula $formula
  }

  set entries [[$box component list] get 0 end]
  set index [lsearch $entries $_rankingFormula]
  if {$index < 0} {
    $box insert list 0 $_rankingFormula
  }
  $box delete entry 0 end 
  $box insert entry 0 $_rankingFormula
}

## ----------------------------------------------------------------------------
## _selectFilterCommand
## ----------------------------------------------------------------------------
body YadaRanking::_selectFilterCommand {} {
  _selectFilter
  clear
}

## ----------------------------------------------------------------------------
## _selectFilter
## ----------------------------------------------------------------------------
body YadaRanking::_selectFilter {{filters ""}} {
    
  set box $itk_component(experimentFilterBox)

  if {$filters == ""} {
    set _experimentFilters [$box get]
  } else {
    set _experimentFilters $filters
  }

  set entries [[$box component list] get 0 end]
  set index [lsearch $entries $_experimentFilters]
  if {$index < 0} {
    $box insert list 0 $_experimentFilters
  }
  $box delete entry 0 end 
  $box insert entry 0 $_experimentFilters
}

## ----------------------------------------------------------------------------
## isFollowData
## ----------------------------------------------------------------------------
body YadaRanking::isFollowData {} {
  return $_followDataFlag
}

## ----------------------------------------------------------------------------
## toggleFollowData
## ----------------------------------------------------------------------------
body YadaRanking::toggleFollowData {} {
  if {$_followDataFlag == 1} {
    set _followDataFlag 0
  } else {
    set _followDataFlag 1
  }

  return $_followDataFlag
}

## ----------------------------------------------------------------------------
## _normalizeRankings
## ----------------------------------------------------------------------------
body YadaRanking::_normalizeRankings {} {
  .main printStatus "... normalizing rankings"

  set sortedRanking [join [lsort -index 1 -decreasing -real $_rankingOf]]
  set table $itk_component(table)
  set rmax [$itk_component(table) cget -rows]
  set experimentNames [$itk_component(table) getCell 0,0 $rmax,0]

  set normalizedRanking 0
  set lastRanking ""
  foreach {experimentName ranking} $sortedRanking {
    if {$ranking != $lastRanking} {
      incr normalizedRanking
    }
    set row [lsearch -exact $experimentNames $experimentName]
    $table setCell $row,1 $normalizedRanking
    set lastRanking $ranking
  }

  .main printStatus ""
}

## ----------------------------------------------------------------------------
## _getRankingOfExperiment
## ----------------------------------------------------------------------------
body YadaRanking::_getRankingOfExperiment {experimentName} {
  foreach {name ranking} $_rankingOf {
    if {$name == $experimentName} {
      return $ranking
    }
  }
}

## ----------------------------------------------------------------------------
## _setRankingOfExperiment
## ----------------------------------------------------------------------------
body YadaRanking::_setRankingOfExperiment {experimentName ranking} {
  _unsetRankingOfExperiment $experimentName
  lappend _rankingOf [list $experimentName $ranking]
}

## ----------------------------------------------------------------------------
## _unsetRankingOfExperiment
## ----------------------------------------------------------------------------
body YadaRanking::_unsetRankingOfExperiment {experimentName} {
  set index 0
  foreach {name ranking} $_rankingOf {
    if {$name == $experimentName} {
      set _rankingOf [lreplace $_rankingOf $index $index]
    }
    incr index
  }

  return
}

## ----------------------------------------------------------------------------
## resortRows
## ----------------------------------------------------------------------------
body YadaRanking::resortRows {} {
  .main printStatus "... resorting rows"
  if {$_lastSorting != ""} {
    eval sortRows $_lastSorting
  }
  .main printStatus ""
}

## ----------------------------------------------------------------------------
## _rowTagCommand
## ----------------------------------------------------------------------------
body YadaRanking::_rowTagCommand {row} {
  if {[expr $row % 2]} {
    return "colored"
  } else {
    return "default"
  }
}

## ----------------------------------------------------------------------------
## _computeStatistics
## ----------------------------------------------------------------------------
body YadaRanking::_computeStatistics {} {
  .main printStatus "... computing statistics"
  set table $itk_component(table)
  set rmax [$table cget -rows]

  set _totalTime 0

  foreach experimentName [$table getCell 1,0 $rmax,0] {
    if {$experimentName == ""} {
      continue
    }
    set experiment [.main getExperiment $experimentName]
    set _totalTime [expr $_totalTime + [$experiment cget -totalTime]]
  }

  .main printStatus ""
  _showStatistics
}

## ----------------------------------------------------------------------------
## _showStatistics
## ----------------------------------------------------------------------------
body YadaRanking::_showStatistics {} {
  $itk_component(totalTime) configure -state normal
  $itk_component(totalTime) delete 0 end
  $itk_component(totalTime) insert 0 "[timeToString $_totalTime]"
  $itk_component(totalTime) configure -state disabled
}
  
